#include <iostream>
#include <fstream>
#include <cereal/archives/binary.hpp>
#include <cereal/archives/json.hpp>

#include <tbb/task_scheduler_init.h>

#include "src/Border.H"
#include "src/Generate.H"
#include "tests/SimpleTests.H"
#include "tests/ConvergenceTests.H"
#include "tests/BorderOwnershipTests.H"
#include "tests/Util.H"

#include "tests/GaborTests.H"

#include "src/ThreadingUtil.H"
#include "config/Options.H"

#include <fenv.h> // for debugging nans

//! Loads an image, converts to floating point [0-255], and returns
//! the color image and its grayscale version
std::pair<cv::Mat, cv::Mat> loadImage( std::string const & file )
{
  auto image = cv::imread( file );
  // convert to float, leave in 0 to 255 range
  if( image.type() != CV_32FC3 )
    image.convertTo( image, CV_32F );

  cv::Mat grayscale;

  cv::cvtColor( image, grayscale, CV_BGR2GRAY );

  return {image, grayscale};
}

cv::Mat doClahe( cv::Mat const & image )
{
  cv::Mat input; image.convertTo( input, CV_8UC1 );

  cv::Ptr<cv::CLAHE> clahe = cv::createCLAHE();
  clahe->setClipLimit(4);
  cv::Mat dst;
  clahe->apply(input, dst);

  cv::Mat result; dst.convertTo( result, CV_32FC1 );

  return result;
}

OptionsStruct Options;

int main( int argc, char ** argv )
{
  // For debugging - crashes on nan or inf
  //feenableexcept(FE_INVALID | FE_OVERFLOW);

  // Handle options
  if( argc == 1 )
  {
    std::cerr << "ERROR: specify a configuraton file" << std::endl;
    return 1;
  }
  else
  {
    //std::ofstream file( argv[1] );
    //cereal::JSONOutputArchive ar( file );
    std::ifstream file( argv[1] );
    cereal::JSONInputArchive ar( file );
    ar( CEREAL_NVP(Options) );
  }

  // initialize pools
  {
    std::cout << "Initializing with " << Options.nthreads << " threads." << std::endl;
    tbb::task_scheduler_init init(Options.nthreads);
    threading::nthreads = Options.nthreads;
  }

  //const bool save = argc < 2;

  // 50, 40, for BO experiments
  // 30, 30 for CL vs Hebb
  const int width  = 10;
  const int height = 10;
  const double scale = 1.0;

  // 10.0f for small image dataset
  const float resolutionScaling = 10.0f;

  // Set up weight parameters
  bd::SystemState state;
  {
    { // Border Ownership
      auto & weights = state.boMaxWeights;
      weights.feedback = 4.0;
      weights.feedbackSkip = 6.0;
      weights.feedbackInhibitory = weights.feedbackSkipInhibitory = 10.0;
    }
    { // Grouping Neurons
      auto & weights = state.gnMaxWeights;
      weights.feedforward = 4.0;
      weights.feedback = weights.feedbackSkip = 4.0;
      weights.feedbackInhibitory = weights.feedbackSkipInhibitory = 10.0;
    }
    { // Proto Object Neurons
      auto & weights = state.poMaxWeights;
      weights.feedforward = 1.2;
      weights.feedforwardSkip = 10.0;
    }
  }

  bd::Border border( 4, // gabor parameters (# orientations)
                     1,    // size of grouping layer column
                     1,    // size of proto layer column
                     bd::CompetitionStrategy::Population, // grouping layer column strategy
                     scale, // scaling size for grouping
                     width, height,
                     resolutionScaling, // resolution scaling
                     state );

  // Hebb adjustments (only uncomment for Hebbian rule)
  //  Hebb needs much lower learning rate otherwise it will immediately learn current input
  //border.itsState.useHebbianLearning = true;
  //border.itsState.learningRate /= 1000;

  // CL adjustments (used for testing)
  //border.itsState.cl_spreading = false;
  //border.itsState.cl_unlearning = false;
  //border.itsState.cl_slt = false;
  //border.itsState.cl_ambiguity = false;
  //border.itsState.learningRate /= 100.0;

  // Dynamics adjustments (used for testing)
  //border.itsState.dy_skip_feedback = false;
  //border.itsState.dy_skip_feedforward = false;
  //border.itsState.dy_grouping_to_proto_feedforward = false;
  //border.itsState.dy_proto_to_grouping_feedback = false;
  border.itsState.dy_bo_to_grouping_feedforward = false;
  border.itsState.dy_grouping_to_bo_feedback = false;

  const int scaledWidth  = width  * resolutionScaling;
  const int scaledHeight = height * resolutionScaling;

  //auto image = bd::generateRectangle( height, scaledWidth, 10, 10, 10, 10 );
  cv::Mat blank = cv::Mat::zeros( scaledHeight, scaledWidth, CV_32FC1 );

  cv::Mat image = cv::Mat::zeros( scaledHeight, scaledWidth, CV_32FC1 );

  // square
  const int squareSize = 10 * scale * resolutionScaling;
  auto shape = bd::rotateShape( bd::generateRectangle( scaledWidth / 3, scaledHeight / 3, squareSize, squareSize, 1.0 ), 0.0_deg );
  // rectangle
  //auto shape = bd::rotateShape( bd::generateRectangle( scaledWidth / 3, scaledHeight / 3,
  //                                                     squareSize * 2.0, squareSize, 1.0 ), 30.0_deg );
  // triangle
  //auto shape = bd::rotateShape( bd::generateEquilateralTriangle( scaledWidth / 2.5, scaledWidth / 2.5, 5.0 * resolutionScaling ), 0.0_deg );

  bd::drawShapesInterpolated( image, {{shape, 255.0, false, (int)resolutionScaling}} );

  //bd::showImage( image, false );


  //auto inputImage = loadImage("../face.png").second;

  //cv::Rect window( 5, 5, 30, 30 );
  //cv::Mat ret;
  //image(window).copyTo( ret );
  //image = ret;


  //===========================================================================
  //===========================================================================
  //===========================================================================
/*
					@param[in] numScales The number of scales to use
					@param[in] scaleFirst The scale of the first filter
					@param[in] scaleStep the step between scales
					@param[in] numOrientations The number of orientations to use (evenly distributed) @param[in] sigma The variance of the gabor filter
					@param[in] freq The spatial frequency of the gabor filter
*/
  //std::cout << "gabor" << std::endl;
  //auto gabor = bd::Gabor( 1, 1, 1, 4,
  //                        0.6 * M_PI, // sigma
  //                        std::sqrt( 0.5f ) ); // freq
  //// 1, 1, 1, 4, .6 pi, sqrt 5 -> fairly even responses with 4 connected 1 width lines
  //itsGabor( numScales, 1, 1, numOrientations, 0.6 * M_PI, std::sqrt( 0.5f ) ),


  //gaborOrientationTest( gabor );
  //fftTests( image, gabor );

  //auto results = gabor.getFilterResponses( image );
  //bd::showImage( image );

  //for( auto const & vec : results )
  //  for( const auto & im : vec )
  //    bd::showImage( im );

  //bd::Sheet<bd::BorderOwnershipNeuron> s1( 100, 100, 500, bd::PopulationStrategy::PoissonDisk );
  //bd::Sheet<bd::BorderOwnershipNeuron> s2( 100, 100, 100, bd::PopulationStrategy::PoissonDisk );
  //auto s3 = bd::mergeSheets(s1, s2);

  //bd::showImage( bd::sheetNeuronImage(s1), false );
  //bd::showImage( bd::sheetNeuronImage(s2), false );
  //bd::showImage( bd::sheetNeuronImage(s3), false );

  //===========================================================================
  //========================================cd /lab/shane/shane/projects/Border/build_n12===================================
  //===========================================================================

  // Prints maxweights for convenience
  auto printMaxWeights = [](bd::SystemState const & state)
  {
    {
      auto & weights = state.boMaxWeights;
      std::cout << "BO maxWeights - FB: " << weights.feedback << ", FB Inhib: " << weights.feedbackInhibitory;
      std::cout << ", FB skip: " << weights.feedbackSkip << std::endl;
    }
    { // Grouping Neurons
      auto & weights = state.gnMaxWeights;
      std::cout << "GN maxWeights - FF: " << weights.feedforward << ", FB: " << weights.feedback << ", FB Inhib: " << weights.feedbackInhibitory;
      std::cout << ", FB skip: " << weights.feedbackSkip << std::endl;
    }
    { // Proto Object Neurons
      auto & weights = state.poMaxWeights;
      std::cout << "PO maxWeights - FF: " << weights.feedforward << ", FF Skip: " << weights.feedforwardSkip << std::endl;
    }
  };

  //bd::test::saveBD( border, "snarf.cereal" );
  bd::setDisplayState( border.itsState );

  if( argc < 4 )
  {
    std::string fileName = "current.cereal";

    if( argc >= 3 )
      fileName = std::string(argv[2]);

    bd::test::loadBD( border, fileName );
    printMaxWeights( border.itsState );
    bd::setDisplayState( border.itsState );

    //border.itsState.cl_ambiguity = false; // for testing

    // Threshold adjustment for raw images - gabor not set up for them, let threshold relax
    {
      auto adjustThreshold = [](std::shared_ptr<bd::Neuron> const & n, double p)
      {
        n->itsActivationThreshold -= (n->itsActivationThreshold - n->itsDecayThreshold) * p;
      };

//      for( auto & sheet : border.itsBOResponses )
//        for( auto & column : sheet )
//          for( auto & neuron : column )
//            adjustThreshold(neuron, 1.0);
//
//      for( auto & column : border.itsGNResponses )
//        for( auto & neuron : column )
//          adjustThreshold(neuron, 0.25);
//
      for( auto & column : border.itsProtoResponses )
        for( auto & neuron : column )
          adjustThreshold(neuron, 0.5);
    }

    const int scaledWidth = (int)border.itsGridWidth * border.itsResolutionScaling;
    const int scaledHeight = (int)border.itsGridHeight * border.itsResolutionScaling;

    cv::Mat blank = cv::Mat::zeros( scaledHeight, scaledWidth, CV_32FC1 );
    bd::test::presentStimulus( border, blank, 20 );

    int counter = 0;
    for( const auto & orientation : border.itsBOResponses )
    {
      double score, avg, stddev, median;
      std::tie( score, avg, stddev, median ) = bd::test::learnedPolarityScore( orientation, std::to_string(counter++) );
      std::cout << "Score: " << score << ", avg: " << avg << ", std: " << stddev << ", med: " << median << std::endl;

      //displayPolarity( orientation, border.itsGNResponses, false );
    }

    //testImage( border, scale, inputImage, TestConfiguration::Test );

    //testSquareOverlap( border );
    //testSquareCross( border );
    //testTetrisPiece1( border );
    //testConstant( border, image, 1, 0 );
    //testFullConvergence( border, scale );

    test4x4Configurations( border, scale, TestConfiguration::Test_output );

    // For Gen 3, square is 0, C shape is 10
    //
    // For Gen 4, sampled responses are:
    // XXX - 2
    //
    // XXXX  - 6
    // X
    //
    // XX - 9
    // XX
    //
    //  XX - 10
    // XX
    //
    // XXX - 16
    // X X
    //
    // XXXX - 26
    // X  X
    //
    // XXX - 36
    // X
    // X
    //
    //  XXX - 40 (38 for short version)
    // XX
    // X
    //
    // XXX - 39
    // XX
    // X
    //
    // XXXX - 43
    // X X
    // X
    //
    // XX   - 46
    // XXX
    // X
    //
    // X XX - 52 makes slight mistake
    // XXX
    // X
    //
    // XXXX - 59 (FIg2 D) 100.0
    // X XX
    // X
    //
    //  X  - 81 (Fig1 G)
    // XXX
    //  X
    //
    // X XX - 1831, only gets 84.69% correct (fig2 F)
    // X X
    // XXXX
    // X  X
    //
    // XX   - 927, 92.74% correct (fig2 G)
    // X  X
    // XXXX
    //  X
    //
    //  XX  - 800, 92.7% correct (fig2 H)
    //   XX
    // XXX
    //  X
    //test4x4SpecificConfiguration( border, scale, 0,
    //                              TestConfiguration::Test_output, SpecificTest::Scale );


    // basic shapes:
    // x - 0
    //
    // xx - 1
    //
    // xx - 4
    // x
    //
    // xx - 9
    // xx
    //
    // xxx - 5
    // x
    //
    // xxx - 7
    //  x
    //
    // xxx - 11
    // xx
    //
    // xxx - 16
    // x x

    //for( auto x : {0, 1, 4, 5, 7, 9, 11, 16} )
    //for( auto x : {2, 6, 9, 10, 26, 36, 40, 39, 43, 46, 52, 59, 81, 1831, 927, 800} )
    //for( auto x : {2, 6, 9, 10, 26, 36, 39, 43, 46, 52, 59, 81, 1831, 927, 800} )
    //{
    //  test4x4SpecificConfiguration( border, scale, x,
    //                                TestConfiguration::Test_output, SpecificTest::Static );
    //  bd::test::presentStimulus( border, blank, 20 );
    //}

    return 0;
  }

  // For resuming
  if( argc >= 4 )
  {
    if( std::string(argv[2]) == "--resume" )
    {
      std::string filename = std::string(argv[3]);
      std::cout << "Resuming " << filename << std::endl;
      bd::test::loadBD( border, filename );
    }
  }

//  std::string fileName = "current.cereal";
//  if( argc >= 3 )
//    fileName = std::string(argv[2]);
//  bd::test::loadBD( border, fileName );

  //bd::test::loadBD( border, "current.cereal" );

  //testConstant( border, image, 10, 0 );
  //testFlip( border, scale );
  //testFlipSlide( border, blank );
  //testSlide( border, blank );
  //testSlideConvergence( border, blank, true );

  //testSimpleConvergence( border );

  //testCrossConvergence( border );
  //testFullConvergence( border, scale );
  //testFullConvergenceSpiral( border, blank );
  //testConvergenceSpecificRow( border, blank );

  //testSquareOverlap( border );
  //testTetrisPiece1( border );

  //testFullConvergenceImages( border, Options.imagesDir );
  //

  test4x4Configurations( border, scale, TestConfiguration::Train );

  return 0;
}
