#include <iostream>
#include <fstream>
#include <cereal/archives/binary.hpp>
#include <cereal/archives/json.hpp>

#include <tbb/task_scheduler_init.h>

#include "tests/V1Tests.H"
#include "tests/Util.H"
#include "src/Combined.H"
#include "src/Generate.H"
#include "tests/CombinedTests.H"

#include "src/ThreadingUtil.H"
#include "config/Options.H"

//! Loads an image, converts to floating point [0-255], and returns
//! the color image and its grayscale version
std::pair<cv::Mat, cv::Mat> loadImage( std::string const & file )
{
  auto image = cv::imread( file );
  // convert to float, leave in 0 to 255 range
  if( image.type() != CV_32FC3 )
    image.convertTo( image, CV_32F );

  cv::Mat grayscale;

  cv::cvtColor( image, grayscale, CV_BGR2GRAY );

  return {image, grayscale};
}

cv::Mat doClahe( cv::Mat const & image )
{
  cv::Mat input; image.convertTo( input, CV_8UC1 );

  cv::Ptr<cv::CLAHE> clahe = cv::createCLAHE();
  clahe->setClipLimit(4);
  cv::Mat dst;
  clahe->apply(input, dst);

  cv::Mat result; dst.convertTo( result, CV_32FC1 );

  return result;
}

OptionsStruct Options;

int main( int argc, char ** argv )
{
  // Handle options
  if( argc == 1 )
  {
    std::cerr << "ERROR: specify a configuraton file" << std::endl;
    return 1;
  }
  else
  {
    std::ifstream file( argv[1] );
    cereal::JSONInputArchive ar( file );
    ar( CEREAL_NVP(Options) );
  }

  // initialize pools
  {
    std::cout << "Initializing with " << Options.nthreads << " threads." << std::endl;
    tbb::task_scheduler_init init(Options.nthreads);
    threading::nthreads = Options.nthreads;
  }

  const int width  = 30;
  const int height = 30;

  bd::Combined combined( 4, width, height );

  //auto image = bd::generateRectangle( height, width, 10, 10, 10, 10 );
  cv::Mat blank = cv::Mat::zeros( height, width, CV_32FC1 );

  cv::Mat image = cv::Mat::zeros( height, width, CV_32FC1 );

  // square
  auto shape = bd::rotateShape( bd::generateRectangle( 10, 10, 10, 10 ), 0.0_deg );
  // rectangle
  //auto shape = bditsState.lateralTotalWeight::rotateShape( bd::generateRectangle( 10, 10, 20, 10 ), 45.0_deg );
  // triangle
  //auto shape = bd::rotateShape( bd::generateEquilateralTriangle( 15, 15, 5 ), 0.0_deg );

  bd::drawShape( image, shape );
  //bd::showImage( image, false );

  if( argc < 3 ) // display
  {
    bd::test::loadBD( combined, "current_combined.cereal" );
    //bd::showImage( sheetNeuronImage( combined.itsEdgeResponses, 8.0 ), false, true, "neurons" );
    //bd::showImage( getV1OrientationPreferencesImage( combined ), false, true, "colors, yo" );
    showConnections( combined.itsEdgeResponses, false, true );

    while( true )
    {
      const double theta = bd::uniformReal( 0.0, 2 * M_PI )();
      auto line  = bd::rotateShape( bd::generateLine( -1, width / 2, width - 1 ), theta );
      cv::Mat image = cv::Mat::zeros( height, width, CV_32FC1 );
      bd::drawShape( image, line );


      //for( int i = 0; i < 10; ++i )
      {
        bd::test::presentStimulus( combined, image, 10 );
        bd::showImage( image, false, false, "input" );
        bd::showImage( sheetActivationImage( combined.itsOnOffResponses, 8.0 ), false, false, "on" );
        bd::showImage( sheetActivationImage( combined.itsOffOnResponses, 8.0 ), false, false, "off" );
        bd::showImage( sheetActivationImage( combined.itsEdgeResponses, 8.0 ), false, true, "v1" );
        bd::showImage( sheetActivationImage( combined.itsEdgeResponses ), false, true, "v1_no_scale" );
      }

      //showConnections( combined.itsEdgeResponses, false, false );
      bd::test::presentStimulus( combined, cv::Mat::zeros(height,width,CV_32FC1), 500 );
    }

  }
  else // run
    testFullCombinedConvergence( combined, blank );


  return 0;
}
