#include <iostream>
#include <fstream>
#include <unordered_map>
#include <cereal/archives/binary.hpp>
#include <cereal/archives/json.hpp>
#include <cereal/types/map.hpp>

#include <tbb/task_scheduler_init.h>

#include "src/ThreadingUtil.H"
#include "config/Options.H"

#include "src/Sheet.H"
#include "src/EdgeNeuron.H"
#include "src/StaticNeuron.H"
#include "src/Wiring.H"

#include "tests/Util.H"

/* This test is for the state network in the paper, showing off the dynamics
   of conflict learning vs Hebbian learning.

   We want to keep track of various states the system enters for a 2 neuron 2 mod neuron case:

   (in terms of strongly learned connections)
   0SL - initial (one case)
   1SL - one learns one modulatory (four cases)
   2SL
    Split   - One learns both modulatory (two cases)
    Shared  - Both learn same modulatory (two cases)
    Desired - End state (two cases)
  3SL - one learns both modulatory (split), one learns single modulatory (1SL), (four cases)
  4SL - both learn to both modulatory (one case)

  To keep track of this, at each stimulus onset we will record how the system transitions.

  SL is defined as greater than .1 and other neuron is less than half my weight

  Note that for this to run properly the threshold in BorderOwnershipNeuron.C should be disabled:
    if( preClamp < itsLTThreshold.pre )
      preClamp = 0.0;
*/

enum StateName
{
  s0SL,         //0
  s1SL,         //1
  s2SL_DESIRED, //2
  s2SL_SPLIT,   //3
  s2SL_SHARED,  //4
  s3SL,         //5
  s4SL,         //6
};

using StateTransitionsMap = std::unordered_map<int, std::unordered_map<int, int>>;


void printStateTransitions( StateTransitionsMap const & transitions )
{
  cereal::JSONOutputArchive ar(std::cout);
  ar(transitions);
  std::cout << std::endl;
}

struct DynamicsNetwork
{
  DynamicsNetwork( bd::SystemState const & state ) :
    itsInputSheet( 1, 1 ), itsCompetitiveSheet( 1, 1 ), itsModulatorySheet( 2, 1 ), itsState( state )
  {
    // feedforward from Input to Competititve, fixed weight
    bd::wireSheetRingUniform( itsCompetitiveSheet, itsInputSheet, 0, 5, 1.0, 1.0, false,
                              bd::ConnectionFlags::Static | bd::ConnectionFlags::Feedforward );


    // inhibition between Competitors
    bd::wireSheetRingUniform( itsCompetitiveSheet, itsCompetitiveSheet, 0, 5.0, 1.0,
                              -state.v1lateralTotalWeight, // lateral inhibition amount, static
                              false, bd::ConnectionFlags::Static | bd::ConnectionFlags::Inhibitory | bd::ConnectionFlags::Lateral );

    // feedback from Modulatory to Competitive
    bd::wireSheetRingUniform( itsCompetitiveSheet, itsModulatorySheet, 0, 5, 1.0, 0.1, true,
                              bd::ConnectionFlags::Dynamic | bd::ConnectionFlags::Feedback );

    // Hard code initial BCM threshold
    for( auto & vec : itsCompetitiveSheet )
      for( auto & neuron : vec )
        neuron->itsBCMThreshold.pre() = neuron->itsBCMThreshold.post() = 0.5;
  }

  DynamicsNetwork( bd::SystemState const & state, StateTransitionsMap const & transitions ) :
    DynamicsNetwork( state )
  {
    itsTransitions = transitions;
  }

  // Randomly turns on one of the modulatory inputs with some optional bias and noise
  void onset( double bias, double noiseScale )
  {
    // Generates a random number between 0 and 2 with some bias and then
    // presents the input
    const double rng = bd::boundedNormalReal( 1.0, bias, 0.0, 2.0 )();
    const int off = static_cast<int>( rng );

    auto noise = bd::uniformReal( 0.0, 1.0 );
    auto interpolate = [](double v, double n, double p)
    { return v * (1.0 - p) + n * p; };

    // Turn on input
    itsInputSheet[0][0]->setActivation( 1 );

    // Turn on one of the modulatory
    itsModulatorySheet[0][0]->setActivation( interpolate( 1.0, noise(), noiseScale ) );
    itsModulatorySheet[1][0]->setActivation( interpolate( 1.0, noise(), noiseScale ) );

    itsModulatorySheet[off][0]->setActivation( interpolate( 0.0, noise(), noiseScale ) );
  }

  void blank( int numIters = 5 )
  {
    itsInputSheet[0][0]->setActivation( 0.0 );
    itsModulatorySheet[0][0]->setActivation( 0.0 );
    itsModulatorySheet[1][0]->setActivation( 0.0 );
    iterate( numIters );
  }

  void iterate( int numIters )
  {
    StateName curState = getState();
    for( int i = 0; i < numIters; ++i )
    {
      // update
      threading::parallel_for( 0lu, itsCompetitiveSheet.size(), [&](size_t i)
      {
        auto & vec = itsCompetitiveSheet[i];
        for( auto & neuron : vec )
          neuron->update( itsState );
      } );
      //learn
      if( itsState.learningRule == bd::LearningRule::Sanger )
        sangerLearn();
      else
      {
        threading::parallel_for( 0lu, itsCompetitiveSheet.size(), [&](size_t i)
        {
          auto & vec = itsCompetitiveSheet[i];
          for( auto & neuron : vec )
            neuron->learn( itsState );
        } );
      }
      // finalize
      bd::finalizeNeuronVariables();
//      threading::parallel_for( 0lu, itsCompetitiveSheet.size(), [&](size_t i)
//      {
//        auto & vec = itsCompetitiveSheet[i];
//        for( auto & neuron : vec )
//          neuron->finalize();
//      } );
    }

      auto newState = getState();
      ++itsTransitions[curState][newState];
      curState = newState;
  }

  //! Performs a weight update using Sanger's Generalized Hebbian learning rule
  /*! This can't occur within Neuron or its derivations because it requires information
      inaccessible via the connections of a single Neuron (the activation of all other neurons),
      and also requires a known ordering (index) for each neuron.

      Kind of a hacky implementation; would require some re-architecting to get this into a more general location */
  void sangerLearn()
  {
    threading::parallel_for( 0lu, itsCompetitiveSheet.size(), [&](size_t index)
    {
      auto & vec = itsCompetitiveSheet[index];
      const int flags = bd::ConnectionFlags::Dynamic | bd::ConnectionFlags::Feedback;

      for( size_t i = 0; i < vec.size(); ++i )
      {
        // The update to apply to each neuron:
        //  dW_ij = x_j y_i - y_i SUM(k=1 to i)( w_kj y_k ) (i = self, j = input, k = other neuron at same level)
        auto const & neuron = vec[i];
        neuron->itsInput.post() = neuron->itsInput.pre();

        const double learningRate = itsState.learningRate;
        const double maxWeight = itsState.maxWeight;
        const double currentActivation = neuron->activationSanger();

        for( size_t j = 0; j < neuron->itsInput.pre().size(); ++j )
        {
          auto const & preEdge = neuron->itsInput.pre()[j];
          auto & postEdge = neuron->itsInput.post()[j];

          auto const inputNeuron = preEdge.neuron; // aka x_j

          if( preEdge.hasFlag( flags ) )
          {
            double decay = 0.0;

            for( size_t k = 0; k <= i; ++k )
            {
              auto const kNeuron = vec[k];
              // Find w_kj
              // We can only look up incoming weights, so we need to search for
              // the link that matches the current incoming neuron
              for( size_t k_i = 0; k_i < kNeuron->itsInput.pre().size(); ++k_i )
              {
                auto const k_i_Neuron = kNeuron->itsInput.pre()[k_i].neuron;
                if( *inputNeuron == *k_i_Neuron )
                {
                  auto const & k_edge = kNeuron->itsInput.pre()[k_i];
                  decay += k_edge.weight * kNeuron->activation();

                  break;
                }
              }
            }

            decay *= currentActivation;

            const double delta = preEdge.neuron->activation() * currentActivation - decay;
            postEdge.weight = bd::clamp( preEdge.weight + learningRate * delta, 0, maxWeight );
          } // end if flags
        } // end for over input
      } // end for over neurons
    } ); // end parallel for
  }

  StateName getState() const
  {
    bool sl[2][2] = { 0 }; // whether neuron-mod pair is strongly learned
    int numSl = 0;
    for( int nIdx = 0; nIdx < 2; ++nIdx) // loop competitive neurons
    {
      double maxWeight = 0;
      for( const auto & in : itsCompetitiveSheet[0][nIdx]->itsInput.pre() ) // loop mod
      {
        if (in.hasFlag(bd::ConnectionFlags::Feedback))
        {
          maxWeight = std::max(maxWeight, in.weight);
        }
      }

      int fbIdx = 0;
      for( const auto & in : itsCompetitiveSheet[0][nIdx]->itsInput.pre() ) // loop mod
      {
        if (in.hasFlag(bd::ConnectionFlags::Feedback))
        {
          bool check = false;

          if( itsState.learningRule == bd::LearningRule::Hebbian )
            check = in.weight > 0.6 && in.weight > maxWeight / 2;
          else
            check = in.weight > 0.1 && in.weight > maxWeight / 2;

          if (check)
          {
            sl[nIdx][fbIdx] = true;
            numSl++;
          }
          fbIdx++;
        }
      }
    }

    switch (numSl)
    {
      case 0:
        return StateName::s0SL;
        break;
      case 1:
        return StateName::s1SL;
        break;
      case 2:
        if ((sl[0][0] && sl[0][1]) || (sl[1][0] && sl[1][1]))
          return StateName::s2SL_SPLIT;
        else if ((sl[0][0] && sl[1][0]) || (sl[0][1] && sl[1][1]))
          return StateName::s2SL_SHARED;
        else
          return StateName::s2SL_DESIRED;
        break;
      case 3:
        return StateName::s3SL;
        break;
      case 4:
        return StateName::s4SL;
        break;
    }

    throw "Did not get a state!";
  } // end getState

  void printNetwork()
  {
    for( int i = 0; i < (int)itsCompetitiveSheet.size(); ++i )
    {
      auto & vec = itsCompetitiveSheet[i];
      for( auto & neuron : vec )
      {
        neuron->printNeuronInfo();
        neuron->printConnections( "Feedforward Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedforward ); } );
        neuron->printConnections( "Feedback Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedback ); } );
        neuron->printConnections( "Lateral Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Lateral ); } );
      }
    }
  }

  StateTransitionsMap const & getTransitions() const { return itsTransitions; }

//  void dumpWeights( std::string const & prefix )
//  {
//    bd::test::dumpWeights( itsCompetitiveSheet, prefix );
//  }

  bd::Sheet<bd::StaticNeuron> itsInputSheet;
  bd::Sheet<bd::BorderOwnershipNeuron> itsCompetitiveSheet;
  bd::Sheet<bd::StaticNeuron> itsModulatorySheet;
  bd::SystemState itsState;

  StateTransitionsMap itsTransitions;

  template <class Archive>
  void serialize( Archive & ar )
  {
    ar( itsInputSheet, itsCompetitiveSheet, itsModulatorySheet, itsState );
  }
};

void load( DynamicsNetwork & network, std::string const & name )
{
  std::cout << "Loading " << name << std::endl;
  std::ifstream file(name, std::ios::binary );
  cereal::BinaryInputArchive ar( file );
  ar( network );
}

void save( DynamicsNetwork & network, std::string const & name )
{
  //std::cout << "Saving " << name << std::endl;
  std::ofstream file(name, std::ios::binary );
  cereal::BinaryOutputArchive ar( file );
  ar( network );
}

OptionsStruct Options;

int main( int argc, char ** argv )
{
  // Handle options
  if( argc == 1 )
  {
    std::cerr << "ERROR: specify a configuraton file" << std::endl;
    return 1;
  }
  else
  {
    std::ifstream file( argv[1] );
    cereal::JSONInputArchive ar( file );
    ar( CEREAL_NVP(Options) );
  }

  // initialize pools
  {
    std::cout << "Initializing with " << Options.nthreads << " threads." << std::endl;
    tbb::task_scheduler_init init(Options.nthreads);
    threading::nthreads = Options.nthreads;
  }

  // flat lowering params: lateral: 5, learn: .01

  auto getHebbianParameters = []()
  {
    bd::SystemState state;
    state.v1lateralTotalWeight = 1.0;
    state.learningRate = 0.01;
    state.boMaxWeights.feedback = 1.0;
    state.learningRule = bd::LearningRule::Hebbian;

    return state;
  };

  auto getSangerParameters = []()
  {
    bd::SystemState state;
    state.v1lateralTotalWeight = 1.0;
    state.learningRate = 0.01;
    state.boMaxWeights.feedback = 1.0;
    state.maxWeight = 1.0; // no weight normalization in Sanger/Oja
    state.learningRule = bd::LearningRule::Sanger;

    return state;
  };

  auto getBCMParameters = []()
  {
    // Hard code threshold to start at 0.5, in between inhibited and active long term averages
    bd::SystemState state;
    state.v1lateralTotalWeight = 1.0;
    state.learningRate = 0.01;
    state.boMaxWeights.feedback = 1.0;
    state.maxWeight = 1.0; // no weight normalization in BCM
    state.learningRule = bd::LearningRule::BCM;

    return state;
  };

  auto getStandardParameters = []()
  {
    bd::SystemState state;
    state.v1lateralTotalWeight = 1.0;
    state.boMaxWeights.feedback = 1.0;
    state.learningRate = 0.01;

    return state;
  };

  //////////////////////////////////////////////////////////////////////////////
  // Experiments
  //////////////////////////////////////////////////////////////////////////////
  auto doTest = [](DynamicsNetwork & network, double bias, double noiseScale, std::string const & fileSuffix)
  {
    const int numPresentIters = 100;
    const int numOverallIters = 101;
    const int numInspectIters = Options.iters;

    for( int i = 0; i < numOverallIters; ++i )
    {
      network.onset( bias, noiseScale );
      network.iterate( numPresentIters );

      //for( int j = 0; j < numPresentIters; ++j )
      //{
      //  network.iterate( 1 );
      //  std::cout << j << std::endl;


      //  network.printNetwork();
      //  std::string x;
      //  std::cout << "ayylamo**************************************" << std::endl;
      //  std::cin >> x;
      //}

      network.blank(10);

      if( i % numInspectIters == 0 && i != 0 )
      {
        //save( network, Options.output + "_" + std::to_string(i) + fileSuffix );
      }
    }
  };

  //////////////////////////////////////////////////////////////////////////////
  // Running Code
  //////////////////////////////////////////////////////////////////////////////
  if( argc < 4 )
  {
    std::string fileName = "current_dynamics.cereal";

    if( argc >= 3 )
      fileName = std::string(argv[2]);

    bd::SystemState state;
    if( fileName.find( "hebbian" ) != std::string::npos )
      state = getHebbianParameters();
    else if( fileName.find( "bcm" ) != std::string::npos )
      state = getBCMParameters();
    else if( fileName.find( "sanger" ) != std::string::npos )
      state = getSangerParameters();
    else
      state = getStandardParameters();


    DynamicsNetwork network( state );
    load( network, fileName );

    if( network.itsState.learningRule == bd::LearningRule::Hebbian )
      std::cout << "Hebbian configuration detected" << std::endl;
    if( network.itsState.learningRule == bd::LearningRule::BCM )
      std::cout << "BCM configuration detected" << std::endl;
    if( network.itsState.learningRule == bd::LearningRule::Sanger )
      std::cout << "Sanger configuration detected" << std::endl;
    else if( network.itsState.learningRule == bd::LearningRule::Conflict )
      std::cout << "Standard configuration detected" << std::endl;
    network.printNetwork();
  }
  else
  {
    const int MAX_ITERS = 30;

    // Uncomment this to do bias testing
    //double stdBias[] = {10, 1, .6, .4, .3, .2, .1};
    double stdBias[] = {100};

    auto coolToString = [](double d, int padding)
    {
      auto s = std::to_string(d);
      int numZeros = padding - (int) s.substr(0, s.find_first_of(".")).size();
      for (int i = 0; i < numZeros; ++i)
        s = "0" + s;
      return s;
    };

    for( const auto std : stdBias )
    {
      std::cout << "On bias " << std << std::endl;
      auto stdStr = coolToString(std, 2);

      auto testRule = [&](bd::SystemState const & state, std::string const & ruleName)
      {
        StateTransitionsMap transitions;

        std::cout << "************ " << ruleName << " ************" << std::endl;

        for( int i = 0; i < MAX_ITERS; ++i )
        {
          auto iterStr = std::to_string( i );
          DynamicsNetwork network( state, transitions );
          doTest( network, std, 0.0, "_dynamics_" + ruleName + "_" + iterStr + "_" + stdStr + ".cereal" );
          transitions = network.getTransitions();
          //network.printNetwork();
          //network.dumpWeights( "dynamics_hebbian_" + iterStr + "_" + stdStr );
        }

        printStateTransitions( transitions );

        std::cout << std::endl;
        std::cout << "*******************************************************************************" << std::endl;
        std::cout << "*******************************************************************************" << std::endl;
        std::cout << "*******************************************************************************" << std::endl;
        std::cout << "*******************************************************************************" << std::endl;
      };

      //testRule( getHebbianParameters(), "hebbian" );
      testRule( getBCMParameters(), "bcm" );
      //testRule( getSangerParameters(), "sanger" );
      //testRule( getStandardParameters(), "standard" );

      std::cout << std::endl;
      std::cout << std::endl;
      for( int i = 0; i < 2; ++i )
        std::cout << "===============================================================================" << std::endl;
      std::cout << std::endl;
    }

    for( int i = 0; i < 8; ++i )
      std::cout << "===============================================================================" << std::endl;

//    // Uncomment this to do noise testing
//    double noiseScale[] = {.01, .1, .2, .3, .4, .6, .8};
//    //double noiseScale[] = {1.0};
//
//    for( const auto std : noiseScale )
//    {
//      std::cout << "On noise " << std << std::endl;
//      auto stdStr = coolToString(std, 2);
//
//      {
//        StateTransitionsMap transitions;
//
//        for( int i = 0; i < MAX_ITERS; ++i )
//        {
//          auto iterStr = std::to_string( i );
//          DynamicsNetwork network( getHebbianParameters(), transitions );
//          doTest( network, 10, std, "_dynamics_noise_hebbian_" + iterStr + "_" + stdStr + ".cereal" );
//          transitions = network.getTransitions();
//          //network.printNetwork();
//          //network.dumpWeights( "dynamics_noise_hebbian_" + iterStr + "_" + stdStr );
//        }
//
//        printStateTransitions( transitions );
//      }
//
//      std::cout << "*******************************************************************************" << std::endl;
//      std::cout << "*******************************************************************************" << std::endl;
//      std::cout << "*******************************************************************************" << std::endl;
//      std::cout << "*******************************************************************************" << std::endl;
//
//      {
//        StateTransitionsMap transitions;
//
//        for( int i = 0; i < MAX_ITERS; ++i )
//        {
//          auto iterStr = std::to_string( i );
//          DynamicsNetwork network( getStandardParameters(), transitions );
//          doTest( network, 10, std, "_dynamics_noise_standard_" + iterStr + "_" + stdStr + ".cereal" );
//          transitions = network.getTransitions();
//          //network.printNetwork();
//          //network.dumpWeights( "dynamics_noise_standard_" + iterStr + "_" + stdStr );
//        }
//
//        printStateTransitions( transitions );
//      }
//
//      std::cout << std::endl;
//      std::cout << std::endl;
//      for( int i = 0; i < 2; ++i )
//        std::cout << "===============================================================================" << std::endl;
//      std::cout << std::endl;
//    }
  } // end main else

  //    // below is normal behavior
  //    bd::SystemState state;
  //    //state = getHebbianParameters();
  //    state = getStandardParameters();
  //
  //    DynamicsNetwork network( state );


  return 0;
}
