#include <iostream>
#include <fstream>
#include <cereal/archives/binary.hpp>
#include <cereal/archives/json.hpp>

#include <tbb/task_scheduler_init.h>

#include "src/ThreadingUtil.H"
#include "config/Options.H"

#include "src/Sheet.H"
#include "src/EdgeNeuron.H"
#include "src/StaticNeuron.H"
#include "src/Wiring.H"

#include "tests/Util.H"

struct ThoughtNetwork
{
  ThoughtNetwork( bd::SystemState const & state ) :
    itsInputSheet( 3, 1 ), itsHiddenSheet( 3, 1 ), itsState( state )
  {
    // inhibition
    bd::wireSheetRingUniform( itsHiddenSheet, itsHiddenSheet, 0, 5.0, 1.0,
                              -state.v1lateralTotalWeight, // lateral inhibition amount, static
                              //false, bd::ConnectionFlags::Static | bd::ConnectionFlags::Inhibitory | bd::ConnectionFlags::Lateral );
                              false, bd::ConnectionFlags::Dynamic | bd::ConnectionFlags::Inhibitory | bd::ConnectionFlags::Lateral );

    // feedforward
    bd::wireSheetRingUniform( itsHiddenSheet, itsInputSheet, 0, 5, 1.0, state.v1MaxFFWeight, true,
                              bd::ConnectionFlags::Dynamic | bd::ConnectionFlags::Feedforward );
  }

  void onset( double bias, double noiseScale )
  {
    // Generates a random number between 1 and 3 with some bias and then
    // presents the input
    const double rng = bd::boundedNormalReal( 1.5, bias, 0, 3.0 )();
    const int off = static_cast<int>( rng );

    //auto noise = bd::normalReal();
    auto noise = bd::uniformReal( 0.0, 1.0 );
    auto interpolate = [](double v, double n, double p)
    { return v * (1.0 - p) + n * p; };

    //itsInputSheet[0][0]->setActivation( 0.0 );
    //itsInputSheet[1][0]->setActivation( 0.0 );
    //itsInputSheet[2][0]->setActivation( 0.0 );
    //itsInputSheet[off][0]->setActivation( 1.0 );

    itsInputSheet[0][0]->setActivation( interpolate( 1.0, noise(), noiseScale ) );
    itsInputSheet[1][0]->setActivation( interpolate( 1.0, noise(), noiseScale ) );
    itsInputSheet[2][0]->setActivation( interpolate( 1.0, noise(), noiseScale ) );

    itsInputSheet[off][0]->setActivation( interpolate( 0.0, noise(), noiseScale ) );
  }

  void blank( int numIters = 5 )
  {
    itsInputSheet[0][0]->setActivation( 0.0 );
    itsInputSheet[1][0]->setActivation( 0.0 );
    itsInputSheet[2][0]->setActivation( 0.0 );
    iterate( numIters );
  }

  void iterate( int numIters )
  {
    for( int i = 0; i < numIters; ++i )
    {
      threading::parallel_for( 0lu, itsHiddenSheet.size(), [&](size_t i)
      {
        auto & vec = itsHiddenSheet[i];
        for( auto & neuron : vec )
          neuron->update( itsState );
      } );
      threading::parallel_for( 0lu, itsHiddenSheet.size(), [&](size_t i)
      {
        auto & vec = itsHiddenSheet[i];
        for( auto & neuron : vec )
          neuron->learn( itsState );
      } );

      bd::finalizeNeuronVariables();
      //threading::parallel_for( 0lu, itsHiddenSheet.size(), [&](size_t i)
      //{
      //  auto & vec = itsHiddenSheet[i];
      //  for( auto & neuron : vec )
      //    neuron->finalize();
      //} );
    }
  }

  void printNetwork()
  {
    for( int i = 0; i < itsHiddenSheet.size(); ++i )
    {
      auto & vec = itsHiddenSheet[i];
      for( auto & neuron : vec )
      {
        neuron->printNeuronInfo();
        neuron->printConnections( "Feedforward Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedforward ); } );
        neuron->printConnections( "Lateral Inhibitory Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Inhibitory ); } );
      }
    }
  }

  void dumpWeights( std::string const & prefix )
  {
    bd::test::dumpWeights( itsHiddenSheet, prefix );
  }

  bd::Sheet<bd::StaticNeuron> itsInputSheet;
  bd::Sheet<bd::EdgeNeuron>   itsHiddenSheet;
  bd::SystemState itsState;

  template <class Archive>
  void serialize( Archive & ar )
  {
    ar( itsInputSheet, itsHiddenSheet, itsState );
  }
};

void load( ThoughtNetwork & network, std::string const & name )
{
  std::cout << "Loading " << name << std::endl;
  std::ifstream file(name, std::ios::binary );
  cereal::BinaryInputArchive ar( file );
  ar( network );
}

void save( ThoughtNetwork & network, std::string const & name )
{
  //std::cout << "Saving " << name << std::endl;
  std::ofstream file(name, std::ios::binary );
  cereal::BinaryOutputArchive ar( file );
  ar( network );
}

OptionsStruct Options;

int main( int argc, char ** argv )
{
  // Handle options
  if( argc == 1 )
  {
    std::cerr << "ERROR: specify a configuraton file" << std::endl;
    return 1;
  }
  else
  {
    std::ifstream file( argv[1] );
    cereal::JSONInputArchive ar( file );
    ar( CEREAL_NVP(Options) );
  }

  // initialize pools
  {
    std::cout << "Initializing with " << Options.nthreads << " threads." << std::endl;
    tbb::task_scheduler_init init(Options.nthreads);
    threading::nthreads = Options.nthreads;
  }

  // flat lowering params: lateral: 5, learn: .01

  auto getHebbianParameters = []()
  {
    bd::SystemState state;
    state.v1MaxFFWeight = 1.0;
    state.v1lateralTotalWeight = 2.0;
    state.learningRate = 0.01;
    state.learningRule = bd::LearningRule::Hebbian;

    return state;
  };

  auto getStandardParameters = []()
  {
    bd::SystemState state;
    state.v1MaxFFWeight = 1.0;
    state.v1lateralTotalWeight = 2.0; // .5 normally
    state.learningRate = 0.01;

    return state;
  };

  //////////////////////////////////////////////////////////////////////////////
  // Experiments
  //////////////////////////////////////////////////////////////////////////////
  auto doTest = [](ThoughtNetwork & network, double bias, double noiseScale, std::string const & fileSuffix)
  {
    const int numPresentIters = 10;
    const int numOverallIters = 20001;
    const int numInspectIters = Options.iters;

    for( int i = 0; i < numOverallIters; ++i )
    {
      network.onset( bias, noiseScale );
      network.iterate( numPresentIters );

//      for( int j = 0; j < numPresentIters; ++j )
//      {
//        network.iterate( 1 );
//        std::cout << j << std::endl;
//
//
//        network.printNetwork();
//        std::string x;
//        std::cout << "ayylamo**************************************" << std::endl;
//        std::cin >> x;
//      }

      network.blank(10);

      if( i % numInspectIters == 0 && i != 0 )
      {
        //save( network, Options.output + "_" + std::to_string(i) + fileSuffix );
      }
    }
  };

  //////////////////////////////////////////////////////////////////////////////
  // Running Code
  //////////////////////////////////////////////////////////////////////////////
  if( argc < 4 )
  {
    std::string fileName = "current_thought.cereal";

    if( argc >= 3 )
      fileName = std::string(argv[2]);

    bd::SystemState state;
    if( fileName.find( "hebbian" ) != std::string::npos )
      state = getHebbianParameters();
    else
      state = getStandardParameters();


    ThoughtNetwork network( state );
    load( network, fileName );

    if( network.itsState.learningRule == bd::LearningRule::Hebbian )
      std::cout << "Hebbian configuration detected" << std::endl;
    if( network.itsState.learningRule == bd::LearningRule::Conflict )
      std::cout << "Standard configuration detected" << std::endl;

    network.printNetwork();
  }
  else
  {
    const int MAX_ITERS = 30;
    for( int i = 0; i < MAX_ITERS; ++i )
    {
      // Uncomment this to do bias testing
      double stdBias[] = {10, 1, .6, .4, .3, .2, .1};
      //double stdBias[] = {.4, .3, .2, .1};
      auto coolToString = [](double d, int padding)
      {
        auto s = std::to_string(d);
        int numZeros = padding - (int) s.substr(0, s.find_first_of(".")).size();
        for (int i = 0; i < numZeros; ++i)
          s = "0" + s;
        return s;
      };

      auto iterStr = std::to_string( i );

      for( const auto std : stdBias )
      {
        std::cout << "On bias " << std << std::endl;
        auto stdStr = coolToString(std, 2);

        {
          ThoughtNetwork network( getHebbianParameters() );
          doTest( network, std, 0.0, "_thought_hebbian_" + iterStr + "_" + stdStr + ".cereal" );
          network.printNetwork();
          network.dumpWeights( "thought_hebbian_" + iterStr + "_" + stdStr );
        }

        std::cout << "*******************************************************************************" << std::endl;
        std::cout << "*******************************************************************************" << std::endl;
        std::cout << "*******************************************************************************" << std::endl;
        std::cout << "*******************************************************************************" << std::endl;

        {
          ThoughtNetwork network( getStandardParameters() );
          doTest( network, std, 0.0, "_thought_standard_" + iterStr + "_" + stdStr + ".cereal" );
          network.printNetwork();
          network.dumpWeights( "thought_standard_" + iterStr + "_" + stdStr );
        }

        std::cout << std::endl;
        std::cout << std::endl;
        std::cout << std::endl;
        std::cout << std::endl;
      }

      for( int i = 0; i < 8; ++i )
        std::cout << "===============================================================================" << std::endl;

      // Uncomment this to do noise testing
      //double noiseScale[] = {.01, .4, .5, .6, .7, .8, .9};
      double noiseScale[] = {.01, .1, .2, .3, .4, .6, .8};
      //double noiseScale[] = {1.0};

      for( const auto std : noiseScale )
      {
        std::cout << "On noise " << std << std::endl;
        auto stdStr = coolToString(std, 2);

        {
          ThoughtNetwork network( getHebbianParameters() );
          doTest( network, 10, std, "_thought_noise_hebbian_" + iterStr + "_" + stdStr + ".cereal" );
          network.printNetwork();
          network.dumpWeights( "thought_noise_hebbian_" + iterStr + "_" + stdStr );
        }

        std::cout << "*******************************************************************************" << std::endl;
        std::cout << "*******************************************************************************" << std::endl;
        std::cout << "*******************************************************************************" << std::endl;
        std::cout << "*******************************************************************************" << std::endl;

        {
          ThoughtNetwork network( getStandardParameters() );
          doTest( network, 10, std, "_thought_noise_standard_" + iterStr + "_" + stdStr + ".cereal" );
          network.printNetwork();
          network.dumpWeights( "thought_noise_standard_" + iterStr + "_" + stdStr );
        }

        std::cout << std::endl;
        std::cout << std::endl;
        std::cout << std::endl;
        std::cout << std::endl;
      }

    }
  } // end MAX_ITERS loop

  //    // below is normal behavior
  //    bd::SystemState state;
  //    //state = getHebbianParameters();
  //    state = getStandardParameters();
  //
  //    ThoughtNetwork network( state );


  return 0;
}
