#include <iostream>
#include <fstream>
#include <cereal/archives/binary.hpp>
#include <cereal/archives/json.hpp>

#include <tbb/task_scheduler_init.h>

#include "src/V1.H"
#include "src/Generate.H"
#include "tests/Util.H"
#include "tests/V1Tests.H"
#include "tests/GaborTests.H"

#include "src/ThreadingUtil.H"
#include "config/Options.H"

//! Loads an image, converts to floating point [0-255], and returns
//! the color image and its grayscale version
std::pair<cv::Mat, cv::Mat> loadImage( std::string const & file )
{
  auto image = cv::imread( file );
  // convert to float, leave in 0 to 255 range
  if( image.type() != CV_32FC3 )
    image.convertTo( image, CV_32F );

  cv::Mat grayscale;

  cv::cvtColor( image, grayscale, CV_BGR2GRAY );

  return {image, grayscale};
}

OptionsStruct Options;

int main( int argc, char ** argv )
{
  // Handle options
  if( argc == 1 )
  {
    std::cerr << "ERROR: specify a configuraton file" << std::endl;
    return 1;
  }
  else
  {
    std::ifstream file( argv[1] );
    cereal::JSONInputArchive ar( file );
    ar( CEREAL_NVP(Options) );
  }

  // initialize pools
  {
    std::cout << "Initializing with " << Options.nthreads << " threads." << std::endl;
    tbb::task_scheduler_init init(Options.nthreads);
    threading::nthreads = Options.nthreads;
  }

  const int width  = 20;
  const int height = 20;

  //auto image = bd::generateRectangle( height, width, 10, 10, 10, 10 );
  cv::Mat blank = cv::Mat::zeros( height, width, CV_32FC1 );

  // square
  auto shape  = bd::rotateShape( bd::generateLine( -1, 10, 19 ), 45.0_deg );
  // rectangle
  //auto shape = bd::rotateShape( bd::generateRectangle( 10, 10, 20, 10 ), 45.0_deg );
  // triangle
  //auto shape = bd::rotateShape( bd::generateEquilateralTriangle( 15, 15, 5 ), 0.0_deg );

  //===========================================================================
  //===========================================================================
  //===========================================================================

  //showConnections( v1.itsEdgeResponses, false, true );
  //bd::test::loadBD( v1, "amd_pc_10000_v1_edge_converge.cereal" );

  //v1.itsState.useHebbianLearning = true;

  auto writeFile = [](cv::Mat const & mat, std::string const & name)
  {
    cv::Mat mat8U = cv::Mat::zeros( mat.rows, mat.cols, CV_8UC3 );
    cv::Mat local = mat.clone();
    local *= 255.0;
    local.convertTo( mat8U, CV_8UC3 );

    std::cout << "Writing image: " << name << std::endl;
    cv::imwrite(name, mat8U);
  };


  auto setV1State = []( bd::V1 & v1, bd::SystemState const & state )
  {
    v1.itsState.v1MaxFFWeight = state.v1MaxFFWeight;
    v1.itsState.v1lateralTotalWeight = state.v1lateralTotalWeight;
    v1.itsState.learningRate = state.learningRate;
    v1.itsState.learningRule = state.learningRule;
  };

  auto getOptimizedOptions = []( bd::SystemState const & state)
  {
    bd::V1::Options options;
    options.useGaussianRing = true;
    options.inhibitionRadius = 1.0;
    options.inhibitionSpread = 0.24333;
    options.inhibitionStrength = 30.0;

    options.excitationRadius = 0.0;
    options.excitationSpread = 0.09;
    options.excitationStrength = 1.0;

    return options;
  };

  auto getDefaultOptions = []( bd::SystemState const & state)
  {
    // default options not tested for conflict learning on gaussian
    const bool useGaussian = true;
    const double excitationRadius = 0.27;
    const double inhibitionRadius = 1.0;

    bd::V1::Options options;
    options.useGaussianRing = useGaussian;

    const double staticRadius = (inhibitionRadius + excitationRadius) / 2.0;
    const double staticSpread = (inhibitionRadius - excitationRadius) / (useGaussian ? 3.0 : 2.0);

    options.inhibitionRadius = useGaussian ? inhibitionRadius : staticRadius;
    options.inhibitionSpread = staticSpread;
    options.inhibitionStrength = useGaussian ? 30.0 : state.v1lateralTotalWeight;

    options.excitationRadius = 0.0;
    options.excitationSpread = useGaussian ? excitationRadius / 3.0 : excitationRadius;
    options.excitationStrength = useGaussian ? 2.0 : 1.0;

    return options;
  };

  auto getHebbianParameters = []()
  {
    // 1.0, 0.5, 0.01 - original set
    // 1.15, 0.5, 0.001 - 'final' set
    bd::SystemState state;
    state.v1MaxFFWeight = 1.15;
    state.v1lateralTotalWeight = 0.5;
    state.learningRate = 0.001;
    state.learningRule = bd::LearningRule::Hebbian;

    return state;
  };

  auto getStandardParameters = []()
  {
    // 1.0,  5.0, 0.01 - original params
    // 1.15, 7.0, 0.01 - 'final' set
    // 2, 5, 0.01 - really close set
    // 2, 5, .05 - closer?
    //
    bd::SystemState state;
    state.v1MaxFFWeight = 1.25;
    state.v1lateralTotalWeight = 7.00; // dynamic
    state.learningRate = 0.01;

    return state;
  };

  const int inputScaling = 15;

  if( argc < 4 )
  {
    std::string fileName = "current_edge.cereal";
    std::string imageName = "image_edge";

    if( argc >= 3 )
    {
      fileName = std::string(argv[2]);
      imageName = std::string(argv[2]);
    }

    bd::SystemState state;
    if( fileName.find( "hebbian" ) != std::string::npos )
      state = getHebbianParameters();
    else
      state = getStandardParameters();

    bd::V1 v1( width, height, state );
    bd::test::loadBD( v1, fileName );

    if( v1.itsState.learningRule == bd::LearningRule::Hebbian )
      std::cout << "Hebbian configuration detected" << std::endl;
    if( v1.itsState.learningRule == bd::LearningRule::Conflict )
      std::cout << "Standard configuration detected" << std::endl;

    //bd::showImage( sheetNeuronImage( v1.itsEdgeResponses, 8.0 ), false, true, "neurons" );
    {
      auto imageAndScore = getV1OrientationPreferencesImageMatlab( v1 );
      writeFile( imageAndScore.first,  imageName + "_matlab.png" );
      std::ofstream scoreFile( imageName + "_score.txt" );
      scoreFile << imageAndScore.second << std::endl;
    }

//    bd::showImage( image, false, true, "colors, yo" );

//    auto image2 = getV1OrientationPreferencesImage( v1, inputScaling );
//    writeFile( image2, imageName + "_default.png" );
//    bd::showImage( image2, false, true, "colors2, yo" );
//
//    showConnections( v1.itsEdgeResponses, false, true );
//
//    while( true )
//    {
//      const int stimulusHeight = height * inputScaling;
//      const int stimulusWidth  = width  * inputScaling;
//
//      const double theta = bd::uniformReal( 0.0, 2 * M_PI )();
//      //const double theta = 0.0;
//      auto shape  = bd::rotateShape( bd::generateLine( -1, stimulusWidth / 2, stimulusWidth - 1 ), theta );
//
//      cv::Mat stimulus = cv::Mat::zeros( stimulusHeight, stimulusWidth, CV_32FC1 );
//
//      cv::Mat stimulus8U = cv::Mat::zeros( stimulusHeight, stimulusWidth, CV_8UC1 );
//      bd::drawShape( stimulus8U, shape, 255.0, false, inputScaling, CV_AA );
//
//      stimulus8U.convertTo( stimulus, CV_32FC1 );
//      stimulus /= 255.0;
//
//      for( int i = 0; i < 10; ++i )
//      {
//        bd::test::presentStimulus( v1, stimulus, 1 );
//        bd::showImage( stimulus, false, false, "input" );
//
//        // normal scaling would be 8.0
//        auto onOffIm = sheetActivationImage( v1.itsOnOffResponses, 2.0 );
//        auto offOnIm = sheetActivationImage( v1.itsOffOnResponses, 2.0 );
//        auto edgeIm  = sheetActivationImage( v1.itsEdgeResponses, 2.0 );
//
//        // Save images
//        {
//          cv::imwrite( "onOff.png", onOffIm * 255 );
//          cv::imwrite( "offOn.png", offOnIm * 255 );
//          cv::imwrite( "edge.png", edgeIm * 255 );
//          cv::imwrite( "input.png", stimulus * 255 );
//        }
//
//        bd::showImage( onOffIm, false, false, "on" );
//        bd::showImage( offOnIm, false, false, "off" );
//        bd::showImage( edgeIm, false, false, "v1" );
//        bd::showImage( sheetActivationImage( v1.itsEdgeResponses ), false, true, "v1_no_scale" );
//      }
//
//      //bd::showImage( image2, false, true, "before" );
//      //image2 = getV1OrientationPreferencesImage( v1, inputScaling );
//      //bd::showImage( image2, false, true, "after" );
//
//    //showConnections( v1.itsEdgeResponses, false, false );
//      bd::test::presentStimulus( v1, cv::Mat::zeros(stimulusHeight,stimulusWidth,CV_32FC1), 500 );
//    }
  }
  else
  {
    // Uncomment this to do bias testing
    //double stdBias[] = {10, 2, 1, .8, .6, .4, .2};
    ////double stdBias[] = {10};

    //for( const auto std : stdBias )
    //{
    //  std::cout << "On bias " << std << std::endl;

    //  {
    //    bd::V1 network( width, height, getHebbianParameters() );
    //    testV1Convergence( network, blank, std, inputScaling, "_v1_hebbian_edge_" + std::to_string(std) + ".cereal" );
    //  }

    //  {
    //    bd::V1 network( width, height, getStandardParameters() );
    //    testV1Convergence( network, blank, std, inputScaling, "_v1_standard_edge_" + std::to_string(std) + ".cereal" );
    //  }
    //}

    // Uncomment this to do noise testing
    double stdNoise[] = {.01, .1, .2, .3, .4, .5, .6, .7};
    //double stdNoise[] = {.01};

    for( const auto std : stdNoise )
    {
      std::cout << "On noise " << std << std::endl;

      {
        auto state = getHebbianParameters();
        //auto options = getDefaultOptions( state );
        auto options = getOptimizedOptions( state );
        bd::V1 network( width, height, state, options );
        network.itsState.staticNoiseScale = std;
        //showConnections( network.itsEdgeResponses, false, true );
        testV1Convergence( network, blank, 10, inputScaling, "_v1_hebbian_edge_noise_" + std::to_string(std) + ".cereal" );
      }

      {
        auto state = getStandardParameters();
        //auto options = getDefaultOptions( state );
        auto options = getOptimizedOptions( state );
        bd::V1 network( width, height, state, options );
        network.itsState.staticNoiseScale = std;

        //network.itsState.cl_spreading = false;
        //network.itsState.cl_unlearning = false;
        //network.itsState.cl_slt = false;
        //network.itsState.learningRate /= 10;

        testV1Convergence( network, blank, 10, inputScaling, "_v1_standard_edge_noise_" + std::to_string(std) + ".cereal" );
      }
    }

//    // below is normal behavior
//    bd::SystemState state;
//    //state = getHebbianParameters();
//    state = getStandardParameters();
//
//    bd::V1 v1( width, height, state );
//    //showConnections( v1.itsEdgeResponses, false, true );
//    //bd::test::loadBD( v1, "current_edge.cereal" );
//    testV1Convergence( v1, blank, 1.0, inputScaling, "_v1_edge.cereal" );
  }

  return 0;
}
