#include <iostream>
#include <fstream>
#include <cereal/archives/binary.hpp>
#include <cereal/archives/json.hpp>

#include <tbb/task_scheduler_init.h>

#include "src/V1.H"
#include "src/Generate.H"
#include "tests/Util.H"
#include "tests/V1Tests.H"
#include "tests/GaborTests.H"

#include "matlab_lib/GaborFit.H"

#include "src/ThreadingUtil.H"
#include "config/Options.H"

OptionsStruct Options;

// Input parameters:
// arg0
// arg1 - configuration JSON file
// arg2 - boolean, true for conflict, false for hebbian
// arg3 - output file name prefix for saved data
// arg4 - options JSON file
int main( int argc, char ** argv )
{
  // Handle options
  if( argc != 5 )
  {
    std::cerr << "Parameters are: configuration JSON, boolean useHebbian, output file prefix, options JSON" << std::endl;
    return 1;
  }

  bool useHebbian = std::stoi( std::string( argv[2] ) );
  std::string outputPrefix( argv[3] );

  bd::V1::Options options;

  {
    std::ifstream file( argv[1] );
    cereal::JSONInputArchive ar( file );
    ar( CEREAL_NVP(Options) );
  }

  {
    std::ifstream file( argv[4] );
    cereal::JSONInputArchive ar( file );
    ar( CEREAL_NVP(options) );
  }

  // initialize pools
  {
    std::cout << "Initializing with " << Options.nthreads << " threads." << std::endl;
    tbb::task_scheduler_init init(Options.nthreads);
    threading::nthreads = Options.nthreads;
  }

  const int width  = 20;
  const int height = 20;

  //auto image = bd::generateRectangle( height, width, 10, 10, 10, 10 );
  cv::Mat blank = cv::Mat::zeros( height, width, CV_32FC1 );

  auto writeFile = [](cv::Mat const & mat, std::string const & name)
  {
    cv::Mat mat8U = cv::Mat::zeros( mat.rows, mat.cols, CV_8UC3 );
    cv::Mat local = mat.clone();
    local *= 255.0;
    local.convertTo( mat8U, CV_8UC3 );

    std::cout << "Writing image: " << name << std::endl;
    cv::imwrite(name, mat8U);
  };

  auto getHebbianParameters = []()
  {
    // 1.0, 0.5, 0.01 - original set
    // 1.15, 0.5, 0.001 - 'final' set
    bd::SystemState state;
    state.v1MaxFFWeight = 1.15;
    state.v1lateralTotalWeight = 0.5;
    state.learningRate = 0.001;
    state.learningRule = bd::LearningRule::Hebbian;

    return state;
  };

  auto getStandardParameters = []()
  {
    // 1.0,  5.0, 0.01 - original params
    // 1.15, 7.0, 0.01 - 'final' set
    // 2, 5, 0.01 - really close set
    // 2, 5, .05 - closer?
    //
    bd::SystemState state;
    state.v1MaxFFWeight = 1.25;
    state.v1lateralTotalWeight = 7.00; // dynamic
    state.learningRate = 0.01;

    return state;
  };

  const int inputScaling = 15;

  auto state = useHebbian ? getHebbianParameters() : getStandardParameters();
  bd::V1 network( width, height, state, options );
  network.itsState.staticNoiseScale = 0.01;
  //showConnections( network.itsEdgeResponses, false, true );

  testV1Convergence( network, blank, 10, inputScaling, outputPrefix + ".cereal" );

  {
    auto imageAndScore = getV1OrientationPreferencesImageMatlab( network );
    const std::string outputImageName = outputPrefix + "_matlab.png";
    writeFile( imageAndScore.first,  outputImageName );
    std::ofstream scoreFile( outputPrefix + "_score.txt" );
    scoreFile << imageAndScore.second << std::endl;

    matlab::generatePinwheelsInputImage( outputImageName, outputPrefix + "_mat.mat" );
  }

  return 0;
}
