function [ biasHebbianData, biasStandardData, noiseHebbianData, noiseStandardData ] = createV1ErrorPlots( directory )
%createV1ErrorPlots Creates error plots (bias, noise) for V1 scores files
% Input directory should contain output .mat files corresponding to the
% last timestep of several iterations of V1 learning. The mat files should
% be generated via the C++ evaluateScores function (GaborFit.C), which
% calls the Matlab functions: colorize and reformatScoresStruct.

bias = struct;
noise = struct;
bias.paramValues = [];
noise.paramValues = [];

files = dir(strcat(directory, '*.mat'));
files = {files(~[files.isdir]).name};

for f = 1:numel(files)    
    file = char(files(f));
    load(char(fullfile(directory, file))); % will load into variable 'scores'
    
    isNoise = ~isempty(strfind(file, 'noise'));
    
    for i = 1:numel(scores)
        conflictName = strcat(scores(i).type(1), num2str(scores(i).bias));
        conflictName = strrep(conflictName, '-', 'N');
        conflictName = strrep(conflictName, '.', 'O');
        
        % comment to use natural distribution
        if ~isempty(strfind(conflictName, 'N'))
            continue;
        end
        
        if isNoise
            if ~isfield(noise, conflictName)
                noise.(conflictName) = [];
                if ~any(noise.paramValues == scores(i).bias)
                    noise.paramValues = [noise.paramValues scores(i).bias];
                end
            end
            
            noise.(conflictName) = [noise.(conflictName) scores(i).data.learningScore];
        else         
            if ~isfield(bias, conflictName)
                bias.(conflictName) = [];
                if ~any(bias.paramValues == scores(i).bias)
                    bias.paramValues = [bias.paramValues scores(i).bias];
                end
            end
            
            bias.(conflictName) = [bias.(conflictName) scores(i).data.scores];
        end        
    end
end
bias.paramValues = sort(bias.paramValues);
noise.paramValues = sort(noise.paramValues);

%% Bias (Skip if no bias data)
if ~isempty(bias.paramValues)
    fields = fieldnames(bias);
    minBiasHebbianSamples = inf;
    minBiasStandardSamples = inf;
    for i = 1:numel(fields)
        if fields{i}(1) == 'h'
            minBiasHebbianSamples = min(numel(bias.(fields{i})), minBiasHebbianSamples);
        elseif fields{i}(1) == 's'
            minBiasStandardSamples = min(numel(bias.(fields{i})), minBiasStandardSamples);
        end
    end
    biasHebbianData = zeros(minBiasHebbianSamples, numel(bias.paramValues));
    biasStandardData = zeros(minBiasStandardSamples, numel(bias.paramValues));
    for i = 1:numel(fields)
        paramValue = fields{i}(2:end);
        paramValue = strrep(paramValue, 'N', '-');
        paramValue = strrep(paramValue, 'O', '.');
        paramValue = str2double(paramValue);
        row = bias.(fields{i});
        if fields{i}(1) == 'h'
            row = row(1:minBiasHebbianSamples);
            biasHebbianData(:, find(bias.paramValues == paramValue)) = row;
        elseif fields{i}(1) == 's'
            row = row(1:minBiasStandardSamples);
            biasStandardData(:, find(bias.paramValues == paramValue)) = row;
        end
    end

    % go from most to least amount of bias STD
    biasHebbianData = fliplr(biasHebbianData);
    biasStandardData = fliplr(biasStandardData);
end

%% Noise (Skip if no noise data)
if ~isempty(noise.paramValues)
    fields = fieldnames(noise);
    minNoiseHebbianSamples = inf;
    minNoiseStandardSamples = inf;
    for i = 1:numel(fields)
        if fields{i}(1) == 'h'
            minNoiseHebbianSamples = min(numel(noise.(fields{i})), minNoiseHebbianSamples);
        elseif fields{i}(1) == 's'
            minNoiseStandardSamples = min(numel(noise.(fields{i})), minNoiseStandardSamples);
        end
    end
    noiseHebbianData = zeros(minNoiseHebbianSamples, numel(noise.paramValues));
    noiseStandardData = zeros(minNoiseStandardSamples, numel(noise.paramValues));
    for i = 1:numel(fields)
        paramValue = fields{i}(2:end);
        paramValue = strrep(paramValue, 'N', '-');
        paramValue = strrep(paramValue, 'O', '.');
        paramValue = str2double(paramValue);
        row = noise.(fields{i});
        if fields{i}(1) == 'h'
            row = row(1:minNoiseHebbianSamples);
            noiseHebbianData(:, find(noise.paramValues == paramValue)) = row;
        elseif fields{i}(1) == 's'
            row = row(1:minNoiseStandardSamples);
            noiseStandardData(:, find(noise.paramValues == paramValue)) = row;
        end
    end
end

%% Plotting
color_map = [0, 0, 1; 1, 0, 0];

if ~isempty(noise.paramValues)
    figure; plotStatisticalError(noiseHebbianData, noiseStandardData); colormap(color_map); title('noise');
end

if ~isempty(bias.paramValues)
    figure; plotStatisticalError(biasHebbianData, biasStandardData); colormap(color_map); title('bias');
end