#include <iostream>
#include <tuple>
#include <opencv2/imgproc.hpp>
#include <opencv2/highgui.hpp>

bool angleInBounds( double angle, double target, double delta )
{
  auto normalize = [](double x)
  {
    x = std::fmod(x, 360.0);
    return x >= 180.0 ? x - 360.0 : x < -180.0 ? x + 360.0 : x;
  };

  const double diff = normalize(angle - target);
  return diff < delta && diff > -delta;
}

std::tuple<double, double, double, double, double, double>
activePolarityScore( cv::Mat const & groundTruth, cv::Mat const & allDirections, const int resolutionScaling )
{
  cv::Mat gt_HSV = groundTruth;
  cv::Mat ad_HSV = allDirections;

  cv::cvtColor( gt_HSV, gt_HSV, CV_BGR2HSV );
  cv::cvtColor( ad_HSV, ad_HSV, CV_BGR2HSV );

  std::vector<double> scores;
  double minScore = 0.0, maxScore = 0.0;
  double totalScore = 0.0;
  int count = 0;
  int countWithin180 = 0; // used to calculate number of answers that are facing the same direction
                          // (plus or minus 90 degrees) of the ground truth

  for( int row = 0; row < allDirections.rows; ++row )
  {
    const int y = row * resolutionScaling;

    for( int col = 0; col < allDirections.cols; ++col )
    {
      const int x = col * resolutionScaling;

      // Ignore values where there is no ground truth available
      auto gtPixel = gt_HSV.at<cv::Vec3f>( y, x );
      if( gtPixel[2] == 0 )
        continue;

      // Also ignore values where the ground truth is solid white
      if( gtPixel[1] == 0 && gtPixel[2] == 255 )
        continue;

      const double gtDegrees = gtPixel[0];
      const double adDegrees = ad_HSV.at<cv::Vec3f>( row, col )[0];
      const double score = std::abs( gtDegrees - adDegrees );

      if( angleInBounds( adDegrees, gtDegrees, 90.0 ) )
        ++countWithin180;

      scores.push_back(score);
      minScore = std::min( minScore, score );
      maxScore = std::max( maxScore, score );
      totalScore += score;
      ++count;
    }
  }

  const double averageScore = totalScore / count;
  double stddev = 0.0;

  // compute standard deviation
  for( int row = 0; row < allDirections.rows; ++row )
  {
    const int y = row * resolutionScaling;

    for( int col = 0; col < allDirections.cols; ++col )
    {
      const int x = col * resolutionScaling;

      // Ignore values where there is no ground truth available
      auto gtPixel = gt_HSV.at<cv::Vec3f>( y, x );
      if( gtPixel[2] == 0 )
        continue;

      // Also ignore values where the ground truth is solid white
      if( gtPixel[1] == 0 && gtPixel[2] == 255 )
        continue;

      const double score = std::abs( gtPixel[0] - ad_HSV.at<cv::Vec3f>( row, col )[0] );

      stddev += std::pow( score - averageScore, 2.0 );
    }
  }

  stddev = std::sqrt( stddev / static_cast<double>(count) );

  // Compute median
  std::nth_element( scores.begin(), scores.begin() + scores.size()/2, scores.end() );
  const double median = scores[scores.size()/2];

  const double score180 = countWithin180 / static_cast<double>(count);

  std::cout << "180Score " << score180 << ", AvgScore " << averageScore << ", min/max: " << minScore << "/" << maxScore << ", std: " << stddev << ", med: " << median << std::endl;

  return std::make_tuple(score180, averageScore, minScore, maxScore, stddev, median);
}

cv::Mat autocrop( cv::Mat const & imageColor, cv::Mat const & imageGray, const int width, const int height )
{
  int minX = 9999;
  int maxX = -1;
  int minY = 9999;
  int maxY = 0;

  // Get bounding rectangle
  for( int y = 0; y < imageGray.rows; ++y )
    for( int x = 0; x < imageGray.cols; ++x )
      if( imageGray.at<float>( y, x ) > 0 )
      {
        minX = std::min( minX, x );
        maxX = std::max( maxX, x );
        minY = std::min( minY, y );
        maxY = std::max( maxY, y );
      }

  const int cWidth = maxX - minX + 1;
  const int cHeight = maxY - minY + 1;

  std::cout << "Cropping to " << cWidth << " " << cHeight << std::endl;

  cv::Rect boundingBox( minX, minY, cWidth, cHeight );

  // Crop dat
  cv::Mat cropped = imageColor( boundingBox ).clone();

  cv::Mat stimulus;
  int padT = (height - cHeight) / 2;
  int padB = (height - cHeight) / 2;
  int padL = (width - cWidth) / 2;
  int padR = (width - cWidth) / 2;

  if( padT + padB + cHeight < height )
    ++padT;

  if( padL + padR + cWidth < width )
    ++padL;

  cv::copyMakeBorder( cropped, stimulus, padT, padB, padL, padR, cv::BORDER_CONSTANT, 0.0f );

  std::cout << "Stimulus size " << stimulus.cols << " " << stimulus.rows << std::endl;

  return stimulus;
}

//! Loads an image, converts to floating point [0-255], and returns
//! the color image and its grayscale version
std::pair<cv::Mat, cv::Mat> loadImage( std::string const & file )
{
  auto image = cv::imread( file );
  // convert to float, leave in 0 to 255 range
  if( image.type() != CV_32FC3 )
    image.convertTo( image, CV_32F );

  cv::Mat grayscale;

  cv::cvtColor( image, grayscale, CV_BGR2GRAY );

  return {image, grayscale};
}

int main( int argc, char ** argv )
{
  if( argc != 3 )
  {
    std::cerr << "First arg: ground truth image, Second arg: all directions border ownership image" << std::endl;
    return 1;
  }

  auto groundTruth = loadImage(std::string(argv[1]));
  auto boImage = loadImage(std::string(argv[2])).first;

  auto stimulus_groundtruth = autocrop( groundTruth.first, groundTruth.second, 800.0f, 800.0f );

  activePolarityScore( stimulus_groundtruth, boImage, 10.0f );

  return 0;
}

