#include "Border.H"

#include "Wiring.H"
#include "Util.H"

#include <fstream>
#include <thread>
#include <cereal/archives/binary.hpp>
#include <cereal/types/vector.hpp>

#include "ThreadingUtil.H"

bd::Border::Border( size_t numOrientations,
                    size_t numGNCompetitors,
                    size_t numProtoCompetitors,
                    CompetitionStrategy strategy,
                    double groupingBaseScaling,
                    size_t gridWidth,
                    size_t gridHeight,
                    float  resolutionScaling,
                    SystemState state ) :
  itsState(state),
  itsV1NumOrientations( numOrientations ),
  itsGridWidth( gridWidth ),
  itsGridHeight( gridHeight ),
  itsResolutionScaling( resolutionScaling ),
  itsGabor( 1, 1, 1, numOrientations, 0.6 * M_PI, std::sqrt( 0.5f ), resolutionScaling ),
  itsV1Images( numOrientations ),
  itsV1Responses( numOrientations ),
  itsBOResponses( numOrientations ),
  itsGNResponses(),
  itsProtoResponses()
{
  // ==========================================================================
  // Initialization
  // ==========================================================================
  for( auto & o : itsV1Images ) o.resize(1);

  // ==========================================================================
  // Parameters
  // ==========================================================================
  static const double BASE_SPREAD = 5 * std::sqrt(2.0); // 7.07, smallest scale spread (diagonal of square size 10)

  const double radius = 0;
  const double probability = 1.0;

  // Initial weights sums - typically the max weight sums will be set higher
  const double feedforwardNormalizer = 1.0 * groupingBaseScaling / static_cast<double>(itsV1NumOrientations);
  const double feedbackNormalizer = 1.0 * groupingBaseScaling;
  const double inhibitoryFeedbackNormalizer = -feedbackNormalizer / 4.0;
  const double lateralNormalizer = 0.25;

  const double groupingSpread = BASE_SPREAD * groupingBaseScaling;
  const double protoSpread = 2.0 * groupingSpread;

  // RF Sizes
  // conv = A + B - 1
  // All before scaling:
  // Gabor -> 9 r=4.5
  // BO    -> 9 r=4.5                                 = 9d
  // GRP   -> 14 r=7    @Input: 7 + BO_r = 11.5r      = 23d (2.5x larger), 2.5x larger than V1
  // PROTO -> 14 r=7    @Input: 7 + GRP_input = 18.5r = 37d (1.6x larger), 4.1x larger than V1
  //
  //
  // GRP   -> 14d r=7   @V1_Input = 7 + BO_r(4.5) = 11.5r    @BO: 14d         @V1: 23d (2.5x)
  // PROTO -> 14d r=7   @Input = 7 + GRP_r(3.5) = 10.5       @BO: 21d (1.5x)
  // should be 10.5 r::
  // PROTO -> 21d r=10.5 @Input = 10.5 + GRP_r(3.5) = 14     @BO: 28d (2x)
  //
  // For 1.6 scaling at all levels
  // GRP -> 6 r=3       @Input: 3 + BO_r = 7.5r = 15d (1.66 larger than V1)
  // PROTO -> 9 r=4.5   @Input = 4.5 + GRP_Input(7.5) = 12r = 24d (2.66 larger than V1), 1.6x larger than prev
  //
  // V1: ~0.25 deg   9 r4.5
  // V2 ~1 deg       36 r18 ->  13.5    -> 28 V1 neurons?
  // V3 ~2 deg       72
  // V4 ~2.5 deg     90

  // ==========================================================================
  // Create Grouping Layer
  // ==========================================================================
  const double groupingScalar2 = groupingBaseScaling * groupingBaseScaling;
  // applied to both gridHeight and gridWidth
  // when factoring number of neurons, so we'll square
  // it here

  auto makeGroupingLayer = [=](){ return Sheet<GroupingNeuron>( gridWidth, gridHeight,
                                                                0.25 * gridHeight * gridWidth / groupingScalar2,
                                                                PopulationStrategy::PoissonDisk ); };

  auto baseGroupingLayer = makeGroupingLayer();
  itsGNResponses = baseGroupingLayer;

  for( size_t gL = 1; gL < numGNCompetitors; ++gL )
  {
    switch( strategy )
    {
      case CompetitionStrategy::Uniform:
        itsGNResponses = mergeSheets( itsGNResponses, baseGroupingLayer );
        break;
      case CompetitionStrategy::Population:
        itsGNResponses = mergeSheets( itsGNResponses, makeGroupingLayer() );
        break;
    }
  }

  // Trim boundary (optional)
  //itsGNResponses.trimBoundary( groupingSpread );

  // ==========================================================================
  // Create Proto Layer
  // ==========================================================================
  const double protoScalar2 = groupingBaseScaling * groupingBaseScaling;
  auto makeProtoLayer = [=](){ return Sheet<ProtoObjectNeuron>( gridWidth, gridHeight,
                                                                0.1 * gridHeight * gridWidth / groupingScalar2,
                                                                PopulationStrategy::PoissonDisk ); };
  auto baseProtoLayer = makeProtoLayer();
  itsProtoResponses = baseProtoLayer;

  for( size_t pL = 1; pL < numProtoCompetitors; ++pL )
  {
    switch( strategy )
    {
      case CompetitionStrategy::Uniform:
        itsProtoResponses = mergeSheets( itsProtoResponses, baseProtoLayer );
        break;
      case CompetitionStrategy::Population:
        itsProtoResponses = mergeSheets( itsProtoResponses, makeProtoLayer() );
        break;
    }
  }

  // Wire Proto to Grouping Layer
  if( numProtoCompetitors )
  {
    const double groupingToProtoSpread = groupingSpread;// * 1.6;

    // (proto to grouping) FEEDBACK
    wireSheetRingUniform( itsGNResponses, itsProtoResponses, radius, groupingToProtoSpread, probability, feedbackNormalizer, true,
        ConnectionFlags::Dynamic | ConnectionFlags::Direct | ConnectionFlags::Feedback | ConnectionFlags::Excitatory );

    // (proto to grouping) FEEDBACK INHIBITORY
    wireSheetRingUniform( itsGNResponses, itsProtoResponses, radius, groupingToProtoSpread, probability, inhibitoryFeedbackNormalizer, true,
        ConnectionFlags::Dynamic | ConnectionFlags::Direct | ConnectionFlags::Feedback | ConnectionFlags::Inhibitory );

    // create reciprocal connection (grouping to proto) FEEDFORWARD
    wireSheetRingUniform( itsProtoResponses, itsGNResponses, radius, groupingToProtoSpread, probability, 1.0, true,
        ConnectionFlags::Dynamic | ConnectionFlags::Direct | ConnectionFlags::Feedforward );
  } // end grouping neuron wiring

  // ==========================================================================
  // Create BO Layer
  // ==========================================================================
  for( size_t o = 0; o < itsV1NumOrientations; ++o )
  {
    // Create static V1 sheets
    itsV1Responses[o] = Sheet<StaticNeuron>( itsGridWidth, itsGridHeight );

    // Create border ownership sheets
    itsBOResponses[o] = Sheet<BorderOwnershipNeuron>( itsGridWidth, itsGridHeight );
    itsBOResponses[o].for_each_neuron( [=]( std::shared_ptr<BorderOwnershipNeuron> const & n ){ n->orientation = o * ( M_PI / static_cast<double>( itsV1NumOrientations ) ); } );

    // Wire V1 to BO
    wireToV1Dense( itsBOResponses[o], itsV1Responses[o], 1 );

    // Wire BO to Grouping Neurons
    {
      // (grouping to border) FEEDBACK EXCITATORY
      wireSheetRingUniform( itsBOResponses[o], itsGNResponses, radius, groupingSpread, probability, feedbackNormalizer, true,
          make_connection_flag( ConnectionFlags::Dynamic, ConnectionFlags::Direct, ConnectionFlags::Feedback, ConnectionFlags::Excitatory ) );

      // (grouping to border) FEEDBACK INHIBITORY
      wireSheetRingUniform( itsBOResponses[o], itsGNResponses, radius, groupingSpread, probability, inhibitoryFeedbackNormalizer, true,
          make_connection_flag( ConnectionFlags::Dynamic, ConnectionFlags::Direct, ConnectionFlags::Feedback, ConnectionFlags::Inhibitory ) );

      // create reciprocal connection (border to grouping) FEEDFORWARD
      wireSheetRingUniform( itsGNResponses, itsBOResponses[o], radius, groupingSpread, probability, feedforwardNormalizer, true,
          make_connection_flag( ConnectionFlags::Dynamic, ConnectionFlags::Direct, ConnectionFlags::Feedforward, ConnectionFlags::Excitatory ) );
    } // end grouping neuron wiring

    // Wire BO to Proto Neurons
    if( numProtoCompetitors )
    {
      wireSheetRingUniform( itsProtoResponses, itsBOResponses[o], radius, protoSpread, probability, feedforwardNormalizer, true,
          ConnectionFlags::Dynamic | ConnectionFlags::Skip | ConnectionFlags::Feedforward | ConnectionFlags::Excitatory );

      // skip FB inhibitory
      wireSheetRingUniform( itsBOResponses[o], itsProtoResponses, radius, protoSpread, probability, inhibitoryFeedbackNormalizer, true,
        make_connection_flag( ConnectionFlags::Dynamic, ConnectionFlags::Skip, ConnectionFlags::Feedback, ConnectionFlags::Inhibitory ) );

      // skip FB
      wireSheetRingUniform( itsBOResponses[o], itsProtoResponses, radius, protoSpread, probability, feedbackNormalizer, true,
        make_connection_flag( ConnectionFlags::Dynamic, ConnectionFlags::Skip, ConnectionFlags::Feedback, ConnectionFlags::Excitatory ) );
    } // end proto neuron wiring
  }

  // ==========================================================================
  // Create Lateral Connections
  // ==========================================================================

  // Border Ownership layer
  {
    for( size_t o = 0; o < itsV1NumOrientations; ++o )
      for( size_t oo = 0; oo < itsV1NumOrientations; ++oo )
      {
        const double lateralSpread = groupingSpread / 2.0;

        // Column competition
        const double columnWeight = (-itsState.boMaxWeights.column / itsV1NumOrientations) * (o == oo ? 0.5 : 1.0); // half weight when only one connection
        wireSheetRingUniform( itsBOResponses[o], itsBOResponses[oo], 0, .01, probability, columnWeight, true,
                              ConnectionFlags::Dynamic | ConnectionFlags::Column | ConnectionFlags::Inhibitory );

        // Lateral competition
//        wireSheetRingUniform( itsBOResponses[o], itsBOResponses[oo], 0, lateralSpread, probability, -itsState.boLateralTotalWeight / itsV1NumOrientations, true,
//                              ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory,
//                              ConnectionFlags::Column /* (exclusions) */ );
//
//        wireSheetRingUniform( itsBOResponses[o], itsBOResponses[oo], 0, lateralSpread, probability, lateralNormalizer, true,
//                              ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Excitatory,
//                              ConnectionFlags::Column /* exclusions */ );
      }
  }

  // Grouping layer
  {
    const double lateralSpread = groupingSpread;// / 2.0;

    // Column competition
    wireSheetRingUniform( itsGNResponses, itsGNResponses, 0, lateralSpread / 8.0, probability, -itsState.gnMaxWeights.column, true,
                          ConnectionFlags::Dynamic | ConnectionFlags::Column | ConnectionFlags::Inhibitory );

    // Lateral competition
    wireSheetRingUniform( itsGNResponses, itsGNResponses, 0, lateralSpread, probability, -itsState.gnMaxWeights.lateral, true,
                          ConnectionFlags::Static | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory,
                          ConnectionFlags::Column /* (exclusions) */ );

    //wireSheetRingUniform( itsGNResponses, itsGNResponses, 0, lateralSpread, probability, lateralNormalizer, true,
    //                      ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Excitatory | ConnectionFlags::Silent ); // silent debug
  }

  // Proto Object layer
  if( numProtoCompetitors )
  {
    const double lateralSpread = groupingSpread;

    // Column competition
    wireSheetRingUniform( itsProtoResponses, itsProtoResponses, 0, lateralSpread / 8.0, probability, -itsState.poMaxWeights.column, true,
                          ConnectionFlags::Dynamic | ConnectionFlags::Column | ConnectionFlags::Inhibitory );
    //wireSheetRingUniform( itsProtoResponses, itsProtoResponses, 0, .01, probability, -itsState.gnColumnTotalWeight, true,
    //                      ConnectionFlags::Dynamic | ConnectionFlags::Column | ConnectionFlags::Inhibitory );

    // Lateral competition
    wireSheetRingUniform( itsProtoResponses, itsProtoResponses, 0, lateralSpread, probability, -itsState.poMaxWeights.lateral, true,
                          ConnectionFlags::Static | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory,
                          ConnectionFlags::Column /* (exclusions) */ );

    //wireSheetRingUniform( itsProtoResponses, itsProtoResponses, 0, lateralSpread, probability, lateralNormalizer, true,
    //                      ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Excitatory );
  }

  // ==========================================================================
  // Normalize Connections
  // ==========================================================================
  calculateBorderOwnershipFBCoefficients( itsBOResponses, itsState );
  calculateCoefficients( itsGNResponses );
  calculateCoefficients( itsProtoResponses );

  // ==========================================================================
  // Finalize Initialization
  // ==========================================================================
  for( size_t o = 0; o < itsV1NumOrientations; ++o )
  {
    threading::parallel_for( 0lu, itsV1Responses[o].size(), [&](size_t i)
    {
      auto & vec = itsV1Responses[o][i];
      for( auto & neuron : vec )
      {
        neuron->itsInput.post() = neuron->itsInput.pre();
        neuron->rebuildCache();
      }
    } );

    threading::parallel_for( 0lu, itsBOResponses[o].size(), [&](size_t i)
    {
      auto & vec = itsBOResponses[o][i];
      for( auto & neuron : vec )
      {
        neuron->itsInput.post() = neuron->itsInput.pre();
        neuron->rebuildCache();
      }
    } );
  }

  threading::parallel_for( 0lu, itsGNResponses.size(), [&](size_t i)
  {
    auto & vec = itsGNResponses[i];
    for( auto & neuron : vec )
    {
      neuron->itsInput.post() = neuron->itsInput.pre();
      neuron->rebuildCache();
    }
  } );

  threading::parallel_for( 0lu, itsProtoResponses.size(), [&](size_t i)
  {
    auto & vec = itsProtoResponses[i];
    for( auto & neuron : vec )
    {
      neuron->itsInput.post() = neuron->itsInput.pre();
      neuron->rebuildCache();
    }
  } );
}

void bd::Border::onset( cv::Mat const & image )
{
  // Downsample input
//  cv::Mat downsampled = image;
//  resizeInPlace( downsampled, cv::Size( itsGridWidth, itsGridHeight ) );

  // Perform gabor filtering
  //  result is orientation X scale array of results
  itsV1Images = itsGabor.getFilterResponses( image );

  //bd::showImage( image, true, true, "input" );
  //bd::showImage( itsV1Images[0][0], true, true, "banana" );

  // Initialize static sheets for v1
  for( size_t o = 0; o < itsV1NumOrientations; ++o )
  {
    cv::Mat const & v1Image = itsV1Images[o][0];

    for( auto & vec : itsV1Responses[o] )
      for( auto & neuron : vec )
      {
        int x, y;
        std::tie(x, y) = neuron->position();
        x *= itsResolutionScaling;
        y *= itsResolutionScaling;

        neuron->setActivation( v1Image.at<float>( y, x ) );
      }
  }
}

void bd::Border::iterate( int numIters )
{
  auto iterateImpl = [&]()
  {
    // UPDATE ===================================================================
    for( size_t o = 0; o < itsV1NumOrientations; ++o )
    {
      //std::cout << "===== Updating BO " << o << ", " << s << " ------------------------------" << std::endl;

      threading::parallel_for( 0lu, itsBOResponses[o].size(), [&](size_t i)
      {
        auto & vec = itsBOResponses[o][i];

        // DEBUGGING PRINT DOES NOT WORK IF MULTI THREADED
        if( o == 0 )
          itsState.debugPrint = true;
        else
          itsState.debugPrint = false;

        for( auto & neuron : vec )
          neuron->update( itsState );
      } );

      //std::cout << std::endl;
    }

    threading::parallel_for( 0lu, itsGNResponses.size(), [&](size_t i)
    {
      auto & vec = itsGNResponses[i];
      for( auto & neuron : vec )
        neuron->update( itsState );
    } );

    threading::parallel_for( 0lu, itsProtoResponses.size(), [&](size_t i)
    {
      auto & vec = itsProtoResponses[i];
      for( auto & neuron : vec )
        neuron->update( itsState );
    } );

    // LEARN ====================================================================
    for( size_t o = 0; o < itsV1NumOrientations; ++o )
    {
      threading::parallel_for( 0lu, itsBOResponses[o].size(), [&](size_t i)
      {
        auto & vec = itsBOResponses[o][i];
        for( auto & neuron : vec )
          neuron->learn( itsState );
      } );
    }

    threading::parallel_for( 0lu, itsGNResponses.size(), [&](size_t i)
    {
      auto & vec = itsGNResponses[i];
      for( auto & neuron : vec )
        neuron->learn( itsState );
    } );

    threading::parallel_for( 0lu, itsProtoResponses.size(), [&](size_t i)
    {
      auto & vec = itsProtoResponses[i];
      for( auto & neuron : vec )
        neuron->learn( itsState );
    } );

    // FINALIZE =================================================================
    finalizeNeuronVariables();
  }; // end iterateImpl

  for( int i = 0; i < numIters; ++i )
    iterateImpl();
}
