#ifndef BORDER_H_
#define BORDER_H_

#include <opencv2/core/core.hpp>

#include "Gabor.H"
#include "Sheet.H"
#include "State.H"

#include "StaticNeuron.H"
#include "BorderOwnershipNeuron.H"
#include "GroupingNeuron.H"
#include "ProtoObjectNeuron.H"

namespace bd
{
  //! Overall border ownership class
  class Border
  {
    public:
      /*! Construct a Border environment
          @param numOrientations Number of edge orientations, used to populate V1 Gabor responses
          @param numGNCompetitors The maximum number of grouping neurons in competition in a column
          @param numProtoCompetitors The maximum number of proto neurons in competition in a column
          @param strategy The column competition strategy for grouping and beyond
          @param groupingBaseScaling A multiplier on the size of the first scale (normally 1.6x larger each scale)
          @param gridWidth Width of simulation (think input pixels)
          @param gridHeight Height of simulation
          @param resolutionScaling Scaling factor for interpolation, used for V1 computation and drawing
          @param state An initial system state, containing various weight settings */
      Border( size_t numOrientations = 4,
              size_t numGNCompetitors = 1,
              size_t numProtoCompetitors = 1,
              CompetitionStrategy strategy = CompetitionStrategy::Uniform,
              double groupingBaseScaling = 1.0,
              size_t gridWidth  = 40,
              size_t gridHeight = 40,
              float  resolutionScaling = 10.0f,
              SystemState state = SystemState() );

      //! Present a new image for the first time (recomputes v1)
      void onset( cv::Mat const & image );

      //! Advances the network
      void iterate( int numIters = 1 );

      template <class Archive>
      void serialize( Archive & ar )
      {
        ar( CEREAL_NVP(itsState),
            CEREAL_NVP(itsV1NumOrientations),
            CEREAL_NVP(itsGridWidth),
            CEREAL_NVP(itsGridHeight),
            CEREAL_NVP(itsV1Responses),
            CEREAL_NVP(itsBOResponses),
            CEREAL_NVP(itsGNResponses),
            CEREAL_NVP(itsProtoResponses) );
        ar.serializeDeferments(); // connections between neurons deferred
      }

    public:
    //private:
      SystemState itsState;

      size_t itsV1NumOrientations;
      size_t itsGridWidth;
      size_t itsGridHeight;
      float  itsResolutionScaling;

      Gabor itsGabor; //!< Gabor filtering mechanism

      //!< Images of Gabor output, organized as [orientation][scale] (scale fixed to 1)
      std::vector<std::vector<cv::Mat>> itsV1Images;

      //! Responses from V1, organized as [orientation]
      std::vector<Sheet<StaticNeuron>>          itsV1Responses;
      //! Responses from BO, organized as [orientation]
      std::vector<Sheet<BorderOwnershipNeuron>> itsBOResponses;
      //! Responses from grouping neurons
      Sheet<GroupingNeuron>                     itsGNResponses;
      //! Responses from proto object
      Sheet<ProtoObjectNeuron>                  itsProtoResponses;
  };
}

#endif // BORDER_H_
