#include "BorderOwnershipNeuron.H"
#include "Random.H"

#include <cereal/types/polymorphic.hpp>

CEREAL_REGISTER_TYPE(bd::BorderOwnershipNeuron)

double bd::BorderOwnershipNeuron::updateImpl( SystemState const & state )
{
  const double noise = normalReal()() * state.noiseScale;

  double feedforward = 0.0;
  double feedback_e = 0, feedback_i = 0; // Split into excitatory and inhibitory
  double ltFeedforward = 0;
  double lateralExcitation = 0.0, lateralInhibition = 0.0;

  const double skipCoefficient = 0.25;

  auto const & preInput = itsInput.pre();

  // Excitation
  for( auto const & preEdge : preInput )
  {
    // skip all silent connections
    if( preEdge.hasFlag( ConnectionFlags::Silent ) )
      continue;

    const double incomingActivation = preEdge.neuron->activation();
    double currentInput   = preEdge.weight * incomingActivation;
    double ltCurrentInput = preEdge.ltWeight * incomingActivation;

    if( preEdge.hasFlag( ConnectionFlags::Skip ) )
      currentInput *= skipCoefficient;

    if( preEdge.hasFlag( ConnectionFlags::Feedforward ) )
    {
      feedforward += currentInput;
      ltFeedforward += ltCurrentInput;
    }
    else if( preEdge.hasFlag( ConnectionFlags::Feedback ) )
    {
      if( preEdge.hasFlag( ConnectionFlags::Direct ) && !state.dy_grouping_to_bo_feedback )
        continue;
      if( preEdge.hasFlag( ConnectionFlags::Skip ) && !state.dy_skip_feedback )
        continue;

      if( currentInput >= 0 )
        feedback_e += currentInput;
      else
        feedback_i += currentInput;
    }
    else if( preEdge.hasFlag( ConnectionFlags::Lateral | ConnectionFlags::Excitatory ) )
      lateralExcitation += currentInput;
  }

  // Ambiguity and feedback
  const double feedback = std::max( feedback_e + feedback_i, 0.0 );
  const double ambiguity = state.cl_ambiguity ? std::min( feedback_e, -feedback_i ) : 0.0;

  // Calculates activation value before threshold and min/max clamping
  auto calculatePreClamp = [=]( double ff, double lateral, double fb, double inhib, double noise )
  {
    const double clampedFF = clamp(ff);
    const double ffGating = clampedFF * clampedFF;

    const double input = ff + lateral * ffGating + fb * ffGating + noise;

    double result = input / ( 1.0 + inhib );

    if( result <= 0 )
      return 0.0;

    result = std::min( result, 1.0 ) + std::max( 0.0, std::log10( result ) );
    return result / itsGainMultiplier;
  };

  // Inhibition
  itsInhibitorySum = calculateInhibitorySum( state, itsInhibitorySum, ConnectionFlags::Column );
  itsLateralInhibitorySum = calculateInhibitorySum( state, itsLateralInhibitorySum, ConnectionFlags::Lateral );

  itsModulatoryInput.update( clamp(feedforward) * clamp(feedforward) * feedback );

  updateGainControls( ConnectionFlags::Column );

  // Activation
  itsUninhibitedActivation.update( calculatePreClamp( feedforward, lateralExcitation, feedback, ambiguity, noise ) );

  double preClamp = calculatePreClamp( feedforward, lateralExcitation, feedback, itsInhibitorySum + itsLateralInhibitorySum + ambiguity, noise );

  updateThresholds( state, ltFeedforward, feedforward,
                    itsInhibitorySum,
                    0.4 * itsNormalizationCoefficients.column, // inhibitory threshold
                    0.001,       // fast
                    0.0005,      // slow
                    0.00001 );   // decay

  // New idea with thresholds:
  //  Thresholds only apply to neurons that were previously in an 'off' state - they can't turn off neurons
  //  that are already on, even if those neurons become inhibited somehow. Rather they use long term weighted activation
  //  to control the initial turn-on, and allow adjustment to this turn on threshold using short term weight.
  //
  //  This scheme allows a neuron to have a good dynamic range of activation once on but still segregates neurons based
  //  on thresholding input.

  const double inverse = 1.0 / itsNormalizationCoefficients.feedforward;
  if( feedforward < itsFastThreshold * inverse )
 // if( feedforward < state.noiseScale * 4 )
  {
    preClamp = 0.0;
    itsUninhibitedActivation.update( 0.0 );
  }

  const double value = clamp( preClamp, 0.0, state.maxActivation );

  // Only used for BCM thought experiment
  const double bcmSmoothing = 0.99999; // proportion of old value of threshold
  itsBCMThreshold.post() = itsBCMThreshold.pre() * bcmSmoothing + clamp(value * value, 0.0, 1.0) * (1.0 - bcmSmoothing);

  return value;
}

void bd::BorderOwnershipNeuron::learnImpl( SystemState const & state, double currentActivation )
{
  if( state.learningRule == LearningRule::Hebbian )
  {
    shortTermHebbianNormalization( state.boMaxWeights.feedback * itsNormalizationCoefficients.feedback,
                                   currentActivation, state,
                                   ConnectionFlags::Dynamic | ConnectionFlags::Feedback );
  }
  else if( state.learningRule == LearningRule::BCM )
  {
    bcmUpdate( currentActivation, state, ConnectionFlags::Dynamic | ConnectionFlags::Feedback );
  }
  else if( state.learningRule == LearningRule::Conflict )
  {
    // Slow params: .5, .05
    // Fast params: .005, .005
    const double alpha = calculateShortTermAlpha(); // excitation
    const double beta = 1.0; // inhibition

    // Most connections will use the same pool of neurons to check what is strongly learned
    const int kappaFlagsFeedback = ConnectionFlags::Feedback | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

    // Feedback Update
    {
      const int flags = ConnectionFlags::Direct | ConnectionFlags::Feedback | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

      // Short term (skip if not on)
      if( currentActivation != 0.0 )
      {
        constexpr double localLearningRate = 1.00;
        constexpr double ltmRate = 1.0; // how much faster LTM should adjust

        shortTermUpdate( state, flags, kappaFlagsFeedback,
                         currentActivation, itsInhibitorySum,
                         localLearningRate, ltmRate,
                         alpha, beta );
      }

      // Long term
      propogateAccumulatedWeight( state, flags,
          state.boMaxWeights.feedback * itsNormalizationCoefficients.feedback,
          0.05,
          0.005,
          100.0 );
    }

    // Feedback Update Inhibitory
    {
      const int flags = ConnectionFlags::Direct | ConnectionFlags::Feedback | ConnectionFlags::Inhibitory | ConnectionFlags::Dynamic;

      // Short term (skip if not on)
      if( currentActivation != 0.0 )
      {
        constexpr double localLearningRate = 1.00;
        constexpr double ltmRate = 1.0; // how much faster LTM should adjust

        shortTermUpdate( state, flags, kappaFlagsFeedback,
        //shortTermUpdate( state, flags, flags,
                         currentActivation, itsInhibitorySum,
                         localLearningRate, ltmRate,
                         alpha, beta );
      }

      // Long term
      propogateAccumulatedWeight( state, flags,
          state.boMaxWeights.feedbackInhibitory * itsNormalizationCoefficients.feedback,
          0.05,
          0.005,
          100.0 );
    }

    { // Feedback Update Skip
      const int flags = ConnectionFlags::Skip | ConnectionFlags::Feedback | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

      // Short term (skip if not on)
      if( currentActivation != 0.0 )
      {
        constexpr double localLearningRate = 1.00;
        constexpr double ltmRate = 1.0; // how much faster LTM should adjust

        shortTermUpdate( state, flags, kappaFlagsFeedback,
                         currentActivation, itsInhibitorySum,
                         localLearningRate, ltmRate,
                         alpha, beta );
      }

      // Long term
      propogateAccumulatedWeight( state, flags,
          state.boMaxWeights.feedbackSkip * itsNormalizationCoefficients.feedbackSkip,
          0.05,
          0.005,
          100.0 );
    }

    // Feedback Update Skip Inhibitory
    {
      const int flags = ConnectionFlags::Skip | ConnectionFlags::Feedback | ConnectionFlags::Inhibitory | ConnectionFlags::Dynamic;

      // Short term (skip if not on)
      if( currentActivation != 0.0 )
      {
        constexpr double localLearningRate = 1.00;
        constexpr double ltmRate = 1.0; // how much faster LTM should adjust

        shortTermUpdate( state, flags, kappaFlagsFeedback,
                         currentActivation, itsInhibitorySum,
                         localLearningRate, ltmRate,
                         alpha, beta );
      }

      // Long term
      propogateAccumulatedWeight( state, flags,
          state.boMaxWeights.feedbackSkipInhibitory * itsNormalizationCoefficients.feedbackSkip,
          0.05,
          0.005,
          100.0 );
    }

    // Lateral Update
    //{
    //  const int flags = ConnectionFlags::Lateral | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

    //  // Short term (skip if not on)
    //  if( currentActivation != 0.0 )
    //  {
    //    constexpr double localLearningRate = 1.0;
    //    constexpr double ltmRate = 1.0; // how much faster LTM should adjust

    //    shortTermUpdate( state, flags, flags,
    //                     currentActivation, itsInhibitorySum,
    //                     localLearningRate, ltmRate,
    //                     alpha, beta );
    //  }

    //  // Long term
    //  propogateAccumulatedWeight( state, ConnectionFlags::Lateral | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic,
    //      state.boLateralTotalWeight * itsNormalizationCoefficients.lateral,
    //      0.5,
    //      0.05,
    //      100.0 );

    //  // inhibitory lateral weights
    //  updateLateralWeights( state, ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory,
    //                        0.5, // hybrid threshold
    //                        state.boLateralTotalWeight, currentActivation,
    //                        itsInhibitorySum, alpha, beta );

    //}
  } // end Conflict learning

  // Column connections learned using a modified LTM hebbian
  columnNormalization( state.boMaxWeights.column,
                       currentActivation,
                       state,
                       ConnectionFlags::Dynamic | ConnectionFlags::Column | ConnectionFlags::Inhibitory );

}
