#include "Combined.H"

#include "Wiring.H"
#include "Util.H"

#include <fstream>
#include <cereal/archives/binary.hpp>
#include <cereal/types/vector.hpp>

#include "ThreadingUtil.H"

bd::Combined::Combined(
                    size_t numBOPairs,
                    size_t gridWidth,
                    size_t gridHeight ) :
  /* General parameters */
  itsState(),
  itsNumBOPairs( numBOPairs ),
  itsGridWidth( gridWidth ),
  itsGridHeight( gridHeight ),
  itsV1NumScales(1),          // hard coded
  itsGroupingNumScales(1),    // hard coded
  /* V1 Layers */
  itsOnOffResponses( gridWidth, gridHeight ),
  itsOffOnResponses( gridWidth, gridHeight ),
  itsEdgeResponses( gridWidth, gridHeight, gridHeight * gridWidth * 16.0, PopulationStrategy::PoissonDisk ),
  /* V2 Layers */
  itsBOResponses( itsNumBOPairs ),
  itsGNResponses( 1 ) // scale hardcoded to one
{
  // Wire V1 /////////////////////////////////////////////////////////////
  {
    // V1 parameters
    const double radius = 0;
    const double spread = 2.5;
    const double lateralRadiusCoefficient = 0.6;
    const double probability = 1.0;

    // Feedforward connections
    wireSheetRingUniform( itsEdgeResponses, itsOnOffResponses,
        radius, spread, probability, 1.0 / 2.0, true,
        ConnectionFlags::Dynamic | ConnectionFlags::Feedforward );

    wireSheetRingUniform( itsEdgeResponses, itsOffOnResponses,
        radius, spread, probability, 1.0 / 2.0, true,
        ConnectionFlags::Dynamic | ConnectionFlags::Feedforward | ConnectionFlags::DisplayAsInhibitory );

    const double staticRadius = (0.85 + 0.37) / 2.0;
    const double staticSpread = (0.85 - 0.37) / 2.0;

    // Inhibitory connections
    wireSheetRingUniform( itsEdgeResponses, itsEdgeResponses, spread * lateralRadiusCoefficient * staticRadius, spread * lateralRadiusCoefficient * staticSpread,
                          probability, -0.50, false, // -itsState.v1lateralTotalWeight, true,
                          ConnectionFlags::Static | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );

    // Normalize connections
    calculateV1Coefficients( itsEdgeResponses, itsState );


    // Excitatory lateral connections
    //                            to             from,   radius, spread,        probability, scaling, flags
    wireSheetRingUniform( itsEdgeResponses, itsEdgeResponses, 0, spread * lateralRadiusCoefficient * 0.37,
                          probability, 1.2 * itsState.v1lateralTotalWeight, true,
                          ConnectionFlags::Static | ConnectionFlags::Lateral );

    // Normalize lateral excitatory connections
    normalizeV1LateralExcitatoryConnections( itsEdgeResponses, itsState );
  }

  // Wire V2 /////////////////////////////////////////////////////////////
  {
    // V2 parameters
    const double radius = 0;
    const double spread = 6.0;
    const double probability = 1.0;

    // Set up the network
    const double feedforwardNormalizer = itsState.gnMaxWeights.feedforward / static_cast<double>(itsNumBOPairs * itsV1NumScales);
    const double feedbackNormalizer = 1.0;

    // Create Grouping sheets
    const double groupingScaleFactor = 1.60; // how much larger subsequent scales should be
    for( size_t oG = 0; oG < itsGroupingNumScales; ++oG )
    {
      const double groupingScalar = std::pow( groupingScaleFactor, (double)oG );
      itsGNResponses[oG] = Sheet<GroupingNeuron>( gridWidth, gridHeight,
                                                  gridHeight * gridWidth / ( groupingScalar * 2.0 ),
                                                  PopulationStrategy::PoissonDisk );
    }

    for( auto & v : itsBOResponses ) v.resize( itsV1NumScales );

    // Create BO sheets and wire
    for( size_t p = 0; p < itsNumBOPairs; ++p )
      for( size_t s = 0; s < itsV1NumScales; ++s )
      {
        // Create border ownership sheets
        itsBOResponses[p][s] = BOSheet( itsGridWidth, itsGridHeight );
        itsBOResponses[p][s].orientation = p * ( M_PI / static_cast<double>( itsNumBOPairs ) );

        // Wire V1 to BO
        wireSheetRandom( itsBOResponses[p][s], itsEdgeResponses, 0.75, 1, 1.0, ConnectionFlags::Static | ConnectionFlags::Feedforward );

        // Wire BO to Grouping Neurons
        for( size_t oG = 0; oG < itsGroupingNumScales; ++oG )
        {
          const double groupingScalar = oG == 0 ? 1.0 : oG * groupingScaleFactor;
          const double groupingSpread = spread * groupingScalar;

          // (grouping to border) FEEDBACK
          wireSheetRingUniform( itsBOResponses[p][s], itsGNResponses[oG], radius, groupingSpread, probability, feedbackNormalizer, true,
              ConnectionFlags::Dynamic | ConnectionFlags::Feedback );

          // create reciprocal connection (border to grouping) FEEDFORWARD
          wireSheetRingUniform( itsGNResponses[oG], itsBOResponses[p][s], radius, groupingSpread, probability, feedforwardNormalizer, true,
              ConnectionFlags::Dynamic | ConnectionFlags::Feedforward );


          // FEEDFORWARD inhibitory connections
          wireSheetRingUniform( itsGNResponses[oG], itsBOResponses[p][s], 0, 1.0, probability, -0.25, false,
              ConnectionFlags::Static | ConnectionFlags::Feedforward | ConnectionFlags::Inhibitory );


          /* note that the coefficients used here for fb and ff (feedbackNormalizer, feedforwardNormalizer) are
             essentially meaningless as they will be individually set by the calculateXXYYCoefficients functions later on */
        } // end grouping neuron wiring
      }

// TODO: this has not been updated since simplifying/removing scale parameter    // Normalize FF and FB connections
//    calculateGroupingFFCoefficients( itsGNResponses, itsState ); // TODO: may need to modify the coefficient in state to be scale dependent
//    calculateBorderOwnershipFBCoefficients( itsBOResponses, itsState );

    // Create lateral Combined Ownership connections
    for( size_t p = 0; p < itsNumBOPairs; ++p )
      for( size_t s = 0; s < itsV1NumScales; ++s )
        for( size_t oo = 0; oo < itsNumBOPairs; ++oo )
          for( size_t ss = 0; ss < itsV1NumScales; ++ss )
          {
            // Inhibitory lateral connections, initially equal weight
            wireSheetRingUniform( itsBOResponses[p][s], itsBOResponses[oo][ss], 0, .01, probability, itsState.boMaxWeights.column, true,
                ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );
          }

    // Lateral inhibition and excitation in grouping layer
    for( size_t oG = 0; oG < itsGroupingNumScales; ++oG )
    {
      const double scaling = (1.0 + oG) * groupingScaleFactor;
      //                             to,              from, radius, spread, probability, scaling, flags
      wireSheetRing( itsGNResponses[oG], itsGNResponses[oG], 0,  0.6 * scaling, 1, 0.5 * scaling, ConnectionFlags::Static | ConnectionFlags::Lateral );
      wireSheetRing( itsGNResponses[oG], itsGNResponses[oG], 3 * scaling,  1.0 * scaling, 1, -10 * scaling, ConnectionFlags::Static | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );
    }

    // Normalize the GN lateral connections
    for( size_t oG = 0; oG < itsGroupingNumScales; ++oG )
      calculateCoefficients( itsGNResponses[oG] );
  } // end V2 setup
}

void bd::Combined::onset( cv::Mat const & image )
{
  // Downsample input
  cv::Mat downsampled = image;
  resizeInPlace( downsampled, cv::Size( itsGridWidth, itsGridHeight ) );

  // Convolution for onOff/offOn
  cv::Mat onOffKernel = (cv::Mat_<float>( 3, 3 ) <<
    -.125, -.25, -.125,
    -.25, 1.5, -.25,
    -.125, -.25, -.125);
  cv::Mat offOnKernel = -onOffKernel;

  cv::Mat onOffResult, offOnResult;

  cv::filter2D( downsampled, onOffResult, -1, onOffKernel );
  cv::filter2D( downsampled, offOnResult, -1, offOnKernel );

  const double modulation = 1.0;

  for( auto & vec : itsOnOffResponses )
    for( auto & neuron : vec )
    {
      int x, y;
      std::tie(x, y) = neuron->position();
      neuron->setActivation( clamp(onOffResult.at<float>( y, x ) * modulation) );
    }

  for( auto & vec : itsOffOnResponses )
    for( auto & neuron : vec )
    {
      int x, y;
      std::tie(x, y) = neuron->position();
      neuron->setActivation( clamp(offOnResult.at<float>( y, x ) * modulation) );
    }
}

void bd::Combined::iterate( int numIters )
{
  auto iterateImpl = [&]()
  {
    // UPDATE ===================================================================
    threading::parallel_for( 0lu, itsEdgeResponses.size(), [&](size_t i)
    {
      auto & vec = itsEdgeResponses[i];
      for( auto & neuron : vec )
        neuron->update( itsState );
    } );

    for( size_t p = 0; p < itsNumBOPairs; ++p )
      for( size_t s = 0; s < itsV1NumScales; ++s )
      {
        threading::parallel_for( 0lu, itsBOResponses[p][s].size(), [&](size_t i)
        {
          auto & vec = itsBOResponses[p][s][i];

          for( auto & neuron : vec )
            neuron->update( itsState );
        } );
      }

    for( size_t oG = 0; oG < itsGroupingNumScales; ++oG )
    {
      threading::parallel_for( 0lu, itsGNResponses[oG].size(), [&](size_t i)
      {
        auto & vec = itsGNResponses[oG][i];
        for( auto & neuron : vec )
          neuron->update( itsState );
      } );
    }

    // learn( itsState ) ====================================================================
    threading::parallel_for( 0lu, itsEdgeResponses.size(), [&](size_t i)
    {
      auto & vec = itsEdgeResponses[i];
      for( auto & neuron : vec )
        neuron->learn( itsState );
    } );

    for( size_t p = 0; p < itsNumBOPairs; ++p )
      for( size_t s = 0; s < itsV1NumScales; ++s )
      {
        threading::parallel_for( 0lu, itsBOResponses[p][s].size(), [&](size_t i)
        {
          auto & vec = itsBOResponses[p][s][i];
          for( auto & neuron : vec )
            neuron->learn( itsState );
        } );
      }

    for( size_t oG = 0; oG < itsGroupingNumScales; ++oG )
    {
      threading::parallel_for( 0lu, itsGNResponses[oG].size(), [&](size_t i)
      {
        auto & vec = itsGNResponses[oG][i];
        for( auto & neuron : vec )
          neuron->learn( itsState );
      } );
    }

    // FINALIZE =================================================================
    finalizeNeuronVariables();
//    threading::parallel_for( 0lu, itsEdgeResponses.size(), [&](size_t i)
//    {
//      auto & vec = itsEdgeResponses[i];
//      for( auto & neuron : vec )
//        neuron->finalize();
//    } );
//
//    for( size_t p = 0; p < itsNumBOPairs; ++p )
//      for( size_t s = 0; s < itsV1NumScales; ++s )
//      {
//        threading::parallel_for( 0lu, itsBOResponses[p][s].size(), [&](size_t i)
//        {
//          auto & vec = itsBOResponses[p][s][i];
//          for( auto & neuron : vec )
//            neuron->finalize();
//        } );
//      }
//
//    for( size_t oG = 0; oG < itsGroupingNumScales; ++oG )
//    {
//      threading::parallel_for( 0lu, itsGNResponses[oG].size(), [&](size_t i)
//      {
//        auto & vec = itsGNResponses[oG][i];
//        for( auto & neuron : vec )
//          neuron->finalize();
//      } );
//    }
  };

  for( int i = 0; i < numIters; ++i )
    iterateImpl();
}
