#include <cmath>
#include <math.h>
#include "Generate.H"
#include "../tests/Util.H"
#include "Display.H"

namespace
{
  // Returns z-component of cross product between AB and BC
  // Negative if C is to the left of A-->B
  double sideOfLine(std::pair<double, double> A, std::pair<double, double> B, std::pair<double, double> C) {
    return (C.first - B.first) * (B.second - A.second) - (C.second - B.second) * (B.first - A.first);
  }

  // Returns the angle between vectors A-->B and B-->C, after translating them so that they start in the same place.
  // Returns a value between -PI and PI.
  double counterClockwiseAngle(std::pair<double, double> A, std::pair<double, double> B, std::pair<double, double> C) {
    double val = std::atan2((B.first - A.first) * (C.second - B.second) - (B.second - A.second) * (C.first - B.first),
                            (B.first - A.first) * (C.first - B.first) + (B.second - A.second) * (C.second - B.second));
    return val;
  }

  bd::SystemState displayState;
}

std::string debugGetImageType(int number)
{
    // find type
    int imgTypeInt = number%8;
    std::string imgTypeString;

    switch (imgTypeInt)
    {
        case 0:
            imgTypeString = "8U";
            break;
        case 1:
            imgTypeString = "8S";
            break;
        case 2:
            imgTypeString = "16U";
            break;
        case 3:
            imgTypeString = "16S";
            break;
        case 4:
            imgTypeString = "32S";
            break;
        case 5:
            imgTypeString = "32F";
            break;
        case 6:
            imgTypeString = "64F";
            break;
        default:
            break;
    }

    // find channel
    int channel = (number/8) + 1;

    std::stringstream type;
    type<<"CV_"<<imgTypeString<<"C"<<channel;

    return type.str();
}


template <class T>
void debugPrint(cv::Mat const & mat, int x, int y, int size)
{
  if( mat.channels() != 3 )
    return;

  std::cout << "Image type is " << debugGetImageType(mat.type()) << std::endl;

  for( int yy = y; yy < y + size; ++yy)
  {
    for( int xx = x; xx < x + size; ++xx)
    {
      T pixel = mat.at<T>( yy, xx );
      std::cout << (int)pixel[0] << ", " << (int)pixel[1] << ", " << (int)pixel[2] << " || ";
    }
    std::cout << std::endl;
  }
}

void bd::setDisplayState( SystemState const & state )
{
  displayState = state;
}

std::string bd::showImage( cv::Mat const & image, bool normalize, bool wait, std::string const & name )
{
  static int counter = 0;
  const auto str = name.empty() ? std::to_string(counter++) : name;
  cv::namedWindow( str, cv::WINDOW_NORMAL );

  auto localImage = image;

  if( normalize )
    cv::normalize( localImage, localImage, 0, 1, cv::NORM_MINMAX );

  if( image.channels() == 3 )
  {
    auto rgb = image;
    rgb *= 255.0;
    rgb.convertTo( localImage, CV_8UC3 );
  }

  cv::imshow( str, localImage );
  if( wait )
    cv::waitKey( 0 );
  return str;
}

std::pair<cv::Mat, cv::Mat> bd::sheetActivationImage( Sheet<BorderOwnershipNeuron> const & sheet )
{
  cv::Mat image = cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC1 );
  cv::Mat image2 = cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC1 );

  for( const auto & cell : sheet )
  {
    const auto & neuron = cell[0];
    int x, y;
    std::tie(x, y) = neuron->position();
    image.at<float>( y, x ) += neuron->activation();// / 2.0;

    const auto & neuron2 = cell[1];
    std::tie(x, y) = neuron2->position();
    image2.at<float>( y, x ) += neuron2->activation();// / 2.0;
  }
  return {image, image2};
}

cv::Mat bd::visualizeInhibitoryPairs( Sheet<BorderOwnershipNeuron> const & sheet )
{
  cv::Mat image = cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC3 );

  for( const auto & cell : sheet )
  {
    const auto & left = cell[0];
    const auto & right = cell[1];

    int x, y;
    std::tie(x, y) = left->position();

    // Left neuron will be green, right will be red
    image.at<cv::Vec3f>( y, x ) = cv::Vec3f(0, left->activation(), right->activation());
  }

  return image;
}

std::pair<double, double> bd::display::getDirectionValues( std::shared_ptr<bd::BorderOwnershipNeuron> const & neuron )
{
  int x, y;
  std::tie(x, y) = neuron->position();

  const auto orientation = neuron->orientation;
  const auto myPosition = neuron->position();

  double posTotal = 0.0;
  double negTotal = 0.0;
  double posCount = 0;
  double negCount = 0;

  for( const auto & in : neuron->itsInput.pre() )
  {
    if( !in.hasFlag( bd::ConnectionFlags::Feedback ) )
      continue;

//    // Only consider neurons that are singletons
//    if( in.neuron->hasCompetitors() )
//      continue;

    if( in.hasFlag( bd::ConnectionFlags::Inhibitory) ) // TODO: properly handle negative weight
      continue;

    const auto inPosition = in.neuron->position();
    //const auto inPosition = in.neuron->centroid( bd::ConnectionFlags::Feedforward );

    // Compute angle between edge and perpendicular
    std::pair<double, double> dummy = {x - std::cos(-orientation + M_PI / 2), y - std::sin(-orientation + M_PI / 2)};
    const double angle = counterClockwiseAngle(dummy, myPosition, inPosition);

    const double orientationWeight = (M_PI/2 - std::abs(angle)); // Weight based on direction
    //const double orientationWeight = std::abs(angle) > M_PI/2 ? -1.0 : 1.0; // Weight only on which side
    //const double orientationWeight = [=]() // Weight by side with no weight given for ambiguous (along bisecting line)
    //{
    //  const double zero = (std::abs(angle) < (M_PI/2 + 1e-8)) && (std::abs(angle) > (M_PI/2 - 1e-10));
    //  return zero ? 0.0 : std::abs(angle) > M_PI/2 ? -1.0 : 1.0;

    //}();

    //double value = in.weight * orientationWeight;
    double value = in.ltWeight * orientationWeight;

    if( value > 0 )
    {
      posTotal += value;
      posCount += std::abs(orientationWeight);
    }
    else if( value < 0 )
    {
      negTotal += value;
      negCount += std::abs(orientationWeight);
    }
  }

  negTotal = std::abs(negTotal);

  const double maxWeight = neuron->itsNormalizationCoefficients.feedback * displayState.boMaxWeights.feedback;
  const double normalizer = (posTotal + negTotal) / maxWeight;
  const double maxTotal = std::max( posTotal, negTotal );

  posTotal *= normalizer / maxTotal;
  negTotal *= normalizer / maxTotal;

  posTotal = std::isnan(posTotal) ? 0.0 : posTotal;
  negTotal = std::isnan(negTotal) ? 0.0 : negTotal;

//  if( posCount > 0 )
//    posTotal /= posCount;
//  if( negCount > 0 )
//    negTotal /= negCount;

  return {posTotal, negTotal};
};

std::pair<int, int> bd::display::getPolarityIndex( std::shared_ptr<bd::BorderOwnershipNeuron> const & neuronA,
                                                   std::shared_ptr<bd::BorderOwnershipNeuron> const & neuronB )
{
  const auto & a = neuronA;
  const auto & b = neuronB;

  const auto valsA = display::getDirectionValues(a);
  const auto valsB = display::getDirectionValues(b);

  bool maxIsA = false;
  auto maxVals = valsB;

  if(std::max(valsA.first, valsA.second) >= std::max(valsB.first, valsB.second))
  {
    maxIsA = true;
    maxVals = valsA;
  }

  const int maxIndex = maxVals.first >= maxVals.second ? 0 : 1;
  const int aIndex = maxIsA ? maxIndex : 1 - maxIndex;
  const int bIndex = 1 - aIndex;

  return {aIndex, bIndex};
}

std::pair<double, double> bd::display::getPolarityAngles( std::shared_ptr<bd::BorderOwnershipNeuron> const & neuronA,
                                                          std::shared_ptr<bd::BorderOwnershipNeuron> const & neuronB )
{
  const auto indices = getPolarityIndex( neuronA, neuronB );
  const double baseOrientation = neuronA->orientation;

  return {baseOrientation + M_PI/2.0 - M_PI * indices.first,
          baseOrientation + M_PI/2.0 - M_PI * indices.second};
}


cv::Mat bd::visualizePrimaryDirection( Sheet<BorderOwnershipNeuron> const & sheet )
{
  cv::Mat image = cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC3 );

  for( const auto & cell : sheet )
  {
    const auto & left = cell[0];
    const auto & right = cell[1];

    int x, y;
    std::tie(x, y) = left->position();

    const double leftAct  = left->activation();
    const double rightAct = right->activation();

    // skip displaying neurons that are off
    if( leftAct < 0.05 && rightAct < 0.05 )
      continue;

    std::pair<double, double> colors = {0, 0};

    if( leftAct > rightAct )
      colors = display::getDirectionValues( left );
    else if( leftAct == rightAct )
      image.at<cv::Vec3f>( y, x ) += cv::Vec3f( 1, 1, 1 );
    else
      colors = display::getDirectionValues( right );

    image.at<cv::Vec3f>( y, x ) += cv::Vec3f( 0, colors.first, colors.second );
  }

  return image;
}

namespace
{
  struct Color {
    double r,g,b;
  };

  //! Maps the value v in the range [vmin, vmax] to a rainbow like color
  Color GetColour(double v,double vmin,double vmax)
  {
    Color c = {1.0,1.0,1.0}; // white
    double dv;

    if (v < vmin)
      v = vmin;
    if (v > vmax)
      v = vmax;
    dv = vmax - vmin;

    if (v < (vmin + 0.25 * dv)) {
      c.r = 0;
      c.g = 4 * (v - vmin) / dv;
    } else if (v < (vmin + 0.5 * dv)) {
      c.r = 0;
      c.b = 1 + 4 * (vmin + 0.25 * dv - v) / dv;
    } else if (v < (vmin + 0.75 * dv)) {
      c.r = 4 * (v - vmin - 0.5 * dv) / dv;
      c.b = 0;
    } else {
      c.g = 1 + 4 * (vmin + 0.75 * dv - v) / dv;
      c.b = 0;
    }

    return(c);
  }
}

cv::Mat bd::visualizeAllDirections( std::vector<Sheet<BorderOwnershipNeuron>> const & sheets, double resolutionScaling )
{
  const int height = sheets[0].height();
  const int width  = sheets[0].width();

  cv::Mat sinAngle = cv::Mat::zeros( height, width, CV_32FC1 );
  cv::Mat cosAngle = cv::Mat::zeros( height, width, CV_32FC1 );
  cv::Mat value    = cv::Mat::zeros( height, width, CV_32FC1 );

  // Aggregate polarity values
  for( const auto & orientationSheet : sheets )
  {
    for( const auto & cell : orientationSheet )
    {
      const auto & neuronA = cell[0];
      const auto & neuronB = cell[1];

      int x, y;
      std::tie(x, y) = neuronA->position();

      const double actA = neuronA->activation();
      const double actB = neuronB->activation();

      const auto angles = bd::display::getPolarityAngles( neuronA, neuronB );

      cosAngle.at<float>( y, x ) += std::cos( angles.first ) * actA + std::cos( angles.second ) * actB;
      sinAngle.at<float>( y, x ) += std::sin( angles.first ) * actA + std::sin( angles.second ) * actB;
      value.at<float>( y, x ) += std::abs( actA - actB );
    }
  }

  // Create map of all ground truth angles, in degrees
  cv::Mat unused = cv::Mat::zeros( height, width, CV_32FC1 );
  cv::Mat hue = cv::Mat::zeros( height, width, CV_32FC1 );
  cv::cartToPolar( sinAngle, cosAngle, unused, hue, true ); // reverse x,y for image coordinates

  // Convert saturation and hue to HSV image
  cv::Mat colorImage = cv::Mat::zeros( height, width, CV_32FC3 );

  std::vector<cv::Mat> channels;
  channels.push_back( hue );
  channels.push_back( cv::Mat::ones( height, width, CV_32FC1 ) );
  channels.push_back( value );
  cv::merge( channels, colorImage );

  // Convert HSV to RGB for final image
  cv::cvtColor( colorImage, colorImage, CV_HSV2BGR );
  cv::normalize( colorImage, colorImage, 0, 1, cv::NORM_MINMAX );

  return colorImage;
}

cv::Mat bd::convertToWhiteBackground( cv::Mat const & image )
{
  // Convert to HSV
  cv::Mat imageHLS = cv::Mat::zeros( image.rows, image.cols, CV_32FC3 ); // used to color BO inputs
  cv::cvtColor( image, imageHLS, CV_BGR2HLS );

  // Set all value to 1 - value
  for( auto it = imageHLS.begin<cv::Vec3f>(), end = imageHLS.end<cv::Vec3f>(); it != end; ++it )
    (*it)[1] = 1.0f - (*it)[1];

  cv::cvtColor( imageHLS, imageHLS, CV_HLS2BGR );
  return imageHLS;
}

namespace display_impl
{
  using Data = std::pair<bd::Sheet<bd::BorderOwnershipNeuron>*, bd::Sheet<bd::GroupingNeuron>*>;

  template <size_t Index> inline
  void windowCallbackBO( int event, int mX, int mY, int /*flags*/, void* userdata )
  {
    if( event == cv::EVENT_LBUTTONDOWN || event == cv::EVENT_RBUTTONDOWN )
    {
      const bool verbose = event == cv::EVENT_RBUTTONDOWN;
      //std::pair<bd::Sheet<bd::BorderOwnershipNeuron>*, bd::Sheet<bd::GroupingNeuron>*> data =
      auto data = *reinterpret_cast<Data*>( userdata );

      auto & sheet  = *data.first;
      auto & gSheet = *data.second;
      //bd::Sheet<bd::BorderOwnershipNeuron> sheet = *reinterpret_cast<bd::Sheet<bd::BorderOwnershipNeuron>*>( userdata );

      // Get neuron closest to mouse click
      std::shared_ptr<const bd::BorderOwnershipNeuron> n;
      double minDist = std::numeric_limits<double>::max();

      for( const auto & cell : sheet )
      {
        const auto & neuron = cell[Index];
        double x, y;
        std::tie(x, y) = neuron->position();
        double dist = (mX - x) * (mX - x) + (mY - y) * (mY - y);
        if ( dist < minDist )
        {
          n = neuron;
          minDist = dist;
        }
      }

      // Create image of neuron connections
      cv::Mat fbImage = cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC3 );
      cv::Mat fbLTImage = cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC3 );
      cv::Mat ffImage = cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC3 );
      cv::Mat latImage = cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC3 );

      // Mark center blue
      int myX, myY; std::tie(myX, myY) = n->position();
      n->printNeuronInfo();

      // Generate feedback and lateral outgoing connections
      for( const auto & in : n->itsInput.post() )
      {
        int inX, inY;
        std::tie(inX, inY) = in.neuron->position();

        cv::Vec3f color;
        cv::Vec3f colorLT;

        if( in.hasFlag( bd::ConnectionFlags::Excitatory ) )
        {
          color = cv::Vec3f(0, in.weight, 0);
          colorLT = cv::Vec3f(0, in.ltWeight, 0);
        }
        else
        {
          color = cv::Vec3f(0, 0, std::abs(in.weight));
          colorLT = cv::Vec3f(0, 0, std::abs(in.ltWeight));
        }

        if( in.hasFlag( bd::ConnectionFlags::Feedback ) )
        {
          fbImage.at<cv::Vec3f>( inY, inX ) += color;
          fbLTImage.at<cv::Vec3f>( inY, inX ) += colorLT;
        }
        else if( in.hasFlag( bd::ConnectionFlags::Lateral ) )
          latImage.at<cv::Vec3f>( inY, inX ) += color;
      }

      // Generate feedforward image
      // if print is true, just prints out connections, else does just image generation
      auto generateFFImage = [&](bool print, std::string const & title)
      {
        std::cout << title << std::endl;

        double weightSum = 0.0;
        double ltSum = 0.0;
        double accSum = 0.0;
        double numWeights = 0;

        for( const auto & cell : gSheet )
          for( const auto & neuron : cell )
          {
            // Find every neuron in grouping that has input from me
            for( const auto & in : neuron->itsInput.post() )
            {
              if( *in.neuron == *n && in.hasFlag( bd::ConnectionFlags::Feedforward ) )
              {
                int x, y;
                std::tie(x, y) = neuron->position();

                cv::Vec3f color;
                if( in.weight > 0 )
                  color = cv::Vec3f(0, in.weight, 0);
                else
                  color = cv::Vec3f(0, 0, -in.weight);

                weightSum += in.weight;
                ltSum += in.ltWeight;
                accSum += in.accumulatedWeight;
                ++numWeights;

                if( print )
                {
                  std::cout << std::setw(2);
                  std::cout << x << ", " << std::setw(2) << y << std::fixed;
                  std::cout << std::setprecision(3) << ", act: " << neuron->activation();
                  std::cout << std::setprecision(6) << ", weight: " << in.weight << ", ltWeight: " << in.ltWeight;
                  std::cout << ", acc: " << in.accumulatedWeight;
                  std::cout << ", bcmthresh: " << neuron->itsBCMThreshold.pre();
                  std::cout << ", maxThresh: " << neuron->itsMaxThreshold;
                  std::cout << ", actThresh: " << neuron->itsActivationThreshold;
                  std::cout << ", decayThresh: " << neuron->itsDecayThreshold << std::endl;
                }

                ffImage.at<cv::Vec3f>( y, x ) += color;

                break; // there will only be one FF projection from me to
                       // any single grouping neuron
              } // end matching neuron
            } // end input loop
          } // end neuron loop

        double stdTotal = 0.0;
        // Calculate standard deviation of accumulator
        for( const auto & cell : gSheet )
          for( const auto & neuron : cell )
          {
            // Find every neuron in grouping that has input from me
            for( const auto & in : neuron->itsInput.post() )
            {
              if( *in.neuron == *n && in.hasFlag( bd::ConnectionFlags::Feedforward ) )
                stdTotal += std::pow( in.accumulatedWeight - (accSum/numWeights), 2.0 );
            }
          }

        std::cout << "Weight average: " << weightSum / numWeights << " sum: " << weightSum;
        std::cout << ", ltsum: " << ltSum << ", accsum: " << accSum << ", accmean: " << accSum / numWeights;
        std::cout << ", accstd: " << std::sqrt(stdTotal/numWeights) << std::endl;
      };

      n->printConnections( "Incoming Column Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Column ); }, true );
      n->printConnections( "Incoming Feedforward Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedforward ); }, verbose );
      n->printConnections( "Feedback Connections(+d)", [](bd::Input const & in)
          { return in.hasFlag( bd::ConnectionFlags::Feedback | bd::ConnectionFlags::Direct | bd::ConnectionFlags::Excitatory ); }, verbose );
      n->printConnections( "Feedback Connections(-d)", [](bd::Input const & in)
          { return in.hasFlag( bd::ConnectionFlags::Feedback | bd::ConnectionFlags::Direct | bd::ConnectionFlags::Inhibitory ); }, verbose );
      n->printConnections( "Feedback Connections(+s)", [](bd::Input const & in)
          { return in.hasFlag( bd::ConnectionFlags::Feedback | bd::ConnectionFlags::Skip | bd::ConnectionFlags::Excitatory ); }, verbose );
      n->printConnections( "Feedback Connections(-s)", [](bd::Input const & in)
          { return in.hasFlag( bd::ConnectionFlags::Feedback | bd::ConnectionFlags::Skip | bd::ConnectionFlags::Inhibitory ); }, verbose );
      n->printConnections( "Incoming Lateral + Connections",   [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Lateral ) && in.weight >= 0.0; }, verbose );
      n->printConnections( "Incoming Lateral - Connections",   [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Lateral | bd::ConnectionFlags::Inhibitory ) ||
                                                                                                (in.hasFlag( bd::ConnectionFlags::Lateral ) && in.weight < 0.0); }, verbose );
      generateFFImage( verbose, "Outgoing Feedforward Connections" );

      static const std::string feedbackName = "BO Incoming Feedback Neuron Connections";
      static const std::string feedbackLTName = "BO Incoming LT Feedback Neuron Connections";
      static const std::string feedforwardName = "BO Outgoing Feedforward Neuron Connections";
      static const std::string lateralName = "BO Incoming Lateral Neuron Connections";

      const std::string savePrefix = "BO_" + std::to_string(myX) + "_" + std::to_string(myY) + "_";

      cv::normalize( fbImage, fbImage, 0, 1, cv::NORM_MINMAX );
      fbImage.at<cv::Vec3f>( myY, myX ) += cv::Vec3f{1.0, 0, 0};

      cv::namedWindow( feedbackName, cv::WINDOW_NORMAL );
      cv::imshow( feedbackName, fbImage );
      cv::imwrite( savePrefix + "feedback.png", fbImage * 255.0 );

      {
        auto whiteBG = bd::convertToWhiteBackground( fbImage );
        cv::namedWindow( "BO Incoming Feedback Neuron Connections WHITE", cv::WINDOW_NORMAL );
        cv::imshow( "BO Incoming Feedback Neuron Connections WHITE", whiteBG );
        cv::imwrite( savePrefix + "feedback_white.png", whiteBG * 255.0 );
      }

      cv::normalize( fbLTImage, fbLTImage, 0, 1, cv::NORM_MINMAX );
      fbLTImage.at<cv::Vec3f>( myY, myX ) += cv::Vec3f{1.0, 0, 0};

      cv::namedWindow( feedbackLTName, cv::WINDOW_NORMAL );
      cv::imshow( feedbackLTName, fbLTImage );
      cv::imwrite( savePrefix + "feedback_lt.png", fbLTImage * 255.0 );

      cv::normalize( ffImage, ffImage, 0, 1, cv::NORM_MINMAX );
      ffImage.at<cv::Vec3f>( myY, myX ) += cv::Vec3f{1.0, 0, 0};

      cv::namedWindow( feedforwardName, cv::WINDOW_NORMAL );
      cv::imshow( feedforwardName, ffImage );
      cv::imwrite( savePrefix + "outgoing_feedforward.png", ffImage * 255.0 );

      cv::normalize( latImage, latImage, 0, 1, cv::NORM_MINMAX );
      latImage.at<cv::Vec3f>( myY, myX ) += cv::Vec3f{1.0, 0, 0};

      cv::namedWindow( lateralName, cv::WINDOW_NORMAL );
      cv::imshow( lateralName, latImage );
      cv::imwrite( savePrefix + "lateral.png", latImage * 255.0 );
    }
  }
}

void bd::showConnections( Sheet<BorderOwnershipNeuron> const & sheet, Sheet<GroupingNeuron> const & gSheet, bool normalize, std::string const & name )
{
  auto data = new display_impl::Data{ const_cast<Sheet<BorderOwnershipNeuron>*>(reinterpret_cast<Sheet<BorderOwnershipNeuron> const *>(&sheet)),
                                      const_cast<Sheet<GroupingNeuron>*>(reinterpret_cast<Sheet<GroupingNeuron> const *>(&gSheet)) };

  std::string name1 = name.empty() ? name : name + "1";
  std::string name2 = name.empty() ? name : name + "2";

  auto images = sheetActivationImage( sheet );
  {
    std::string sheetWindow = showImage(images.first, normalize, false, name1);
    cv::setMouseCallback( sheetWindow, &display_impl::windowCallbackBO<0>, data );
  }

  {
    std::string sheetWindow = showImage(images.second, normalize, false, name2);
    cv::setMouseCallback( sheetWindow, &display_impl::windowCallbackBO<1>, data );
  }
  //cv::waitKey(0);
}

void bd::displayPolarity( Sheet<BorderOwnershipNeuron> const & sheet, Sheet<GroupingNeuron> const & gSheet, bool normalize, std::string const & name )
{
  // Create image of neuron connections
  cv::Mat images[] = {cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC3 ),
                      cv::Mat::zeros( sheet.height(), sheet.width(), CV_32FC3 )};

  for(int index = 0; index < 2; ++index)
  {
    for( const auto & cell : sheet )
    {
      const auto & neuron = cell[index];
      int x, y;
      std::tie(x, y) = neuron->position();

      auto colors = display::getDirectionValues( neuron );
      images[index].at<cv::Vec3f>( y, x ) += cv::Vec3f( 0, colors.first, colors.second );
    }
  } // END INDEX

  auto data = new display_impl::Data{ const_cast<Sheet<BorderOwnershipNeuron>*>(reinterpret_cast<Sheet<BorderOwnershipNeuron> const *>(&sheet)),
                                      const_cast<Sheet<GroupingNeuron>*>(reinterpret_cast<Sheet<GroupingNeuron> const *>(&gSheet)) };

  std::string name1 = name.empty() ? name : name + "1";
  std::string name2 = name.empty() ? name : name + "2";

  auto window0 = showImage(images[0], normalize, false, name1);
  cv::setMouseCallback( window0, &display_impl::windowCallbackBO<0>, data );
  cv::imwrite( "polarity0.png", images[0] );

  auto window1 = showImage(images[1], normalize, false, name2);
  cv::setMouseCallback( window1, &display_impl::windowCallbackBO<1>, data );
  cv::imwrite( "polarity1.png", images[1] );
  //cv::waitKey(0);
}
