#ifndef DISPLAY_H_
#define DISPLAY_H_

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <string>
#include <iomanip>
#include <functional>

#include "BorderOwnershipNeuron.H"
#include "GroupingNeuron.H"
#include "Sheet.H"

namespace bd
{
  //! Sets a global state for all display functions to use
  /*! This is a hacky workaround of accessing the system state without having to pass it to every function */
  void setDisplayState( SystemState const & state );

  //! Creates a window showing an image
  /*! @param image The image to show
      @param noramlize Whether the channels should be normalized
      @param wait Whether we should pause for user response
      @param name The name of the window; if empty, a new name will be generated. If existing, the window will be updated
      @return The name of this window */
  std::string showImage( cv::Mat const & image, bool normalize = true, bool wait = true, std::string const & name = {} );

  /*! ========================================================================*/
  //! Creates a red-green opponency image from the activation of BO neurons
  /*! This image is non-interactive.  Shows the dominant BO neuron at each location as
      either red or green, or a mixing (yellow) paired with a black if one has overtaken
      the other */
  cv::Mat visualizeInhibitoryPairs( Sheet<BorderOwnershipNeuron> const & sheet );

  //! For currently active BO neurons, shows the primary direction currently selected
  /*! This works on a single BO sheet and will use a red-green color scheme */
  cv::Mat visualizePrimaryDirection( Sheet<BorderOwnershipNeuron> const & sheet );

  //! For all currently active BO neurons, shows the primary direction currently selected
  /*! This works over all BO sheets and will give each prominent direction a color */
  cv::Mat visualizeAllDirections( std::vector<Sheet<BorderOwnershipNeuron>> const & sheets, double resolutionScaling );

  /*! ========================================================================*/
  //! Converts a color image with a black background to one with a white background
  cv::Mat convertToWhiteBackground( cv::Mat const & image );

  /*! ========================================================================*/
  //! Create an image of this sheet with pixels representing neurons
  template <class NeuronT>
  cv::Mat sheetNeuronImage( Sheet<NeuronT> const & sheet, double scaling = 1.0 );

  //! Create an image of this sheet using a parameter function to create a value from each neuron
  template <class NeuronT, class F>
  cv::Mat sheetNeuronValueImage( Sheet<NeuronT> const & sheet,
                                 F && func = [](std::shared_ptr<NeuronT> const & n){ return n->activation(); },
                                 double scaling = 1.0 );

  //! Create an image of this sheet with pixels representing activation
  template <class NeuronT>
  cv::Mat sheetActivationImage( Sheet<NeuronT> const & sheet, double scaling = 1.0 );

  //! Specialization for BO sheets, produces an image for both pairs of BO neurons
  std::pair<cv::Mat, cv::Mat> sheetActivationImage( Sheet<BorderOwnershipNeuron> const & sheet );

  /*! ========================================================================*/
  //! Displays an interactive image that allows inspections of connection weights
  template <class NeuronT>
  void showConnections( Sheet<NeuronT> const & sheet, bool normalize = false, bool showPositions = false, std::string const & name = {} );

  //! Special overload for BO neurons
  void showConnections( Sheet<BorderOwnershipNeuron> const & sheet, Sheet<GroupingNeuron> const & gSheet, bool normalize = false, std::string const & name = {} );

  /*! ========================================================================*/
  //! Displays the preferred polarity of every neuron in a BO sheet
  /*! Callback enables clicking to see incoming feedback and lateral connections
      Also displays outgoing feedforward connections to the grouping sheet */
  void displayPolarity( Sheet<BorderOwnershipNeuron> const & sheet, Sheet<GroupingNeuron> const & gSheet, bool normalize = false, std::string const & name = {} );

  namespace display
  {
    //! Get direction values for a specific neuron in a BD sheet
    /*! This function computes a score for all learned weights on each side of the oriented boundary the
        neuron receives input from. */
    std::pair<double, double> getDirectionValues( std::shared_ptr<bd::BorderOwnershipNeuron> const & neuron );

    //! Returns an assigned polarity index (0 or 1) for two BO neurons that form a pair
    std::pair<int, int> getPolarityIndex( std::shared_ptr<bd::BorderOwnershipNeuron> const & neuronA,
                                          std::shared_ptr<bd::BorderOwnershipNeuron> const & neuronB );

    //! Returns the primary direction assigned to a polarity for two BO neurons that form a pair
    std::pair<double, double> getPolarityAngles( std::shared_ptr<bd::BorderOwnershipNeuron> const & neuronA,
                                                 std::shared_ptr<bd::BorderOwnershipNeuron> const & neuronB );
  }
}

#include "DisplayImpl.H"

#endif // DISPLAY_H_
