#ifndef DISPLAYIMPL_H_
#define DISPLAYIMPL_H_

#include <iostream>
#include "EdgeNeuron.H"

template <class NeuronT> inline
cv::Mat bd::sheetNeuronImage( Sheet<NeuronT> const & sheet, double scaling )
{
  return sheetNeuronValueImage( sheet,
                                [](std::shared_ptr<NeuronT> const & n){ return 1.0f; },
                                scaling );
}

template <class NeuronT> inline
cv::Mat bd::sheetActivationImage( Sheet<NeuronT> const & sheet, double scaling )
{
  return sheetNeuronValueImage( sheet,
                                [](std::shared_ptr<NeuronT> const & n){ return n->activation(); },
                                scaling );
}

template <class NeuronT, class F> inline
cv::Mat bd::sheetNeuronValueImage( Sheet<NeuronT> const & sheet,
                                   F && func,
                                   double scaling )
{
  cv::Mat image = cv::Mat::zeros( sheet.height() * scaling, sheet.width() * scaling, CV_32FC1 );

  for( const auto & cell : sheet )
    for( const auto & neuron : cell )
    {
      double x, y;
      std::tie(x, y) = neuron->position();
      image.at<float>( y * scaling, x * scaling ) += func( neuron );
    }

  return image;
}

namespace display_impl
{
  template <class NeuronT> inline
  void windowCallback( int event, int mX, int mY, int flags, void* userdata )
  {
    if( event == cv::EVENT_LBUTTONDOWN || event == cv::EVENT_RBUTTONDOWN )
    {
      const bool verbose = event == cv::EVENT_RBUTTONDOWN;
      bd::Sheet<NeuronT> & sheet = *reinterpret_cast<bd::Sheet<NeuronT>*>( userdata );

      int maxPerGrid = 0;
      for( auto const & cell : sheet )
        maxPerGrid = std::max( maxPerGrid, (int)cell.size() );
      const double scaling    = maxPerGrid;

      // Get neuron closest to mouse click
      std::shared_ptr<const bd::Neuron> n;
      double minDist = std::numeric_limits<double>::max();

      for( const auto & cell : sheet )
        for( const auto & neuron : cell )
        {
          double x, y;
          std::tie(x, y) = neuron->position();
          x *= scaling; y *= scaling;

          double dist = (mX - x) * (mX - x) + (mY - y) * (mY - y);
          if ( dist < minDist )
          {
            n = neuron;
            minDist = dist;
          }
        }

      // Assemble all neurons in column
      std::vector<std::shared_ptr<const bd::Neuron>> column;
      column.push_back( n );
      for( const auto & in : n->itsInput.pre() )
        if( in.hasFlag( bd::ConnectionFlags::Column ) )
          column.push_back( in.neuron );

      const int columnSize = column.size();

      // dimensions for column connection grid image
      const int nColRows = std::max( 1, (columnSize / 4) + (columnSize % 4 ? 1 : 0));
      const int nColCols = std::min( 4, columnSize );

      // Returns the grid position in the column image for a given column index
      auto colIndexToRowCol = [=](int index){ return std::make_pair( index / 4, index % 4 ); };

      const double scaledHeight = sheet.height() * scaling;
      const double scaledWidth = sheet.width() * scaling;

      // Create image of neuron connections
      cv::Mat ffImage    = cv::Mat::zeros( scaledHeight * nColRows, scaledWidth * nColCols, CV_32FC3 );
      cv::Mat ffImageRaw = cv::Mat::zeros( scaledHeight * nColRows, scaledWidth * nColCols, CV_32FC3 ); // for normalizing pure positive FF
      cv::Mat ffImageHSV = cv::Mat::zeros( scaledHeight * nColRows, scaledWidth * nColCols, CV_32FC3 ); // used to color BO inputs
      cv::Mat ffLTImage    = cv::Mat::zeros( scaledHeight * nColRows, scaledWidth * nColCols, CV_32FC3 );
      cv::Mat ffLTImageRaw = cv::Mat::zeros( scaledHeight * nColRows, scaledWidth * nColCols, CV_32FC3 ); // for normalizing pure positive FF
      cv::Mat ffLTImageHSV = cv::Mat::zeros( scaledHeight * nColRows, scaledWidth * nColCols, CV_32FC3 ); // used to color BO inputs
      cv::Mat fbImage    = cv::Mat::zeros( scaledHeight * nColRows, scaledWidth * nColCols, CV_32FC3 );
      cv::Mat latImage   = cv::Mat::zeros( scaledHeight * nColRows, scaledWidth * nColCols, CV_32FC3 );
      cv::Mat latImageRaw   = cv::Mat::zeros( scaledHeight * nColRows, scaledWidth * nColCols, CV_32FC3 ); // for normalizing pure positive LAT

      // Returns a reference to the pixel belonging to the specified column
      auto imageRef = [&](double yy, double xx, cv::Mat & image, int colIndex) -> cv::Vec3f &
      {
        const auto rowCol = colIndexToRowCol( colIndex );
        return image.at<cv::Vec3f>( yy + scaledHeight * rowCol.first,
                                    xx + scaledWidth * rowCol.second );
      };

      // Draws a 3x3 pixel at the target location
      auto drawBigPixel = [&]( double yy, double xx, cv::Mat & image, int colIndex, cv::Vec3f const & color )
      {
        const auto rowCol = colIndexToRowCol( colIndex );

        const int y = yy + scaledHeight * rowCol.first;
        const int x = xx + scaledWidth * rowCol.second;

        if( scaling == 1 )
          image.at<cv::Vec3f>( y, x ) += color;
        else
          for( int i = -1; i < 2; ++i )
            for( int j = -1; j < 2; ++j )
              image.at<cv::Vec3f>( y + i, x + j ) += color;
      };

      double myX, myY; std::tie(myX, myY) = n->position();
      myX *= scaling; myY *= scaling;

      n->printNeuronInfo();

      // Color in activations for images
      for( int c = 0; c < columnSize; ++c )
      {
        const auto & colN = column[c];

        struct boMetadata
        {
          std::shared_ptr<bd::BorderOwnershipNeuron> n;
          double stWeight;
          double ltWeight;
        };

        std::map<std::pair<int, int>, std::vector<boMetadata>> ffBOMetadata;

        for( const auto & in : colN->itsInput.pre() )
        {
          double inX, inY;
          std::tie(inX, inY) = in.neuron->position();
          inX *= scaling; inY *= scaling;

          cv::Vec3f color;
          if( in.weight >= 0 && !(in.hasFlag( bd::ConnectionFlags::Inhibitory ) || in.hasFlag( bd::ConnectionFlags::DisplayAsInhibitory )) )
            color = cv::Vec3f(0, in.weight, 0);
          else // assume inhibition
            color = cv::Vec3f(0, 0, std::abs(in.weight));

          // long term
          cv::Vec3f colorLT;
          if( in.ltWeight >= 0 && !(in.hasFlag( bd::ConnectionFlags::Inhibitory ) || in.hasFlag( bd::ConnectionFlags::DisplayAsInhibitory )) )
            colorLT = cv::Vec3f(0, in.ltWeight, 0);
          else // assume inhibition
            colorLT = cv::Vec3f(0, 0, std::abs(in.ltWeight));

          if( in.hasFlag( bd::ConnectionFlags::Feedforward ) )
          {
            auto bo_in = std::dynamic_pointer_cast<bd::BorderOwnershipNeuron>( in.neuron );
            if( bo_in && in.weight > 0 )
            {
              const double st = in.weight;
              const double lt = in.ltWeight;
              ffBOMetadata[std::make_pair<int, int>(inY, inX)].push_back({bo_in, st, lt});
            }
            else
            {
              if( in.weight > 0 )
                imageRef( inY, inX, ffImageRaw, c ) += color;
              else
                imageRef( inY, inX, ffImage, c ) += color;

              if( in.ltWeight > 0 )
                imageRef( inY, inX, ffLTImageRaw, c ) += colorLT;
              else
                imageRef( inY, inX, ffLTImage, c ) += colorLT;
            }
          }
          else if( in.hasFlag( bd::ConnectionFlags::Lateral ) )
          {
            if( in.weight > 0 )
              imageRef( inY, inX, latImageRaw, c ) += color;
            else
              imageRef( inY, inX, latImage, c ) += color;
          }
          else if( in.hasFlag( bd::ConnectionFlags::Feedback ) )
            //imageRef( inY, inX, fbImage, c ) += color;
            drawBigPixel( inY, inX, fbImage, c, color );
        } // end coloring current column image

        if( ffBOMetadata.size() )
        {
          for( auto metaIter = ffBOMetadata.begin(), end = ffBOMetadata.end(); metaIter != end; ++metaIter )
          {
            double inY, inX;
            std::tie( inY, inX ) = metaIter->first;

            double xxST = 0.0;
            double yyST = 0.0;
            double totalST = 0.0;

            double xxLT = 0.0;
            double yyLT = 0.0;
            double totalLT = 0.0;

            const auto & vec = metaIter->second;
            for( size_t o = 0; o < 4; ++o )
            {
              const auto & aMeta = vec[o*2+0];
              const auto & bMeta = vec[o*2+1];

              auto angles = bd::display::getPolarityAngles( aMeta.n, bMeta.n );

              xxST += std::cos( angles.first ) * aMeta.stWeight + std::cos( angles.second ) * bMeta.stWeight;
              yyST += std::sin( angles.first ) * aMeta.stWeight + std::sin( angles.second ) * bMeta.stWeight;
              totalST += std::abs(aMeta.stWeight - bMeta.stWeight);

              xxLT += std::cos( angles.first ) * aMeta.ltWeight + std::cos( angles.second ) * bMeta.ltWeight;
              yyLT += std::sin( angles.first ) * aMeta.ltWeight + std::sin( angles.second ) * bMeta.ltWeight;
              totalLT += std::abs(aMeta.ltWeight - bMeta.ltWeight);
            }

            // reverse x and y to atan2 since we are in image coordinates
            const double thetaST = std::fmod( std::atan2( xxST, yyST ) + 2.0 * M_PI, 2.0 * M_PI ) * 180.0 / M_PI;
            const double thetaLT = std::fmod( std::atan2( xxLT, yyLT ) + 2.0 * M_PI, 2.0 * M_PI ) * 180.0 / M_PI;

            //imageRef( inY, inX, ffImageHSV, c )   += cv::Vec3f{(float)thetaST, 1.0f, (float)totalST};
            //imageRef( inY, inX, ffLTImageHSV, c ) += cv::Vec3f{(float)thetaLT, 1.0f, (float)totalLT};
            drawBigPixel( inY, inX, ffImageHSV, c, cv::Vec3f{(float)thetaST, 1.0f, (float)totalST} );
            drawBigPixel( inY, inX, ffLTImageHSV, c, cv::Vec3f{(float)thetaLT, 1.0f, (float)totalLT} );
          }

          if( c == columnSize - 1) // convert back to RGB on last column (when done with coloring)
          {
            cv::cvtColor( ffImageHSV, ffImageHSV, CV_HSV2BGR );
            cv::cvtColor( ffLTImageHSV, ffLTImageHSV, CV_HSV2BGR );
          }
        }
      }

      const std::string savePrefix = "Neuron_" + std::to_string(myX) + "_" + std::to_string(myY) + "_";

      cv::normalize( ffImageRaw, ffImageRaw, 0, 1, cv::NORM_MINMAX );
      cv::normalize( ffImageHSV, ffImageHSV, 0, 1, cv::NORM_MINMAX );
      cv::imwrite( savePrefix + "hsv.png", ffImageHSV * 255.0 );

      {
        auto whiteBG = bd::convertToWhiteBackground( ffImageHSV );
        cv::namedWindow( "hsv WHITE", cv::WINDOW_NORMAL );
        cv::imshow( "hsv WHITE", whiteBG );
        cv::imwrite( savePrefix + "hsv_white.png", whiteBG * 255.0 );
      }

      cv::namedWindow( "hsv", cv::WINDOW_NORMAL );
      cv::imshow( "hsv", ffImageHSV );
      cv::namedWindow( "raw", cv::WINDOW_NORMAL );
      cv::imshow( "raw", ffImageRaw );

      ffImageRaw += ffImageHSV;
      ffLTImageRaw += ffLTImageHSV;

      // TODO: do each part independently?
      cv::normalize( ffImageRaw, ffImageRaw, 0, 1, cv::NORM_MINMAX );
      ffImage += ffImageRaw;

      cv::normalize( ffLTImageRaw, ffLTImageRaw, 0, 1, cv::NORM_MINMAX );
      ffLTImage += ffLTImageRaw;

      cv::normalize( latImageRaw, latImageRaw, 0, 1, cv::NORM_MINMAX );
      cv::normalize( latImage, latImage, 0, 1, cv::NORM_MINMAX );
      latImage += latImageRaw;

      n->printConnections( "Incoming Column Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Column ); } );
      n->printConnections( "Incoming Feedforward(d) Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedforward | bd::ConnectionFlags::Direct ); }, verbose );
      n->printConnections( "Incoming Feedforward(s) Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedforward | bd::ConnectionFlags::Skip ); }, verbose );
      n->printConnections( "Incoming Lateral + Connections",   [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Lateral ) && in.weight >= 0.0; }, verbose );
      n->printConnections( "Incoming Lateral - Connections",   [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Lateral | bd::ConnectionFlags::Inhibitory ) ||
                                                                                                (in.hasFlag( bd::ConnectionFlags::Lateral ) && in.weight < 0.0); }, verbose );
      n->printConnections( "Incoming Feedback Connections(+d)", [](bd::Input const & in)
          { return in.hasFlag( bd::ConnectionFlags::Feedback | bd::ConnectionFlags::Direct | bd::ConnectionFlags::Excitatory ); }, verbose );
      n->printConnections( "Incoming Feedback Connections(-d)", [](bd::Input const & in)
          { return in.hasFlag( bd::ConnectionFlags::Feedback | bd::ConnectionFlags::Direct | bd::ConnectionFlags::Inhibitory ); }, verbose );

      static const std::string feedforwardName = "Incoming Feedforward Neuron Connections";
      static const std::string feedforwardLTName = "Incoming LT Feedforward Neuron Connections";
      static const std::string feedbackName = "Incoming Feedback Neuron Connections";
      static const std::string lateralName = "Incoming Lateral Neuron Connections ";


      // Mark centers gray
      cv::normalize( ffImage,  ffImage, 0, 1, cv::NORM_MINMAX );
      cv::normalize( ffLTImage,  ffLTImage, 0, 1, cv::NORM_MINMAX );
      cv::normalize( fbImage,  fbImage, 0, 1, cv::NORM_MINMAX );
      cv::normalize( latImage, latImage, 0, 1, cv::NORM_MINMAX );
      for( int c = 0; c < columnSize; ++c )
      {
        imageRef( myY, myX, ffImage, c ) += cv::Vec3f{0.5, 0.5, 0.5};
        imageRef( myY, myX, ffLTImage, c ) += cv::Vec3f{0.5, 0.5, 0.5};
        imageRef( myY, myX, fbImage, c ) += cv::Vec3f{0.5, 0.5, 0.5};
        imageRef( myY, myX, latImage, c ) += cv::Vec3f{0.5, 0.5, 0.5};
      }

      // Mark centroids white
      auto drawCentroid = [&](cv::Mat & image, const int flags)
      {
        double cX, cY; std::tie(cX, cY) = n->centroid(flags);
        cX *= scaling; cY *= scaling;

        if( cX < 0.0 || cY < 0.0 ) // do not draw invalid centroids
          return;

        for( int c = 0; c < columnSize; ++c )
          imageRef( cY, cX, image, c ) += cv::Vec3f{1.0, 1.0, 1.0};
      };

      drawCentroid( ffImage,  bd::ConnectionFlags::Feedforward );
      drawCentroid( fbImage,  bd::ConnectionFlags::Feedback );
      drawCentroid( latImage, bd::ConnectionFlags::Lateral );

      cv::namedWindow( feedforwardName, cv::WINDOW_NORMAL );
      cv::imshow( feedforwardName, ffImage );
      cv::imwrite( savePrefix + "feedforward.png", ffImage * 255.0 );

      cv::namedWindow( feedforwardLTName, cv::WINDOW_NORMAL );
      cv::imshow( feedforwardLTName, ffLTImage );
      cv::imwrite( savePrefix + "feedforward_lt.png", ffLTImage * 255.0 );

      cv::namedWindow( feedbackName, cv::WINDOW_NORMAL );
      cv::imshow( feedbackName, fbImage );
      cv::imwrite( savePrefix + "feedback.png", fbImage * 255.0 );

      {
        auto whiteBG = bd::convertToWhiteBackground( fbImage );
        cv::imwrite( savePrefix + "feedback_white.png", whiteBG * 255.0 );
      }

      cv::namedWindow( lateralName, cv::WINDOW_NORMAL );
      cv::imshow( lateralName, latImage );
      //bd::showImage( latImage, true, false, lateralName );
      cv::imwrite( savePrefix + "lateral.png", latImage * 255.0 );
    }
  }
} // end anon namespace

template <class NeuronT> inline
void bd::showConnections( Sheet<NeuronT> const & sheet, bool normalize, bool showPositions, std::string const & name )
{
  int maxPerGrid = 0;
  for( auto const & cell : sheet )
    maxPerGrid = std::max( maxPerGrid, (int)cell.size() );
  const double scaling = maxPerGrid;

  std::string sheetWindow = showImage( showPositions ? sheetNeuronImage( sheet, scaling ) :
                                                       sheetActivationImage( sheet, scaling ),
                                       normalize, false, name );
  cv::setMouseCallback( sheetWindow, &display_impl::windowCallback<NeuronT>, const_cast<Sheet<NeuronT>*>(&sheet) );
  //cv::waitKey(0);
}

#endif // DISPLAYIMPL_H_
