#include "EdgeNeuron.H"
#include "Random.H"

#include <cereal/types/polymorphic.hpp>

#include <opencv2/highgui/highgui.hpp>

CEREAL_REGISTER_TYPE(bd::EdgeNeuron)

bd::EdgeNeuron::EdgeNeuron( ) :
  Neuron( ),
  itsFFNormalizationCoefficient(1.0),
  itsLateralNormalizationCoefficient(1.0)
{
}

bd::EdgeNeuron::EdgeNeuron( std::pair<double, double> pos ) :
  Neuron( pos ),
  itsFFNormalizationCoefficient(1.0),
  itsLateralNormalizationCoefficient(1.0)
{
}

bd::EdgeNeuron::EdgeNeuron( EdgeNeuron const & other, std::pair<double, double> pos ) :
  Neuron( other, pos ),
  itsFFNormalizationCoefficient( other.itsFFNormalizationCoefficient ),
  itsLateralNormalizationCoefficient( other.itsLateralNormalizationCoefficient )
{
}

double bd::EdgeNeuron::updateImpl( SystemState const & state )
{
  const double noise = normalReal()() * state.noiseScale;

  double input = 0.0;

  double feedback = 0, feedforward = 0;
  double feedbackAverage = 0.0;
  double maxFeedback = 0;
  double lateralExcitation = 0;
  int feedbackCount = 0;

  double totalLearnedWeight = 0.0;

  auto const & preInput = itsInput.pre();

  // Excitation
  for( auto const & preEdge : preInput )
  {
    if( preEdge.weight > 0 )
    {
      // skip all silent connections
      if( preEdge.hasFlag( ConnectionFlags::Silent ) )
        continue;

      if( preEdge.hasFlag( ConnectionFlags::Lateral ) && !state.useLateralConnections )
        continue;

      const double currentInput = preEdge.weight * preEdge.neuron->activation();

      if( preEdge.hasFlag( ConnectionFlags::Feedforward ) )
      {
        feedforward += currentInput;
        totalLearnedWeight += preEdge.ltWeight;
      }
      else if( preEdge.hasFlag( ConnectionFlags::Feedback ) )
      {
        feedback += currentInput;
        maxFeedback = std::max( maxFeedback, currentInput );

        if( preEdge.weight > preEdge.initialWeight )
        {
          feedbackAverage += preEdge.weight;
          ++feedbackCount;
        }
      }
      else if( preEdge.hasFlag( ConnectionFlags::Lateral ) )
      {
        lateralExcitation += currentInput;
      }
    }
  }

  feedbackAverage /= feedbackCount;
  if( feedbackCount == 0 )
    feedbackAverage = 0.0;

  input += feedforward;
  input += maxFeedback * 2.5 * feedforward * feedforward; // note: FF modulates FB
  input += lateralExcitation;
  input += noise;

  // Inhibition
  itsInhibitorySum = calculateInhibitorySum( state, itsInhibitorySum, ConnectionFlags::Column );

  // current activation
  double preClamp = std::max( input / (itsInhibitorySum + 1.0), 0.0 );

  // DEBUG DISABLE
//  //const double averageSmoothing = 0.01;
//  const double averageSmoothing = 0.01;
//
//  double newThreshold = itsLTThreshold.pre();
//
//  if( state.learningRate != 0.0 )
//  {
//    const double onThreshold = state.noiseScale * 4.0;
//    if( feedforward > onThreshold )
//      ++itsNumOnIterations;
//    else
//      itsNumOnIterations = 0;
//
//    if( itsInhibitorySum < 0.2 && feedforward > 0.3 ) // may need hebb with original .7
//    {
//      ++itsNumOveractiveIterations;
//      if( itsNumOveractiveIterations > 1 )
//        newThreshold = averageSmoothing * feedforward * 0.7 + (1.0 - averageSmoothing) * itsLTThreshold.pre();
//    }
//    else
//      itsNumOveractiveIterations = 0;
//  }
//
////  if( itsRollingAverage > 0.7 ) // prinnnt
////  {
////    printNeuronInfo();
////    printConnections( "Incoming Feedforward Connections", [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedforward ); } );
////    printConnections( "Incoming Lateral + Connections",   [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Lateral ) && in.weight >= 0.0; } );
////    printConnections( "Incoming Lateral - Connections",   [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Lateral | bd::ConnectionFlags::Inhibitory ); } );
////    printConnections( "Incoming Feedback Connections",    [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedback ); } );
////    cv::waitKey(0);
////  }
//
//  // Apply threshold
//  if( preClamp < itsLTThreshold.pre() )
//    preClamp = 0.0;

  const double value = clamp( preClamp, 0.0, state.maxActivation );

  if( state.learningRate == 0.0 ) // If learning is off, don't adjust thresholds
    return value;

  // DEBUG DISABLE
//  itsLTThreshold.post() = newThreshold;

  return value;
}

void bd::EdgeNeuron::learnImpl( SystemState const & state, double currentActivation )
{
  if( state.learningRule == LearningRule::Hebbian )
  {
    shortTermHebbianNormalization( state.v1MaxFFWeight * itsFFNormalizationCoefficient,
                                   currentActivation, state,
                                   ConnectionFlags::Dynamic | ConnectionFlags::Feedforward );
    return;
  }
  else if( state.learningRule == LearningRule::Conflict )
  {
    if( currentActivation == 0.0 ) // learning can't happen if we aren't on at all
    {
      propogateAccumulatedWeight( state );
      return;
    }

    const double learningRate = state.learningRate;

    auto const & preInput = itsInput.pre();
    auto & postInput = itsInput.post();

    auto computeMaxLearnedActivation = [&](ConnectionFlags type)
    {
      double maxWeight = -std::numeric_limits<double>::max();

      for( auto const & preEdge : preInput )
      {
        if( preEdge.hasFlag( type ) && preEdge.hasFlag( ConnectionFlags::Dynamic ) )
          maxWeight = std::max( maxWeight, preEdge.ltWeight );
      }

      const double weightCutoff = 0.5 * maxWeight;

      double maxLearnedActivation = 0.0;

      // Find maximum weight over cutoff
      for( auto const & preEdge : preInput )
      {
        if( preEdge.hasFlag( type ) && preEdge.hasFlag( ConnectionFlags::Dynamic ) )
        {
          if( preEdge.weight > weightCutoff )
            maxLearnedActivation = std::max( maxLearnedActivation, preEdge.neuron->activation() );
        }
      }

      return maxLearnedActivation;
    };

    const double ffMaxLearnedActivation = state.cl_spreading ? computeMaxLearnedActivation( ConnectionFlags::Feedforward ) : 1.0;
    //const double maximumFeedforwardTotalWeight = state.v1MaxFFWeight * itsFFNormalizationCoefficient;

    // Sum all dynamic edges
    for( size_t i = 0; i < preInput.size(); ++i )
    {
      auto const & preEdge = preInput[i];
      auto & postEdge = postInput[i];

      const double snarf = 1.0;

      if( preEdge.hasFlag( ConnectionFlags::Dynamic | ConnectionFlags::Feedforward ) )
      {
        const double inputActivation = preEdge.neuron->activation();

        const double lowerBound = 0.0;
        const double upperBound = state.maxWeight;

        const double targetWeight = clamp(inputActivation * currentActivation);

        double ltmUpdate = 0.0;

        // Inhibition
        if( state.cl_unlearning )
        {
          // Short term weight delta
          const double inhibitoryModulation = clamp( itsInhibitorySum );
          const double beta = 1.0 * snarf;
          const double update = beta * learningRate * inhibitoryModulation * targetWeight;// * ffMaxLearnedActivation;

          postEdge.weight = clamp( postEdge.weight - update, lowerBound, upperBound );
          ltmUpdate -= update;
        }

        // Excitation
        {
          // Short term weight delta
          const double inhibitoryModulation = state.cl_unlearning ? (1.0 - clamp( itsInhibitorySum )) : 1.0;
          const double alpha = 1.0 * snarf; // alpha is used to balance learning vs unlearning rate (learning should never be faster than unlearning)
          const double update = alpha * learningRate * inhibitoryModulation * targetWeight * ffMaxLearnedActivation;

          postEdge.weight = clamp( postEdge.weight + update, lowerBound, upperBound );
          ltmUpdate += update;
        }

        // Force unlearning to take us towards 'unlearned' (original) state
        //  Unlearning lowers weight until we go below 'unlearned' value, at which point
        //  it increases.  Unlearning always moves us towards 'unlearned' value.
        if( postEdge.weight < preEdge.weight )
        {
          const double unlearningModulator = 1.00; // how fast unlearning is relative to learning
          const double difference = (preEdge.weight - postEdge.weight) * unlearningModulator;

          if( preEdge.weight >= preEdge.initialWeight )
            postEdge.weight = clamp( preEdge.weight - difference, preEdge.initialWeight, upperBound );
          else
            postEdge.weight = clamp( preEdge.weight + difference, lowerBound, preEdge.initialWeight );
        }

        // Accumulator adjustment
        const double ltmAlpha = 100.0; // todo: vestigial from when we used decay
        const double difference = std::abs(postEdge.ltWeight - postEdge.weight) > 0.01 ? 1.0 : 0.0;
        postEdge.accumulatedWeight = std::max( preEdge.accumulatedWeight + ltmAlpha * ltmUpdate * difference, 0.0 );
      }
    } // end for

    propogateAccumulatedWeight( state );

    hebbianNormalization( state.v1lateralTotalWeight,
                          currentActivation,
                          state,
                          ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );
  } // end NOT use hebbian learning
}

void bd::EdgeNeuron::propogateAccumulatedWeight( SystemState const & state )
{
  /**
   * Long term weights are closely related to the distribution of the accumulators.
   *
   * At each iteration, the long term weights will move towards their current short term weights
   * at a rate dependent upon their accumulator and short term weights.
   *
   * At the same time, the short term weights constantly decay towards their long term weight
   * (increasing or decreasing).
   *
   * If the long term weight sum exceeds some maximum, it will be normalized.
   *
   * All weights are restricted to be in the range [0, 1]
   */

  double accumulatorSum = 0.0;
  double longTermSum = 0.0;
  int numConnections = 0;

  auto & postInput = itsInput.post();

  for( auto const & postEdge : postInput )
  {
    if( postEdge.hasFlag( ConnectionFlags::Dynamic ) && postEdge.hasFlag( ConnectionFlags::Feedforward ) )
    {
      accumulatorSum += postEdge.accumulatedWeight;
      longTermSum += postEdge.ltWeight;
      ++numConnections;
    }
  }

  // Increment each long term weight based upon the current distribution of the accumulators
  //  and normalize inhibited weights by the total accumulator sum
  double postWeightSum = 0.0;

  for( auto & postEdge : postInput )
  {
    if( postEdge.hasFlag( ConnectionFlags::Dynamic ) && postEdge.hasFlag( ConnectionFlags::Feedforward ) )
    {
      {
        const double accDistribution = accumulatorSum ? postEdge.accumulatedWeight / accumulatorSum : 0.0;
        const double ltDistribution  = longTermSum    ? postEdge.ltWeight / longTermSum             : 0.0;

        // how far we are from the long term distribution, ranges [0, 1.0]
        const double difference = std::abs(ltDistribution - accDistribution );

        const double goal = postEdge.weight; // new target for long term weight

        const bool growUp   = (goal - postEdge.ltWeight) >= 0; // True if weight will increase
        const bool desireUp = (accDistribution - ltDistribution ) >= 0; // True if accumulator is higher than LT

        // How quickly we move towards our goal (multiplier), ranges from [1.0 to 20.0]
        //  in practice we're usually off by around 5% distribution maximum, so we'll scale to this
        const double convergenceRate = 1.0 + difference * 199;

        // Increasing this value makes us favor short term weights; decreasing it long term weights
        const double baseRate = 0.05; // base rate of change given no accumulator
        //const double baseRate = 0.0001; // base rate of change given no accumulator
        //const double proportion = clamp( baseRate * ( (growUp ^ desireUp) ? 1.0 / convergenceRate : convergenceRate ) );
        const double proportion = clamp( baseRate * ( (growUp ^ desireUp) ? 1.0 / 20.0 : convergenceRate ) );

        postEdge.ltWeight = proportion * goal + (1.0 - proportion) * postEdge.ltWeight; // move towards short term weight
      }

      postWeightSum += postEdge.ltWeight;
    }
  } // end for

  // Normalization ensures that our post-shaped distribution still sums up to its original values, or
  // the maximum if we exceeded that
  const double maximumFeedforwardTotalWeight = state.v1MaxFFWeight * itsFFNormalizationCoefficient;
  {
    const double weightSumTarget = std::min( postWeightSum, maximumFeedforwardTotalWeight );
    const double normalizer = weightSumTarget / postWeightSum;

    for( auto & postEdge : postInput )
    {
      if( postEdge.hasFlag( ConnectionFlags::Dynamic ) && postEdge.hasFlag( ConnectionFlags::Feedforward ) )
        postEdge.ltWeight = clamp( postEdge.ltWeight * normalizer, 0.0, state.maxWeight );
    }
  }

  // Move short term towards long term
  //  larger values make us follow long term weight more strongly; lower values short term weights
  const double chaseFactor = 0.05; // .05 normally, .1 thought
  double postWeightSumST = 0.0;
  for( auto & postEdge : postInput )
    if( postEdge.hasFlag( ConnectionFlags::Dynamic ) && postEdge.hasFlag( ConnectionFlags::Feedforward ) )
    {
      if( state.cl_slt )
        postEdge.weight = (1 - chaseFactor) * postEdge.weight + chaseFactor * postEdge.ltWeight;
      postWeightSumST += postEdge.weight;
    }

  // Normalization on the short term
  {
    const double weightSumTarget = std::min( postWeightSumST, maximumFeedforwardTotalWeight );
    const double normalizer = weightSumTarget / postWeightSumST;

    for( auto & postEdge : postInput )
    {
      if( postEdge.hasFlag( ConnectionFlags::Dynamic ) && postEdge.hasFlag( ConnectionFlags::Feedforward ) )
        postEdge.weight = clamp( postEdge.weight * normalizer, 0.0, state.maxWeight );
    }
  } // end ST normalization
}
