#include "Gabor.H"
#include "Util.H"
#include "ThreadingUtil.H"

#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>

#include <omp.h>

namespace
{
  bool isSame( cv::Mat const & a, cv::Mat const & b )
  {
    if( a.size() == b.size() &&
        a.depth() == b.depth() &&
        a.data == b.data )
      return true;
    else
      return false;
  }
}

bd::Gabor::Gabor() : itsValid( false ), itsPrevImage()
{
}

bd::Gabor::Gabor( int numScales, float scaleFirst, float scaleStep, int numOrientations, float sigma, float freq, float resolutionScaling ) :
  itsNumOrientations( numOrientations ),
  itsNumScales( numScales ),
  itsValid( true ), itsLargestFilter( 0 ), itsPaddedSize(), itsPrevImage()
{
  // set up threads
  if( fftwf_init_threads() == 0 )
    throw("FFT init failure");

  fftwf_plan_with_nthreads( threading::nthreads );

  // allocate space
  itsFiltersSpatial.resize( numOrientations );
  itsFiltersFrequency.resize( numOrientations );
  itsResponseRealCache.resize( numOrientations );
  itsResponseImagCache.resize( numOrientations );

  for( int i = 0; i < numOrientations; ++i )
  {
  	itsFiltersSpatial[i].resize( numScales );
  	itsFiltersFrequency[i].resize( numScales );
    itsResponseRealCache[i].resize( numScales );
    itsResponseImagCache[i].resize( numScales );
  }

  // fill out filters
  for( int o = 0; o < numOrientations; ++o )
  {
    const float theta = o * ( M_PI / static_cast<float>( numOrientations ) );

    float sF = scaleFirst;
    for( int s = 0; s < numScales; ++s, sF += scaleStep )
    {
      const float K = M_PI / 2.0f / std::pow( freq, sF ) / resolutionScaling;

      float width = std::round( sigma / K * 8 + 1 );
      if( std::fmod( width, 2.0f ) == 0 )
        ++width;

      const int size = static_cast<int>( width );
			itsLargestFilter = std::max( itsLargestFilter, size );

      cv::Mat filterReal( size, size, CV_32FC1 );
      cv::Mat filterImag( size, size, CV_32FC1 );

      for( int i = 0; i < size; ++i )
        for( int j = 0; j < size; ++j )
        {
          const int x = i - ( size - 1 ) / 2;
          const int y = j - ( size - 1 ) / 2;

          const float common = ( std::pow( K, 2.0f ) / std::pow( sigma, 2.0f ) ) *
                                 std::exp( -( std::pow( static_cast<float>( x ), 2.0f ) +
                                              std::pow( static_cast<float>( y ), 2.0f ) ) *
                                           std::pow( K, 2.0f ) / ( 2.0f * std::pow( sigma, 2.0f ) ) );

          const float real = std::cos( K * std::cos( theta ) * x + K * std::sin( theta ) * y ) - std::exp( - std::pow( sigma, 2.0f ) / 2.0f );
          const float imag = std::sin( K * std::cos( theta ) * x + K * std::sin( theta ) * y );

					filterReal.at<float>( i, j ) = common * real;
          filterImag.at<float>( i, j ) = common * imag;
        }

			itsFiltersSpatial[o][s] = cv::Mat( size, size, CV_32FC2 );
      cv::Mat merge[] = { filterReal, filterImag };
      cv::merge( merge, 2, itsFiltersSpatial[o][s] );

      // DEBUG viewing
      //bd::showImage( filterReal, true, true, "real before" + std::to_string(o) );
      //bd::showImage( filterImag, true, true, "imag before " + std::to_string(o) );
    }
  }
}

bd::Gabor::~Gabor()
{
  fftwf_cleanup_threads();
}

std::vector<std::vector<cv::Mat>> bd::Gabor::getFilterResponses( cv::Mat const & image, ResponseType rt )
{
  if( !itsValid )
    return std::vector<std::vector<cv::Mat>>();

  bool useCache = false;

  if( ::isSame( image, itsPrevImage ) )
  {
    useCache = itsUseCache;
  }
  else
  {
    itsPrevImage = image;
    itsUseCache = false;
    auto prevPaddedSize = itsPaddedSize;

    // FFT stuff
		cv::Size biggestFilter( itsLargestFilter, itsLargestFilter );
		itsPaddedSize = getPaddedSize( image.size(), biggestFilter );
		itsPrevImageFFT = padImageFFT( image, biggestFilter, itsPaddedSize );
		//cv::dft( itsPrevImageFFT, itsPrevImageFFT );
    FFTW_DFT( itsPrevImageFFT );

    //{ // DEBUG
		//		std::vector<cv::Mat> realImag(2);
		//		cv::split( itsPrevImageFFT, realImag );
    //    bd::showImage( realImag[0], true, true, "real padded" );
    //} // end DEBUG

    if( prevPaddedSize != itsPaddedSize ) // padded size change, pad the filters
    {
      for( size_t o = 0; o < itsFiltersSpatial.size(); ++o )
        for( size_t s = 0; s < itsFiltersSpatial[0].size(); ++s )
        {
          itsFiltersFrequency[o][s] = padKernelFFT( itsFiltersSpatial[o][s], itsPaddedSize );
          //cv::dft( itsFiltersFrequency[o][s], itsFiltersFrequency[o][s] );
          FFTW_DFT( itsFiltersFrequency[o][s] );

          //{ // DEBUG
          //  cv::Mat temp;
          //  cv::idft( itsFiltersFrequency[o][s], temp );

          //  std::vector<cv::Mat> realImag(2);
          //  cv::split( temp, realImag );

          //  fftshift(realImag[0]);
          //  fftshift(realImag[1]);
          //  bd::showImage( realImag[0], true, true, "real after" + std::to_string(o) );
          //  bd::showImage( realImag[1], true, true, "imag after " + std::to_string(o) );
          //} // end DEBUG
        }
    } // end pad filters
  }

  const int numOrientations = itsFiltersFrequency.size();
  const int numScales = itsFiltersFrequency[0].size();

  // set up result - outer vector over orientations, inner over scale
  std::vector<std::vector<cv::Mat>> result( numOrientations );
	for( auto & i : result )
    i.resize( numScales );

	// outer loop for orientations
  for( int o = 0; o < numOrientations; ++o )
    for( int s = 0; s < numScales; ++s )
    {
			cv::Mat temp;

			if( !useCache )
			{
				cv::mulSpectrums( itsPrevImageFFT, itsFiltersFrequency[o][s], temp, cv::DFT_COMPLEX_OUTPUT );
				//cv::idft( temp, temp );
        FFTW_IDFT( temp );

				std::vector<cv::Mat> realImag(2);
				cv::split( temp, realImag );

				itsResponseRealCache[o][s] = uncropFFT( realImag[0], image.size() );
				itsResponseImagCache[o][s] = uncropFFT( realImag[1], image.size() );
				itsUseCache = true;
			}

      switch( rt )
      {
        case ResponseType::real:
          result[o][s] = itsResponseRealCache[o][s];
          break;
        case ResponseType::imag:
          result[o][s] = itsResponseImagCache[o][s];
          break;
        case ResponseType::magnitude:
          cv::magnitude( itsResponseRealCache[o][s], itsResponseImagCache[o][s], result[o][s] );
          break;
        break;
      }

			cv::normalize( result[o][s], result[o][s], 0, 1, cv::NORM_MINMAX );
    }

  return result;
}


void bd::Gabor::FFTW_DFT( cv::Mat & image )
{
  fftwf_plan plan( fftwf_plan_dft_1d( image.rows * image.cols,
                   reinterpret_cast<fftwf_complex*>( image.data ),
                   reinterpret_cast<fftwf_complex*>( image.data ), FFTW_FORWARD, FFTW_ESTIMATE ) );
  fftwf_execute( plan );
  fftwf_destroy_plan( plan );
}

void bd::Gabor::FFTW_IDFT( cv::Mat & image )
{
  fftwf_plan plan( fftwf_plan_dft_1d( image.rows * image.cols,
                   reinterpret_cast<fftwf_complex*>( image.data ),
                   reinterpret_cast<fftwf_complex*>( image.data ), FFTW_BACKWARD, FFTW_ESTIMATE ) );
  fftwf_execute( plan );
  fftwf_destroy_plan( plan );
}
