#ifndef GABOR_H_
#define GABOR_H_

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <cmath> // for M_PI
#include <fftw3.h>
#include "Display.H"

namespace bd
{
	//! A class for creating a bank of gabor filters
  class Gabor
  {
    public:
			enum class ResponseType
			{
				real,
				imag,
				magnitude
			};

      //! Create an empty gabor object - can't be used for anything
			Gabor();

      //! Prepares for log gabor filter creation
      /*! Does some initial calculations involved in creating Log Gabor filters
          for an image of a given size
					@param[in] numScales The number of scales to use
					@param[in] scaleFirst The scale of the first filter
					@param[in] scaleStep the step between scales
					@param[in] numOrientations The number of orientations to use (evenly distributed)
					@param[in] sigma The variance of the gabor filter
					@param[in] freq The spatial frequency of the gabor filter
          @param[in] resolutionScaling Adjusts the pixel density of the gabor representation */
			Gabor( int numScales, float scaleFirst, float scaleStep, int numOrientations, float sigma = 1 * M_PI, float freq = std::sqrt( 2.0f ),
             float resolutionScaling = 1.0f );

      ~Gabor();

			//! Gets the response of the filter bank to an image
			/*! @param[in] image the input image
			 	  @param[in] rt The response type desired
					@return for each orientation, a vector of scale responses of rt */
			std::vector<std::vector<cv::Mat>> getFilterResponses( cv::Mat const & image, ResponseType rt = ResponseType::magnitude );

      size_t itsNumOrientations;
      size_t itsNumScales;

    protected:
      void FFTW_DFT( cv::Mat & image );
      void FFTW_IDFT( cv::Mat & image );

			bool itsValid;

			int itsLargestFilter;
			cv::Size itsPaddedSize;

			std::vector<std::vector<cv::Mat>> itsFiltersSpatial;
			std::vector<std::vector<cv::Mat>> itsFiltersFrequency;

			std::vector<std::vector<cv::Mat>> itsResponseRealCache;
			std::vector<std::vector<cv::Mat>> itsResponseImagCache;

			cv::Mat itsPrevImage;
			cv::Mat itsPrevImageFFT;

			bool itsUseCache;
  };
}

#endif // GABOR_H_
