#include "Generate.H"
#include "V1.H"
#include "Combined.H"

namespace
{
  inline int toInt( double d )
  {
    return static_cast<int>( std::round( d ) );
  }

  // Rotate a point about some centroid by some number of radians
  cv::Point2d rotatePoint( cv::Point2d const & p, cv::Point2d const & centroid, double radians )
  {
    const auto s = std::sin(radians);
    const auto c = std::cos(radians);

    const double xx = p.x - centroid.x;
    const double yy = p.y - centroid.y;

    const double xnew = xx * c - yy * s + centroid.x;
    const double ynew = xx * s + yy * c + centroid.y;

    return {xnew, ynew};
  }

  // Returns the perpendicular angle going clockwise, from the given vector,
  // adjusted to fit color maps from the paper
  double perpendicularAngle( cv::Point2d const & a, cv::Point2d const & b )
  {
    const double x = b.x - a.x;
    const double y = b.y - a.y;

    return std::atan2( -y, x );
  }
}

bd::Shape::Shape( std::vector<cv::Point2d> && v, cv::Point2d && c, bool hasGroundTruth ) :
  vertices( std::move( v ) ),
  centroid( std::move( c ) ),
  groundTruth()
{
  if( hasGroundTruth )
  {
    groundTruth.reserve( vertices.size() );

    for( size_t i = 0; i < vertices.size(); ++i )
    {
      const size_t next = i == vertices.size() - 1 ? 0 : i + 1;
      groundTruth.push_back( perpendicularAngle( vertices[i], vertices[next] ) );
    }
  }
}

template <class Network>
bool bd::isValid( Network const & network, Shape const & shape, float resolutionScaling )
{
  const int width  = network.itsGridWidth * resolutionScaling;
  const int height = network.itsGridHeight * resolutionScaling;

  auto isLocValid = [&]( cv::Point const & p )
  {
    if( p.x < 0 || p.x >= width )
      return false;

    if( p.y < 0 || p.y >= height )
      return false;

    return true;
  };

  for( const auto & vertex : shape.vertices )
    if( isLocValid( vertex ) )
      return true;

  return false;
}

template bool bd::isValid<bd::Border>( bd::Border const &, Shape const &, float );
template bool bd::isValid<bd::V1>( bd::V1 const &, Shape const &, float );
template bool bd::isValid<bd::Combined>( bd::Combined const &, Shape const &, float );

bool bd::isValid( int width, int height, Shape const & shape )
{
  auto isLocValid = [&]( cv::Point const & p )
  {
    if( p.x < 0 || p.x >= width )
      return false;

    if( p.y < 0 || p.y >= height )
      return false;

    return true;
  };

  for( const auto & vertex : shape.vertices )
    if( isLocValid( vertex ) )
      return true;

  return false;
}

cv::Mat bd::generateRectangle( int imgH, int imgW,
    int topLeftX, int topLeftY,
    int width, int height, double angle )
{
  cv::Mat image = cv::Mat::zeros( imgH, imgW, CV_32FC1 );
  drawRectangle( image, topLeftX, topLeftY, width, height );
  return image;
}

void bd::drawRectangle( cv::Mat & image,
    int topLeftX, int topLeftY,
    int width, int height, double angle )
{
  cv::rectangle( image, cv::Point( topLeftX, topLeftY ),
                 cv::Point( topLeftX + width - 1, topLeftY + height - 1 ),
                 1.0, // color
                 1 // thickness
      );
}

bd::Shape bd::generateLine( int topLeftX, int topLeftY, int length, bool groundTruth )
{
  return {{{(double)topLeftX, (double)topLeftY},
           {(double)topLeftX + length, (double)topLeftY}},
          {topLeftX + length / 2.0 + 0.5, (double)topLeftY},
          groundTruth};
}

bd::Shape bd::generateLine( int startX, int startY, int endX, int endY, bool groundTruth )
{
  return {{{(double)startX, (double)startY},
           {(double)endX, (double)endY}},
          {(startX + endX) / 2.0 + 0.5, (startY + endY) / 2.0 + 0.5},
          groundTruth};
}

bd::Shape bd::generateRectangle( int topLeftX, int topLeftY,
                                 int width, int height,
                                 double scale )
{
  const double w = width * scale;
  const double h = height * scale;

  return {{{(double)topLeftX,   (double)topLeftY},
           {topLeftX + w - 1.0, (double)topLeftY},
           {topLeftX + w - 1.0, topLeftY + h - 1.0},
           {(double)topLeftX,   topLeftY + h - 1.0}},
          {topLeftX + w / 2.0 - 0.5, topLeftY + h / 2.0 - 0.5}};
}

bd::Shape bd::generateEquilateralTriangle( int centerX, int centerY,
    int radius, double scale )
{
  const double r = radius * scale;
  static const double inv_sqrt3 = 1.0 / std::sqrt( 3.0 );
  const double length = 6 * r * inv_sqrt3; // length of triangle side
  const double half_length = length / 2.0;

  // (0, 0) is upper left corner
  return { {{centerX - half_length, (double)centerY + r},
            {(double)centerX,       centerY - length * inv_sqrt3},
            {centerX + half_length, (double)centerY + r}},
           {static_cast<double>(centerX), static_cast<double>(centerY)} };
}

bd::Shape bd::rotateShape( Shape const & shape, double radians )
{
  bd::Shape result = shape;
  for( auto & vertex : result.vertices )
    vertex = rotatePoint( vertex, shape.centroid, radians );

  for( auto & angle : result.groundTruth )
    angle = std::fmod( angle - radians + 2.0 * M_PI, 2.0 * M_PI );

  return result;
}

bd::Shape bd::translateShape( Shape const & shape, cv::Point2d const & translation )
{
  bd::Shape result = shape;

  for( auto & vertex : result.vertices )
    vertex += translation;

  result.centroid += translation;
  return result;
}

bd::Shape bd::scaleShape( Shape const & shape, double scale )
{
  // We'll do the scaling by moving all points towards the centroid (or away from it)
  bd::Shape result = shape;
  const auto centroid = result.centroid;

  for( auto & vertex : result.vertices )
  {
    const auto v_c = vertex - centroid;
    vertex = centroid + v_c * scale;
  }

  return result;
}

void bd::drawShape( cv::Mat & image, bd::Shape const & shape, double color, bool fill, int edgeWidth, int edgeType )
{
  std::vector<cv::Point> points(shape.vertices.size());
  std::transform( shape.vertices.begin(), shape.vertices.end(),
                  points.begin(),
                  [](cv::Point2d const & p)
                  {
                    return cv::Point( toInt(p.x), toInt(p.y) );
                  } );

  if( shape.vertices.size() == 2 ) // point
    cv::line( image, points[0], points[1], color, edgeWidth, edgeType );
  else
  {
    if( fill )
      cv::fillConvexPoly( image, points, color, edgeType );
    else
      cv::polylines( image, points, true, color, edgeWidth, edgeType );
  }
}

void bd::drawShapesInterpolated( cv::Mat & image, std::vector<bd::ShapeMetaData> const & shapes )
{
  cv::Mat stimulus8U = cv::Mat::zeros( image.rows, image.cols, CV_8UC1 );

  for( const auto & s : shapes )
    bd::drawShape( stimulus8U, s.shape, s.color * 255.0, s.fill, s.edgeWidth, CV_AA );

  stimulus8U.convertTo( image, CV_32FC1 );
  image /= 255.0;
}

cv::Mat bd::drawGroundTruth( cv::Mat & image, std::vector<bd::ShapeMetaData> const & shapes )
{
  cv::Mat canvas8U   = cv::Mat::zeros( image.rows, image.cols, CV_8UC1 );
  cv::Mat value      = cv::Mat::zeros( image.rows, image.cols, CV_32FC1 );
  cv::Mat sinAngle   = cv::Mat::zeros( image.rows, image.cols, CV_32FC1 );
  cv::Mat cosAngle   = cv::Mat::zeros( image.rows, image.cols, CV_32FC1 );

  // Draw line from vertex i to i+1, wrapping around, with ground truth for coloring
  for( const auto & s : shapes )
  {
    // For lines, only draw a single line instead of wrapping around
    const size_t iterCount = s.shape.vertices.size() == 2 ? 1 : s.shape.vertices.size();
    for( size_t i = 0; i < iterCount; ++i )
    {
      const size_t next = i == s.shape.vertices.size() - 1 ? 0 : i + 1;
      const double angle = s.shape.groundTruth[i];

      cv::Mat local8U = cv::Mat::zeros( image.rows, image.cols, CV_8UC1 );
      cv::line( local8U, s.shape.vertices[i], s.shape.vertices[next], 255, s.edgeWidth, CV_AA );
      cv::line( canvas8U, s.shape.vertices[i], s.shape.vertices[next], 255, s.edgeWidth, CV_AA );

      cv::Mat local32F = cv::Mat::zeros( image.rows, image.cols, CV_32FC1 );

      // value
      local8U.convertTo( local32F, CV_32FC1 );
      value += local32F;

      // calculate hue components
      cv::Mat localBinary = local8U;
      for( auto iter = localBinary.begin<unsigned char>(), end = localBinary.end<unsigned char>(); iter != end; ++iter )
        *iter = (bool)*iter;

      // Y component
      localBinary.convertTo( local32F, CV_32FC1 );
      local32F *= std::sin(angle);
      sinAngle += local32F;

      // X component
      localBinary.convertTo( local32F, CV_32FC1 );
      local32F *= std::cos(angle);
      cosAngle += local32F;
    }
  }

  // value
  canvas8U.convertTo( value, CV_32FC1 );

  // Create map of all ground truth angles, in degrees, and store in input image
  cv::Mat unused = cv::Mat::zeros( image.rows, image.cols, CV_32FC1 );
  cv::cartToPolar( cosAngle, -sinAngle, unused, image, true );

  // Convert saturation and angles to HSV image
  cv::Mat colorImage = cv::Mat::zeros( image.rows, image.cols, CV_32FC3 );

  std::vector<cv::Mat> channels;
  channels.push_back( image );
  channels.push_back( cv::Mat::ones( image.rows, image.cols, CV_32FC1 ) );
  channels.push_back( value );
  cv::merge( channels, colorImage );

  // Convert HSV to RGB for final image
  cv::cvtColor( colorImage, colorImage, CV_HSV2BGR );
  cv::normalize( colorImage, colorImage, 0, 1, cv::NORM_MINMAX );

  return colorImage;
}
