#ifndef GENERATE_H_
#define GENERATE_H_

#include <opencv2/core/core.hpp>
#include "Border.H"

//! User defined literal that converts to radians
inline double operator"" _deg( long double degrees )
{ return degrees * M_PI / 180.0; }

//! User defined literal that converts to degrees
inline double rad2deg( long double radians )
{ return radians * 180.0 / M_PI; }

namespace bd
{
  //! Represents a shape to draw
  /*! Shapes are represented by an ordered list of vertices and a centroid.

      The groundTruth holds the angle, in radians, that points towards the interior of the
      object for the side created between vertex[i] and vertex[i+1]. If the object has
      no defined interior, this is empty. */
  struct Shape
  {
    std::vector<cv::Point2d> vertices;
    cv::Point2d centroid;

    std::vector<double> groundTruth;

    Shape() = default;
    Shape( std::vector<cv::Point2d> && v, cv::Point2d && c, bool hasGroundTruth = true );
  };

  //! Checks if a shape can validly fit inside of an environment
  /*! Fit inside means at least one pixel of the shape is in the environment */
  template <class Network>
  bool isValid( Network const & network, Shape const & shape, float resolutionScaling = 1.0f );
  bool isValid( int width, int height, Shape const & shape );

  //! Draw a shape onto an image
  /*! Note that depending on the type of the image, interpolation may not be possible.  Use Interpolated version to
      force interpolation */
  void drawShape( cv::Mat & image, Shape const & shape, double color = 1.0, bool fill = false, int edgeWidth = 1, int edgeType = 4 );

  //! Metadata used for drawing a list of shapes
  struct ShapeMetaData
  {
    Shape shape;
    double color;
    bool fill;
    int edgeWidth;
  };

  //! Draw a shape onto an image with proper interpolation
  void drawShapesInterpolated( cv::Mat & image, std::vector<ShapeMetaData> const & shapes );

  //! Generates a horizontal line with some length
  Shape generateLine( int topLeftX, int topLeftY, int length, bool groundTruth = false );

  //! Generate a line between two points
  Shape generateLine( int startX, int startY, int endX, int endY, bool groundTruth = false );

  //! Generates a rectangle
  Shape generateRectangle( int topLeftX, int topLeftY,
                           int width, int height,
                           double scale = 1.0 );

  //! Generate an equilateral triangle centered at the provided location
  //! such that a circle could be inscribed within it
  Shape generateEquilateralTriangle( int centerX, int centerY,
                                     int radius,
                                     double scale = 1.0 );

  //! Rotates a shape about its centroid clockwise
  Shape rotateShape( Shape const & shape, double radians );

  //! Translates a shape
  Shape translateShape( Shape const & shape, cv::Point2d const & translation );

  //! Scales a shape while keeping centroid constant
  /*! This scales by adjusting the distance from each vertex to the centroid */
  Shape scaleShape( Shape const & shape, double scale );

  //! Generate a rectangle exactly on the parameter location
  /*! All rectangles have 1 pixel stroke widths */
  cv::Mat generateRectangle( int imgH, int imgW,
                             int topLeftX, int topLeftY,
                             int width, int height,
                             double angle = 0.0 );

  //! Draw a rectangle on the parameter image as above
  /*! All rectangles have 1 pixel stroke widths */
  void drawRectangle( cv::Mat & image,
                      int topLeftX, int topLeftY,
                      int width, int height,
                      double angle = 0.0 );

  //! Draw the ground truth (in radians) into the image
  /*! Returns an RGB colorized image for display purposes of the ground truth */
  cv::Mat drawGroundTruth( cv::Mat & image, std::vector<bd::ShapeMetaData> const & shapes );
}
#endif //  GENERATE_H_
