#include "GroupingNeuron.H"
#include "Random.H"

#include <iostream>
#include <cassert>
#include <cereal/types/polymorphic.hpp>
#include <mutex>

CEREAL_REGISTER_TYPE(bd::GroupingNeuron)

double bd::GroupingNeuron::updateImpl( SystemState const & state )
{
  const double noise = normalReal()() * state.noiseScale;

  double feedforward = 0;
  double ltFeedforward = 0;
  double feedback_e = 0, feedback_i = 0;
  double lateralExcitation = 0;

  auto const & preInput = itsInput.pre();

  // Excitation
  for( auto const & preEdge : preInput )
  {
    const double incomingActivation = preEdge.neuron->activation();
    const double currentInput = preEdge.weight * incomingActivation;
    const double ltCurrentInput = preEdge.ltWeight * incomingActivation;

    if( preEdge.hasFlag( ConnectionFlags::Silent ) )
      continue;

    if( preEdge.hasFlag( ConnectionFlags::Feedforward ) )
    {
      if( preEdge.hasFlag( ConnectionFlags::Direct ) && !state.dy_bo_to_grouping_feedforward )
        continue;

      feedforward += currentInput;
      ltFeedforward += ltCurrentInput;
    }
    else if( preEdge.hasFlag( ConnectionFlags::Feedback ) )
    {
      if( preEdge.hasFlag( ConnectionFlags::Direct ) && !state.dy_proto_to_grouping_feedback )
        continue;

      if( currentInput >= 0 )
        feedback_e += currentInput;
      else
        feedback_i += currentInput;
    }
    else if( preEdge.hasFlag( ConnectionFlags::Lateral | ConnectionFlags::Excitatory ) )
      lateralExcitation += currentInput;
  }

  // Ambiguity and feedback
  const double feedback = std::max( feedback_e + feedback_i, 0.0 );
  const double ambiguity = state.cl_ambiguity ? std::min( feedback_e, -feedback_i ) : 0.0;

  itsModulatoryInput.update( feedback );

  // Calculates activation value before threshold and min/max clamping
  auto calculatePreClamp = [=]( double ff, double lateral, double fb, double inhib, double noise )
  {
    const double clampedFF = clamp(ff);
    const double ffGating = clampedFF * clampedFF;

    const double input = ff + lateral * ffGating + fb * ffGating + noise;

    double result = input / ( 1.0 + inhib );

    if( result <= 0 )
      return 0.0;

    result = std::min( result, 1.0 ) + std::max( 0.0, std::log10( result ) );
    return result / itsGainMultiplier;
  };

  // Inhibition
  itsInhibitorySum = calculateInhibitorySum( state, itsInhibitorySum, ConnectionFlags::Column );
  itsLateralInhibitorySum = calculateInhibitorySum( state, itsLateralInhibitorySum, ConnectionFlags::Lateral );

  updateGainControls( ConnectionFlags::Column );

  // Activation
  itsUninhibitedActivation.update( calculatePreClamp( feedforward, lateralExcitation, feedback, ambiguity, noise ) );
  double preClamp = calculatePreClamp( feedforward, lateralExcitation, feedback, itsInhibitorySum + itsLateralInhibitorySum + ambiguity, noise );

  // DEBUG
//  if( itsActivationThreshold == 0.0 )
//  {
//    itsActivationThreshold = itsFastThreshold = itsDecayThreshold = itsMaxThreshold = 0.2;
//  }

  updateThresholds( state, ltFeedforward, feedforward,
                    itsInhibitorySum,
                    0.4 * itsNormalizationCoefficients.column, // inhibitory threshold
                    0.01,       // fast
                    0.0025,     // slow
                    0.0001 );   // decay

  // DEBUG
  //if( itsActivationThreshold == 0.0 )
  //{
  //  itsActivationThreshold = itsFastThreshold = itsDecayThreshold = itsMaxThreshold = 0.2;
  //}

  // New idea with thresholds:
  //  Thresholds only apply to neurons that were previously in an 'off' state - they can't turn off neurons
  //  that are already on, even if those neurons become inhibited somehow. Rather they use long term weighted activation
  //  to control the initial turn-on, and allow adjustment to this turn on threshold using short term weight.
  //
  //  This scheme allows a neuron to have a good dynamic range of activation once on but still segregates neurons based
  //  on thresholding input.


  const double inverse = 1.0 / itsNormalizationCoefficients.feedforward;
  if( feedforward < itsFastThreshold * inverse )
  {
    preClamp = 0.0;
    itsUninhibitedActivation.update( 0.0 );
  }

  // Compute final output
  const double value = clamp( preClamp, 0.0, state.maxActivation );

  return value;
}

void bd::GroupingNeuron::learnImpl( SystemState const & state, double currentActivation )
{
  if( state.learningRule == LearningRule::Hebbian )
  {
    shortTermHebbianNormalization( state.gnMaxWeights.feedforward * itsNormalizationCoefficients.feedforward,
                                   currentActivation, state,
                                   ConnectionFlags::Dynamic | ConnectionFlags::Feedforward );
    return;
  }
  else if( state.learningRule == LearningRule::Conflict )
  {
    // Slow params: .5, .05
    // Fast params: .05, .005
    const double alpha = calculateShortTermAlpha(); // excitation
    const double beta = 1.0; // inhibition

    const int kappaFlagsFeedback    = ConnectionFlags::Feedback | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;
    const int kappaFlagsFeedforward = ConnectionFlags::Feedforward | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

    // Feedforward Update
    {
      const int flags = ConnectionFlags::Direct | ConnectionFlags::Feedforward | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

      // Short term (skip if not on)
      if( currentActivation != 0.0 )
      {
        constexpr double localLearningRate = 0.01; // feedforward needs to be slower moving
        constexpr double ltmRate = 1.0; // how much faster LTM should adjust

        shortTermUpdate( state, flags, kappaFlagsFeedforward,
                         currentActivation, itsLateralInhibitorySum,
                         localLearningRate, ltmRate,
                         alpha, beta );
      }

      // Long term
      propogateAccumulatedWeight( state, ConnectionFlags::Feedforward | ConnectionFlags::Dynamic,
          state.gnMaxWeights.feedforward * itsNormalizationCoefficients.feedforward,
          0.5, // st
          0.05, // lt
          100.0 );
    }

    // Feedback Update
    {
      const int flags = ConnectionFlags::Direct | ConnectionFlags::Feedback | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

      // Short term (skip if not on)
      if( currentActivation != 0.0 )
      {
        constexpr double localLearningRate = 1.00; // feedforward needs to be slower moving
        constexpr double ltmRate = 1.0; // how much faster LTM should adjust

        shortTermUpdate( state, flags, kappaFlagsFeedback,
                         currentActivation, itsInhibitorySum,
                         localLearningRate, ltmRate,
                         alpha, beta );
      }

      // Long term
      propogateAccumulatedWeight( state, flags,
          state.gnMaxWeights.feedback * itsNormalizationCoefficients.feedback,
          0.05,
          0.005,
          100.0 );
    }

    // Feedback Update Inhibitory
    {
      const int flags = ConnectionFlags::Direct | ConnectionFlags::Feedback | ConnectionFlags::Inhibitory | ConnectionFlags::Dynamic;

      // Short term (skip if not on)
      if( currentActivation != 0.0 )
      {
        constexpr double localLearningRate = 1.00; // feedforward needs to be slower moving
        constexpr double ltmRate = 1.0; // how much faster LTM should adjust

        shortTermUpdate( state, flags, kappaFlagsFeedback,
                         currentActivation, itsInhibitorySum,
                         localLearningRate, ltmRate,
                         alpha, beta );
      }

      // Long term
      propogateAccumulatedWeight( state, flags,
          state.gnMaxWeights.feedbackInhibitory * itsNormalizationCoefficients.feedback,
          0.05,
          0.005,
          100.0 );
    }

    // Lateral Update
    //{
    //  const int flags = ConnectionFlags::Lateral | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

    //  // Short term (skip if not on)
    //  if( currentActivation != 0.0 )
    //  {
    //    constexpr double localLearningRate = 1.0;
    //    constexpr double ltmRate = 100.0; // how much faster LTM should adjust

    //    shortTermUpdate( state, flags, flags,
    //                     currentActivation, itsLateralInhibitorySum,
    //                     localLearningRate, ltmRate,
    //                     alpha, beta );
    //  }

    //  // Long term
    //  propogateAccumulatedWeight( state, ConnectionFlags::Lateral | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic,
    //      state.gnMaxWeights.lateral * itsNormalizationCoefficients.lateral,
    //      0.05,
    //      0.05,
    //      10.0 );
    //}
  } // end Conflict learning

  // Column connections learned using a modified LTM hebbian
  columnNormalization( state.gnMaxWeights.column,
                       currentActivation,
                       state,
                       ConnectionFlags::Dynamic | ConnectionFlags::Column | ConnectionFlags::Inhibitory );

  //updateLateralWeights( 0.8, // hybrid threshold
  //                      state.gnLateralTotalWeight, currentActivation, state,
  //                      ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );

  //  // OLD STUFF BELOW (Long term Hebbian)
  //  const double accumulatorScaling = 1.0; // makes learning happen faster
  //  hebbianNormalization( state.groupingTotalWeight * itsNormalizationCoefficients.feedforward,
  //                        currentActivation * accumulatorScaling,
  //                        ConnectionFlags::Dynamic | ConnectionFlags::Feedforward );
}
