#include "InhibitoryInterNeuron.H"
#include "Random.H"

#include <iostream>
#include <cassert>
#include <cereal/types/polymorphic.hpp>

CEREAL_REGISTER_TYPE(bd::InhibitoryInterNeuron)

bd::InhibitoryInterNeuron::InhibitoryInterNeuron() : Neuron()
{
}

bd::InhibitoryInterNeuron::InhibitoryInterNeuron( std::pair<double, double> pos ) :
  Neuron( pos )
{
}

double bd::InhibitoryInterNeuron::updateImpl( SystemState const & /*state*/ )
{
  // These refer to average weighted input of all incoming FF neurons
  static const double lowTarget    = 0.1;   //!< Level we need to drop below to decrease inhibition
  static const double stableTarget = 0.2;  //!< Target level for activation
  static const double highTarget   = 0.3;   //!< Level at which we begin inhibition

  const double scale = 0.01;
  const double noise = normalReal()() * scale;

  double input = 0.0;

  double feedforward = 0;
  int numFF = 0;

  // Excitation
  for( size_t i = 0; i < itsInput.pre().size(); ++i )
  {
    const auto & in = itsInput.pre()[i];

    if( in.weight > 0 && in.hasFlag( ConnectionFlags::Feedforward ) )
    {
      const double currentInput = in.weight * in.neuron->activation();
      feedforward += currentInput;
      ++numFF;
    }
  }

  const double averageFeedforward = feedforward / numFF;

  input += averageFeedforward;
  input += noise;

  // Calculate current value
  double value = activation();

  if( input > highTarget )
  {
    // increase inhibition proportional to how high input is and current level of inhibition
    value += input - lowTarget;
  }
  else if( input < lowTarget )
  {
    // decrease inhibition etc
    value -= lowTarget - input;
  }

  if( value > 0 )
    std::cerr << "INHIB BY " << value << " " << numFF << " " << averageFeedforward << std::endl;

  return std::max( value, 0.0 );
}

void bd::InhibitoryInterNeuron::learnImpl( SystemState const & /*state*/, double /*currentActivation*/ )
{
}
