#include "Neuron.H"
#include "Random.H"

#include <cmath>
#include <fstream>
#include <cassert>

#include <iostream>
#include <iomanip>

#include <future>

#include "EdgeNeuron.H"
#include "GroupingNeuron.H"
#include "BorderOwnershipNeuron.H"

constexpr double bd::Input::INITIAL_ACC_WEIGHT;
constexpr double bd::Input::INITIAL_LATERAL_ACC_WEIGHT;

namespace bd
{
  int preIndex  = 0; //!< Marks current pre index for all NeuronVariables
  int postIndex = 1; //!< Marks current post index for all NeuronVariables
}

bd::Neuron::Neuron() :
  itsInput{{}},
  itsBCMThreshold{0.0},
  itsMaxThreshold{0},
  itsFastThreshold{0},
  itsActivationThreshold{0},
  itsDecayThreshold{0},
  itsGainMultiplier{1.0},
  itsModulatoryGainMultiplier{1.0},
  itsUninhibitedActivation{0},
  itsInhibitorySum{0},
  itsLateralInhibitorySum{0},
  itsModulatoryInput{0},
  itsActivation{0},
  itsPosition( {-1, -1} )
{ }

bd::Neuron::Neuron( std::pair<double, double> pos ) :
  itsInput{{}},
  itsBCMThreshold{0.0},
  itsMaxThreshold{0},
  itsFastThreshold{0},
  itsActivationThreshold{0},
  itsDecayThreshold{0},
  itsGainMultiplier{1.0},
  itsModulatoryGainMultiplier{1.0},
  itsUninhibitedActivation{0},
  itsInhibitorySum{0},
  itsLateralInhibitorySum{0},
  itsModulatoryInput{0},
  itsActivation{0},
  itsPosition( pos )
{ }

bd::Neuron::Neuron( Neuron const & other, std::pair<double, double> pos ) :
  itsInput( other.itsInput ),
  itsBCMThreshold( other.itsBCMThreshold ),
  itsMaxThreshold( other.itsMaxThreshold ),
  itsFastThreshold( other.itsFastThreshold ),
  itsActivationThreshold( other.itsActivationThreshold ),
  itsDecayThreshold( other.itsDecayThreshold ),
  itsGainMultiplier( other.itsGainMultiplier ),
  itsModulatoryGainMultiplier( other.itsModulatoryGainMultiplier ),
  itsUninhibitedActivation( other.itsUninhibitedActivation ),
  itsInhibitorySum( other.itsInhibitorySum ),
  itsLateralInhibitorySum( other.itsLateralInhibitorySum ),
  itsModulatoryInput( other.itsModulatoryInput ),
  itsActivation( other.itsActivation ),
  itsPosition( pos )
{ }

double bd::Neuron::activation( bool useSigmoid ) const
{
//  if( useSigmoid )
//    return kumaraswamy( itsActivation.pre );
//  else

  return itsActivation.pre();
}

double bd::Neuron::activationSanger() const
{
    return itsActivation.post();
}

double bd::Neuron::modulatoryInput() const
{
  return itsModulatoryInput.pre();
}

std::pair<double, double> bd::Neuron::position() const
{
  return itsPosition;
}

std::pair<double, double> bd::Neuron::centroid( const int flags ) const
{
  auto input = observer( itsInput.pre(), flags );
  //const auto & input = itsInput.pre();

  // Step 1: calculate max weight
  double maxWeight = 0.0;

  for( const auto & in : input )
    maxWeight = std::max( maxWeight, std::abs(in.get().weight) );

//  for( size_t i = 0; i < input.size(); ++i )
//  {
//    const auto & in = input[i];
//    if( in.hasFlag( flags ) )
//      maxWeight = std::max( maxWeight, std::abs(in.weight) );
//  }

  // Step 2: Create weighted centroid using normalized weight values
  double xx = 0.0, yy = 0.0;
  double weightSum = 0.0;

  for( const auto & in : input )
  //for( size_t i = 0; i < input.size(); ++i )
  {
    //const auto & in = input[i];

    //if( in.hasFlag( flags ) )
    {
      double x, y;
      std::tie( x, y ) = in.get().neuron->position();

      const double weight = std::abs(in.get().weight) / maxWeight;

      xx += x * weight;
      yy += y * weight;
      weightSum += weight;
    }
  }

  if( maxWeight == 0.0 || weightSum == 0.0 )
    return {-1.0, -1.0};
  else
    return {xx / weightSum, yy / weightSum};
}

bool bd::Neuron::hasCompetitors()
{
  auto const & indices = getInputIndices( ConnectionFlags::Column );
  return !indices.empty();
}

void bd::Neuron::addInput( Input in )
{
  itsInput.pre().push_back( in );
}

void bd::Neuron::update( SystemState const & state )
{
  itsActivation.update( updateImpl( state ) );
  //itsActivation.updateSmooth( updateImpl( state ), 0.5 );
}

void bd::Neuron::learn( SystemState const & state )
{
  learnImpl( state, itsActivation.post() );
}

void bd::Neuron::setActivation( double activation, double noiseScale )
{
  const double noise = normalReal()() * noiseScale;
  const double noisyActivation = clamp( activation + noise );
  itsActivation.initialize(noisyActivation);
}

double bd::Neuron::calculateInhibitorySum( SystemState const & state, const double prevInhibition, const int flags )
{
  // Used for visualizing V1 - disables all lateral interaction (including columns)
  if( !state.useLateralConnections )
    return 0.0;

  const double myActivation = activation();
  //const double myActivation = itsUninhibitedActivation.pre();

  /* Notes:
   *  Comparing to non-inhibited activation is great for our pair because it lets a pure "who has more input"
   *  decide the winner.
   *
   *  It is bad for the non-dominant pair because they will only receive inhibition from the dominant neuron.
   *  In essence it distributes the same inhibition to everyone in the column, only from the absolute winner.
   *
   *  ---
   *
   *  The downsides to raw activation comparisons is that it makes it more difficult to correct a noise based
   *  decision that receives some reinforcement.
   */

  double inhib = 0;

  auto const & indices = getInputIndices( flags );
  auto const & preInput = itsInput.pre();

  for( auto i : indices )
  {
    const auto & in = preInput[i];

    // TODO: re-enable silent
//    if( in.hasFlag( ConnectionFlags::Silent ) )
//      continue;

    const double inActivation = in.neuron->activation();
    //const double inActivation = in.neuron->itsUninhibitedActivation.pre();
    if( inActivation >= myActivation )
      inhib += inActivation * std::abs(in.weight);
  }

  return inhib;
}

double bd::Neuron::calculateShortTermAlpha( const double inhibProportion, const double maxInflection )
{
  /* Want to set inflection point equal to max inhibition weight.
   * Using inflection point equation with beta fixed to 1.0:
   *  Inhib = alpha / (alpha + beta)
   *  => alpha = Inhib / (1 - Inhib) */

  assert( inhibProportion > 0.0 && inhibProportion <= 1.0 );
  assert( maxInflection > 0.0 && maxInflection < 1.0 );

  auto const & indices = getInputIndices( ConnectionFlags::Column | ConnectionFlags::Dynamic );
  auto const & preInput = itsInput.pre();

  // Get max
  double maxInhib = 0.0;
  for( auto i : indices )
    maxInhib = std::max( maxInhib, std::abs(preInput[i].weight) );

  const double target = inhibProportion * clamp( maxInhib );
  const double alpha = target / (1.0 - target);
  const double maxAlpha = maxInflection / (1.0 - maxInflection );

  if( std::isnormal(alpha) )
    return std::min( alpha, maxAlpha );
  else
    return maxAlpha;
}

double bd::Neuron::computeMaxLearnedActivation( const int type, const double learnThreshold )
{
  double maxWeight = -std::numeric_limits<double>::max();
  const int flags = type | ConnectionFlags::Dynamic;

  auto const & indices = getInputIndices( flags );
  auto const & preInput = itsInput.pre();

  if( indices.empty() ) // do not use kappa if no connections available
    return 1.0;

  for( auto i : indices )
  {
    const auto & preEdge = preInput[i];
    maxWeight = std::max( maxWeight, std::abs(preEdge.weight) );
  }

  const double weightCutoff = learnThreshold * maxWeight;
  double maxLearnedActivation = 0.0;

  for( auto i : indices )
  {
    const auto & preEdge = preInput[i];
    if( std::abs(preEdge.weight) > weightCutoff )
      maxLearnedActivation = std::max( maxLearnedActivation, preEdge.neuron->activation() );
  }

  return clamp( maxLearnedActivation );
}

double bd::Neuron::computeHybridThreshold( double percentile, const int flags ) const
{
  const auto & preInput = itsInput.pre();
  std::vector<double> values; values.reserve( 70 );

  for( const auto & preEdge : preInput )
  {
    if( preEdge.hasFlag( flags ) )
      values.push_back( preEdge.neuron->activation() );
  }

  const size_t index = values.size() * percentile;
  std::nth_element( values.begin(), values.begin() + index, values.end() );

  return values[index];
}

bool bd::Neuron::passesHybridThreshold( double currentActivation, double percentile, const int flags ) const
{
  return currentActivation > computeHybridThreshold( percentile, flags );
}

void bd::Neuron::learnImpl( SystemState const & /*state*/, double /*currentActivation*/ )
{
  throw "Not implemented";
}

double bd::Neuron::updateImpl( SystemState const & /*state*/ )
{
  throw "Not Implemented";
  return 0;
}

bool bd::Neuron::operator==( Neuron const & other ) const
{
  return this == &other;
}

bool bd::Neuron::operator!=( Neuron const & other ) const
{
  return !operator==( other );
}

void bd::Neuron::dumpWeights( std::string filename ) const
{
  std::ofstream out( filename );

  for( const auto & in : itsInput.pre() )
    out << in.weight << std::endl;
}

void bd::Neuron::printNeuronInfo() const
{
  std::cout << "===================================================" << std::endl;
  std::cout << "I am at " << position().first << ", " << position().second << " (" << this << ")";
  std::cout << " with activation: " << activation();
  std::cout << ", lateralInhibSum: " << itsLateralInhibitorySum;
  std::cout << ", inhibSum: " << itsInhibitorySum;
  std::cout << ", gainMult: " << itsGainMultiplier;
  std::cout << ", modGainMult: " << itsModulatoryGainMultiplier.post();
  std::cout << ", modInput: " << modulatoryInput();
  //std::cout << ", bcmthresh: " << itsBCMThreshold.pre();
  std::cout << ", maxThresh: " << itsMaxThreshold << ", fastThresh: " << itsFastThreshold;
  std::cout << ", actThresh: " << itsActivationThreshold << ", deccaythresh: " << itsDecayThreshold << std::endl;

  std::cout << "Old activation: " << itsActivation.post() << ", uninhibActivation: " << itsUninhibitedActivation.post() << std::endl;

  auto edgeNeuron = dynamic_cast<bd::EdgeNeuron const *>( this );
  if( edgeNeuron )
  {
    std::cout << "Display Cache: ";
    for( const auto val : edgeNeuron->itsDisplayCache )
      std::cout << val << " ";
    std::cout << std::endl;
  }

  //std::cout << "Max FB: " << itsNormalizationCoefficients.feedback * SystemState::feedbackTotalWeight << ", ";
  std::cout << "FB coef: " << itsNormalizationCoefficients.feedback << std::endl;

  //std::cout << "Max FF: " << itsNormalizationCoefficients.feedforward * SystemState::groupingMaxWeight << ", ";
  std::cout << "FF coef: " << itsNormalizationCoefficients.feedforward << std::endl;

  //std::cout << "Max FFs: " << itsNormalizationCoefficients.feedforwardSkip * SystemState::groupingMaxWeight << ", ";
  std::cout << "FFs coef: " << itsNormalizationCoefficients.feedforwardSkip << std::endl;

  std::cout << "LAT coef: " << itsNormalizationCoefficients.lateral << std::endl;
}

void bd::Neuron::printWeights( std::ostream & os, std::function<bool(Input const &)> && check ) const
{
  for( const auto & in : itsInput.pre() )
  {
    if( !check( in ) ) continue;
    os << in.weight << " ";
  }
  os << std::endl;
}

void bd::Neuron::printConnections( std::string const & heading, std::function<bool(Input const &)> && check, bool verbose ) const
{
  std::cout << heading << std::endl;

  double weightSumPosOnly = 0.0;
  double weightSum = 0.0;
  double ltSum = 0.0;
  double accSum = 0.0;
  double numWeights = 0;
  double weightedInput = 0.0;
  double ltWeightedInput = 0.0;

  for( const auto & in : itsInput.post() )
  {
    if( !check( in ) ) continue;

    int inX, inY; std::tie(inX, inY) = in.neuron->position();

    weightSum += in.weight;
    if( in.weight > 0 )
      weightSumPosOnly += in.weight;
    ltSum += in.ltWeight;
    accSum += in.accumulatedWeight;
    weightedInput += in.weight * in.neuron->itsActivation.post();
    ltWeightedInput += in.ltWeight * in.neuron->itsActivation.post();

    ++numWeights;

    if( verbose )
    {
      std::cout << std::setw(2);
      std::cout << inX << ", " << std::setw(2) << inY << std::fixed;

      auto bo_in = std::dynamic_pointer_cast<bd::BorderOwnershipNeuron>( in.neuron );
      if( bo_in )
        std::cout << " (" << static_cast<int>(bo_in->orientation * 180 / M_PI) << ")";

      std::cout << std::setprecision(3) << ", act: " << in.neuron->itsActivation.post();
      std::cout << std::setprecision(6) << ", weight: " << in.weight << ", ltWeight: " << in.ltWeight;
      std::cout << ", acc: " << in.accumulatedWeight;
      if( in.hasFlag( ConnectionFlags::Column ) )
      {
        std::cout << ", inhibSum: " << in.neuron->itsInhibitorySum;
        std::cout << ", modGain: " << in.neuron->itsModulatoryGainMultiplier.post();
        std::cout << ", modInput: " << in.neuron->modulatoryInput();
        std::cout << ", latInhibSum: " << in.neuron->itsLateralInhibitorySum;
      }
      std::cout << ", bcmthresh: " << in.neuron->itsBCMThreshold.post();
      std::cout << ", maxThresh: " << in.neuron->itsMaxThreshold;
      std::cout << ", fastThresh: " << in.neuron->itsFastThreshold;
      std::cout << ", activeThresh: " << in.neuron->itsActivationThreshold;
      std::cout << ", decayThresh: " << in.neuron->itsDecayThreshold << std::endl;
    }
  }

  double stdTotal = 0.0;
  for( const auto & in : itsInput.post() )
  {
    if( !check( in ) ) continue;
    if( numWeights > 0 )
      stdTotal += std::pow( in.accumulatedWeight - (accSum/numWeights), 2.0 );
  }

  const double accmean = numWeights > 0 ? accSum / numWeights : 0.0;
  const double accstd = numWeights > 0 ? std::sqrt(stdTotal/numWeights ) : 0.0;

  std::cout << "Weight sum: " << weightSum << ", posWeightSum: " << weightSumPosOnly << ", weightedInput: " << weightedInput << ", ltWeightedInput: " << ltWeightedInput;
  std::cout << ", ltsum: " << ltSum << ", accsum: " << accSum << ", accmean: " << accmean;
  std::cout << ", accstd: " << accstd << ", num: " << numWeights << std::endl;
}


void bd::Neuron::columnNormalization( double normalizationCoefficient, double currentActivation, SystemState const & state, int flags )
{
  double totalAccumulatedWeight = 0.0;

  auto const & indices = getInputIndices( flags );
  auto const & preInput = itsInput.pre();
  auto & postInput = itsInput.post();

  // Sum all dynamic edges
  for( auto i : indices )
  {
    const auto & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    // Weighted hebbian update modulated by inhibition signal
    const double update = preEdge.neuron->activation() * std::abs(preEdge.weight) *
                          currentActivation * clamp( 1.0 - itsInhibitorySum );

    postEdge.accumulatedWeight = preEdge.accumulatedWeight + update;
    totalAccumulatedWeight += postEdge.accumulatedWeight;
  }

  // Normalize weights based on accumulated weights
  const double normalizer = normalizationCoefficient / totalAccumulatedWeight;

  for( auto i : indices )
  {
    const auto & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    postEdge.weight = clamp( postEdge.accumulatedWeight * normalizer * sgn(preEdge.weight), -state.maxWeight, state.maxWeight );
  }
}

void bd::Neuron::hebbianNormalization( double normalizationCoefficient, double currentActivation, SystemState const & state, int flags )
{
  double totalAccumulatedWeight = 0.0;

  const auto & preInput = itsInput.pre();
  auto & postInput = itsInput.post();

  const double normalizedActivation = normalizeActivation( currentActivation, state.maxActivation );

  // Sum all dynamic edges
  for( size_t i = 0; i < preInput.size(); ++i )
  {
    auto const & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    if( preEdge.hasFlag( flags ) )
    {
      const double inActivation = normalizeActivation( preEdge.neuron->activation(), state.maxActivation );

      // TESTING:
      if( preEdge.hasFlag( ConnectionFlags::Column ) )
      {
        postEdge.accumulatedWeight = preEdge.accumulatedWeight + inActivation * std::abs(preEdge.weight) * normalizedActivation *
                                     clamp( 1.0 - itsInhibitorySum );
      }
      else
        postEdge.accumulatedWeight = preEdge.accumulatedWeight + inActivation * preEdge.weight * normalizedActivation;

      totalAccumulatedWeight += postEdge.accumulatedWeight;
    }
  } // end for

  // Normalize weights based on accumulated weights
  const double normalizer = normalizationCoefficient / totalAccumulatedWeight;

  for( size_t i = 0; i < preInput.size(); ++i )
  {
    auto const & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    if( preEdge.hasFlag( flags ) )
      postEdge.weight = clamp( postEdge.accumulatedWeight * normalizer * sgn(preEdge.weight), -state.maxWeight, state.maxWeight );
  } // end for
}

void bd::Neuron::shortTermHebbianNormalization( double normalizationCoefficient, double currentActivation, SystemState const & state, int flags )
{
  const double learningRate = state.learningRate;
  double totalWeight = 0.0;

  auto const & preInput = itsInput.pre();
  auto & postInput = itsInput.post();

  // Sum all dynamic edges
  for( size_t i = 0; i < preInput.size(); ++i )
  {
    auto const & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    if( preEdge.hasFlag( flags ) )
    {
      postEdge.weight = preEdge.weight + learningRate * preEdge.neuron->activation() * currentActivation;

      totalWeight += postEdge.weight;
    }
  } // end for

  // Normalize weights based on desired total weights
  const double normalizer = normalizationCoefficient / totalWeight;

  for( size_t i = 0; i < preInput.size(); ++i )
  {
    auto const & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    if( preEdge.hasFlag( flags ) )
    {
      postEdge.weight = clamp( postEdge.weight * normalizer * sgn(preEdge.weight), -state.maxWeight, state.maxWeight );
    }
  } // end for
}

void bd::Neuron::bcmUpdate( double currentActivation, SystemState const & state, int flags )
{
  const double learningRate = state.learningRate;
  const double maxWeight = state.maxWeight;

  const auto & preInput = itsInput.pre();
  auto & postInput = itsInput.post();

  // BCM Update equation: dW = xy(y-theta)
  for( size_t i = 0; i < preInput.size(); ++i )
  {
    auto const & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    if( preEdge.hasFlag( flags ) )
    {
      const double delta = learningRate * preEdge.neuron->activation() * currentActivation * ( currentActivation - itsBCMThreshold.post() );
      postEdge.weight = clamp( postEdge.weight + delta, 0, maxWeight );
    }
      // optional term: subtract epison * preEdge.weight;
  } // end for
}

void bd::Neuron::updateLateralWeights( SystemState const & state, const int flags,
                                       const double threshold, const double normalizationCoefficient, const double currentActivation,
                                       const double inhibitorySum, const double alpha, const double beta )
{
  // Lateral (inhibitory) weigths are updated in an anti-Hebbian fashion. The current implementation uses
  // a scheme derived from a Hybrid Learning Rule (Jain et al 2015?) to achieve a multi-map like configuration.
  //
  // Inhibition is increased if exclusively this or an incoming neuron is above the hybrid threshold. If both are
  // above threshold, inhibition is decreased.
  //
  // Inhibitory weights use long term accumulator to set their weights, which are normalized to a total value.
  //
  // ME THEM
  // 0  0     do nothing
  // 0  1     increase inhibition
  // 1  0     increase inhibition
  // 1  1     decrease inhibition
  //
  // In addition, when a neuron is experiencing column inhibition, it will act as if it is not passing its hybrid
  // threshold.
  //
  // if( inhib )
  // ME THEM
  // X  0     do nothing
  // X  1     increase inhibition

  const double normalizedActivation = currentActivation;//normalizeActivation( currentActivation, state.maxActivation );
  const double ltmAlpha = 1.0;

  const double hybridThreshold = computeHybridThreshold( threshold, flags );
  const bool passesHybridThreshold = currentActivation > hybridThreshold;

  const double inflectionPoint = 1.0 / (1.0 + beta / alpha);
  const bool isInhibited = inhibitorySum > inflectionPoint;

  const double learningRate = state.learningRate;
  const auto & preInput = itsInput.pre();
  auto & postInput = itsInput.post();

  double accumulatedWeight = 0.0;

  for( size_t i = 0; i < preInput.size(); ++i )
  {
    auto const & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    if( preEdge.hasFlag( flags ) )
    {
      const double normalizedPreAct = preEdge.neuron->activation();//normalizeActivation( preEdge.neuron->activation(), state.maxActivation );
      const double targetWeight = normalizedActivation * normalizedPreAct;

      const bool incomingPassesHybrid = preEdge.neuron->activation() > hybridThreshold;
      double update = 0.0;

      if( isInhibited )
      {
        if( incomingPassesHybrid )
        {
          update += learningRate * normalizedPreAct;
        }

        // if incoming does not pass, do nothing
      }
      else
      {
        if( passesHybridThreshold ^ incomingPassesHybrid )
        {
          // Only one passes threshold; increase inhibition
          update += learningRate * normalizedPreAct;
        }
        else if( passesHybridThreshold & incomingPassesHybrid )
        {
          // Both pass threshold; decrease inhibition
          update -= learningRate * targetWeight;
        }
        // If neither pass threshold, do nothing
      }

      postEdge.accumulatedWeight = std::max( preEdge.accumulatedWeight + ltmAlpha * update, 0.0 );
      accumulatedWeight += postEdge.accumulatedWeight;
    }
  } // end for

  // Normalize weights based on accumulated weights
  const double normalizer = accumulatedWeight ? normalizationCoefficient / accumulatedWeight : 1.0;

  for( size_t i = 0; i < preInput.size(); ++i )
  {
    auto const & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    if( preEdge.hasFlag( flags ) )
      postEdge.weight = clamp( -1.0 * postEdge.accumulatedWeight * normalizer, -state.maxWeight, state.maxWeight );
  } // end for
}

void bd::Neuron::shortTermUpdate( SystemState const & state,
    const int flags,
    const int maxLearnedFlags,
    const double currentActivation, const double inhibitorySum,
    const double localLearningRate, const double ltmRate,
    const double alpha, const double beta )
{
  const bool isInhibitory = (flags & ConnectionFlags::Inhibitory) == ConnectionFlags::Inhibitory;
  const double learningRate = state.learningRate;
  // undo the local learning rate on ltm growth
  const double ltmAlpha = ltmRate / localLearningRate;

  // max learned activation
  const double kappa = state.cl_spreading ? computeMaxLearnedActivation( maxLearnedFlags ) : 1.0;

  auto const & indices = getInputIndices( flags );
  auto const & preInput = itsInput.pre();
  auto & postInput = itsInput.post();

  // Takes a min or max as appropriate; when we are doing an inhibitory update, keep the max at zero, otherwise
  // zero is the minimum
  auto accumulatorUpdate = isInhibitory ? []( double newVal ){ return std::min( newVal, 0.0 ); } :
                                          []( double newVal ){ return std::max( newVal, 0.0 ); };

  for( auto i : indices )
  {
    const auto & preEdge = preInput[i];
    auto & postEdge = postInput[i];

    const double inputActivation = preEdge.neuron->activationSanger();
    const double targetWeight = inputActivation * currentActivation;

    const double lowerBound = isInhibitory ? -state.maxWeight : preEdge.initialWeight;
    const double upperBound = isInhibitory ? preEdge.initialWeight : state.maxWeight;

    double ltmUpdate = 0.0;

    // Inhibition
    if( state.cl_unlearning )
    {
      const double inhibitoryModulation = clamp( inhibitorySum );
      const double b = beta * localLearningRate;
      const double update = b * learningRate * inhibitoryModulation * targetWeight;

      postEdge.weight = clamp( postEdge.weight - update, lowerBound, upperBound );
      ltmUpdate -= update;
    }

    // Excitation
    {
      // Short term weight delta
      const double inhibitoryModulation = state.cl_unlearning ? (1.0 - clamp( inhibitorySum )) : 1.0;
      const double a = alpha * localLearningRate; // alpha is used to balance learning vs unlearning rate (learning should never be faster than unlearning)
      const double update = a * learningRate * inhibitoryModulation * targetWeight * kappa;

      postEdge.weight = clamp( postEdge.weight + update, lowerBound, upperBound );
      ltmUpdate += update;
    }

    postEdge.accumulatedWeight = accumulatorUpdate( preEdge.accumulatedWeight + ltmAlpha * ltmUpdate );
  }
}

void bd::Neuron::propogateAccumulatedWeight( SystemState const & state, const int flags,
    const double maxWeight, // should include ff/fb/etc coefficient
    const double stChaseFactor,
    const double ltChaseFactor,
    const double ltSlowdown )
{
  const bool isInhibitory = (flags & ConnectionFlags::Inhibitory) == ConnectionFlags::Inhibitory;

  /**
   * Long term weights are closely related to the distribution of the accumulators.
   *
   * At each iteration, the long term weights will move towards their current short term weights
   * at a rate dependent upon their accumulator and short term weights.
   *
   * At the same time, the short term weights constantly decay towards their long term weight
   * (increasing or decreasing).
   *
   * If the long term weight sum exceeds some maximum, it will be normalized.
   *
   * All weights are restricted to be in the range [0, 1]
   */

  double accumulatorSum = 0.0;
  double longTermSum = 0.0;
  int numConnections = 0;

  auto const & indices = getInputIndices( flags );
  auto & postInput = itsInput.post();

  for( auto i : indices )
  {
    const auto & postEdge = postInput[i];

    accumulatorSum += std::abs(postEdge.accumulatedWeight);
    longTermSum += std::abs(postEdge.ltWeight);
    ++numConnections;
  }

  // Increment each long term weight based upon the current distribution of the accumulators
  //  and normalize inhibited weights by the total accumulator sum
  double postWeightSum = 0.0;

  for( auto i : indices )
  {
    auto & postEdge = postInput[i];

    // Long term weights are drawn towards their short term weights at a rate dependent upon
    // the direction they are moving and how far away from the accumulator distribution they are.
    // LT Weights want to be at their long term distribution, so the further away they move from
    // it the slower they will move. Moving towards the accumulator happens at a fixed rate.

    const double accDistribution = accumulatorSum ? postEdge.accumulatedWeight / accumulatorSum : 0.0;
    const double ltDistribution  = longTermSum    ? postEdge.ltWeight / longTermSum             : 0.0;

    // NEW IDEAS HERE:
    // New idea: Long term weights move towards short term weights at a rate dependent on how far
    //           they are from their accumulated weight distribution. Weight changes that differ
    //           little are faster, but weight changes that create dramatic shifts are much slower.
    //
    {
      const double goal = postEdge.weight; // new target for long term weight

      const bool growUp   = (goal - postEdge.ltWeight) >= 0; // True if weight will increase
      const bool desireUp = (accDistribution - ltDistribution ) >= 0; // True if accumulator is higher than LT

      // gU dU
      // 0   0 weight will not increase, accumulator is lower than LT = GOTTA GO FAST
      // 0   1 weight will not increase, accumulator higher than LT = PUMP THE BRAKES
      // 1   0 weight will increase, accumulator lower than LT = PUMP THE BRAKES
      // 1   1 weight will increase, accumulator higher than LT = GOTTA GO FAST

      // Compute net slowdown applied to chase factor; if we are moving against the desired direction, divide by the
      // ratio [0, 1.0] of acc and lt distribution, made smaller by the external slowdown factor
      const double slowdown = ltSlowdown;
      const double minDist = std::min( std::abs(accDistribution), std::abs(ltDistribution) );
      const double maxDist = std::max( std::abs(accDistribution), std::abs(ltDistribution) );

      const double distFactor = accDistribution == 0.0 || ltDistribution == 0.0 ? 1.0 :
                                   growUp ^ desireUp ? minDist / (maxDist * slowdown) : 1.0;

      const double chaseFactor = ltChaseFactor * distFactor;

      postEdge.ltWeight = chaseFactor * goal + (1.0 - chaseFactor) * postEdge.ltWeight; // move towards short term weight
      postWeightSum += std::abs(postEdge.ltWeight);
    }
  } // end for

  // Normalization ensures that our post-shaped distribution still sums up to its original values, or
  // the maximum if we exceeded that

  const double lowerBound = isInhibitory ? -state.maxWeight : 0.0;
  const double upperBound = isInhibitory ? 0.0 : state.maxWeight;

  // Normalization on the long term
  {
    const double weightSumTarget = std::min( postWeightSum, maxWeight );
    const double normalizer = weightSumTarget / postWeightSum;

    for( auto i : indices )
    {
      auto & postEdge = postInput[i];
      postEdge.ltWeight = clamp( postEdge.ltWeight * normalizer, lowerBound, upperBound );
    }
  } // end LT normalization

  // Move short term towards long term
  //  larger chase values make us follow long term weight more strongly; lower values short term weights
  double postWeightSumST = 0.0;
  for( auto i : indices )
  {
    auto & postEdge = postInput[i];
    if( state.cl_slt )
      postEdge.weight = (1 - stChaseFactor) * postEdge.weight + stChaseFactor * postEdge.ltWeight;
    postWeightSumST += std::abs(postEdge.weight);
  }

  // Normalization on the short term
  {
    const double weightSumTarget = std::min( postWeightSumST, maxWeight );
    const double normalizer = weightSumTarget / postWeightSumST;

    for( auto i : indices )
    {
      auto & postEdge = postInput[i];
      postEdge.weight = clamp( postEdge.weight * normalizer, lowerBound, upperBound );
    }
  } // end ST normalization
}

void bd::Neuron::updateThresholds( SystemState const & state,
    const double ltFeedforward, const double feedforward,
    const double inhibition, const double inhibitionThreshold,
    const double fastSmoothing, const double slowSmoothing, const double decayRate )
{
  // Threshold adjustment:
  //  Neurons have several thresholds they use to control activation
  //    * Maximum threshold - this is a rolling average threshold of the neuron when it is allowed to fire
  //      - Updated if actual_activation > activation_threshold
  //      - Indirectly affects other thresholds and activation
  //    * Activation threshold - a lower threshold than the maximum that controls when the neuron can fire
  //      - Updated if silent_activation > decay_threshold
  //      - Directly controls activation
  //    * Decay threshold - a yet lower threshold that controls when the neuron is allowed to adjust its
  //                        activation threshold
  //      - Updated regardless
  //
  //
  // FF is LT FF
  // feedforward is total output of neuron prior to being clamped to (min, max) range (essentially equal to output)
  //
  //
  // Active regime
  //  When the neuron is above its activation threshold. It may not be actually firing, but it is still considered
  //  'active' for purposes of the threshold.
  //
  // Inactive regime
  //  When the neuron is below its activation threshold but above its decay threshold. The decay threshold will move
  //  towards the
  //
  // Decay regime
  //  When the neuron is below its decay threshold. The decay threshold will passively decay in this state.

  const double minThreshold = state.noiseScale * 4.0; // 4 standard deviations

  const double isInhibited = inhibitionThreshold ? inhibition >= inhibitionThreshold : false;
  const double isOverActive = feedforward > 1.1 && itsGainMultiplier < 1.01;
  const double adjustThresholds = !isInhibited && !isOverActive;

  {
    // Active regime
    if( ltFeedforward > itsActivationThreshold && adjustThresholds )
    {
      const double fastSmoothingDownward = fastSmoothing * 0.1; // Controls the rate at which fast moving thresholds decrease
      const double fastThreshFactor = 3.0; // Controls how much faster the "fast" threshold is versus "max"

      itsMaxThreshold = exponentialAvg( ltFeedforward, itsMaxThreshold, fastSmoothing, fastSmoothingDownward );
      itsFastThreshold = exponentialAvg( itsMaxThreshold, itsFastThreshold, fastSmoothing * fastThreshFactor, fastSmoothingDownward );
    }

    if( ltFeedforward > itsDecayThreshold && adjustThresholds )
    {
      if( ltFeedforward <= itsMaxThreshold )
        itsActivationThreshold = exponentialAvg( ltFeedforward, itsActivationThreshold, slowSmoothing );

      if( ltFeedforward <= itsActivationThreshold )
        itsDecayThreshold = exponentialAvg( itsActivationThreshold, itsDecayThreshold, slowSmoothing );
    }
    else
    {
      itsDecayThreshold = exponentialAvg( minThreshold, itsDecayThreshold, decayRate );
      itsFastThreshold = exponentialAvg( itsActivationThreshold, itsFastThreshold, .5 );
    }
  }
}

void bd::Neuron::updateGainControls( const int flags )
{
  /*
   * Two gain controls are used to normalize activation within a column.
   *
   * The first, called itsGainMultiplier, ensures that no neuron within a column activates
   * above 1.0, divisively normalizing activation by the maximum of any neuron in the column.
   *
   * A second, called itsModulatoryGainMultiplier, dampens activation when the clumn is receiving
   * large amounts of modulatory feedback. This usually indicates that the column is receiving
   * conflicting feedback signals and thus we want to lower activation until the modulatory input
   * can be corrected.
   *
   * The modulatory multiplier is applied simultaneously to the activation along with the gain
   * multiplier.
   *
   * Since the modulatory multiplier is an additional dampening mechanism, we don't want to
   * consider it when first calculating the gain multiplier. For this reason, the maximum activation
   * in the column will proceed as if there was no modulatory multiplier applied */

  // Get maximum activation amongst local neighborhood (including self)
  auto const & indices = getInputIndices( flags );
  auto const & preInput = itsInput.pre();

  double maxActivation = activation();// * itsModulatoryGainMultiplier.pre();
  //double sumModulatory = 0.0;//modulatoryInput();
  //double maxModDiff = 0.0, minModDiff = 100.0;

  //const double myModulatoryInput = modulatoryInput();

  for( auto i : indices )
  {
    const auto & in = preInput[i];
    maxActivation = std::max( maxActivation, in.neuron->activation() );
    //maxActivation = std::max( maxActivation, in.neuron->activation() * in.neuron->itsModulatoryGainMultiplier.pre() );

    //sumModulatory += in.neuron->modulatoryInput() * std::abs(in.weight);
    //const double modDiff = std::abs(myModulatoryInput - in.neuron->modulatoryInput());
    //maxModDiff = std::max( maxModDiff, modDiff );
    //minModDiff = std::min( minModDiff, modDiff );
  }

  /*
   * The target for max activation is 1.0.
   *
   * Error = 1.0 - Non adjusted max activation;
   *
   * e.g.: 1 - 1.5 = -.5 err
   *
   * p = -.5
   * d = (0 -- .5) = 5
   * i = -.5
   *
   * out = 1.0?
   */

  // Gain multiplier
  {
    const double newGain = std::max( 1.0, itsGainMultiplier * maxActivation );

    //itsGainMultiplier = newGain;//exponentialAvg( newGain, itsGainMultiplier, 0.5 );
    itsGainMultiplier = exponentialAvg( newGain, itsGainMultiplier, 0.5 );
  }

  // Modulatory gain

//  if( indices.size() )
//  {
//    const double newGain = 1.0 / ( 1.0 - std::min( 1.0, std::abs( maxModDiff - minModDiff )) ); // what we should be reduced by
//    //const double newGain = std::max( 1.0, maxModDiff / minModDiff );
//    //const double newGain = std::max( 1.0, myModulatoryInput * sumModulatory2 );
//    //const double newGain = std::max( 1.0, sumModulatory );
//
//    itsModulatoryGainMultiplier.update( newGain );
//    //itsModulatoryGainMultiplier.update( exponentialAvg( newGain, itsModulatoryGainMultiplier.pre(), 0.5 ) );
//  }
}

void bd::Neuron::rebuildCache()
{
  for( auto & cache : itsCache )
    cache.second.first = true;
}

auto bd::Neuron::getInputIndices( int flags ) -> std::vector<size_t> const &
{
  // we can get away with using just input.pre since the flags on pre and post
  // should always be the same
  auto lookup = itsCache.find( flags );

  if( lookup == itsCache.end() )
    lookup = itsCache.emplace_hint( lookup, flags, std::make_pair(false, filter_input( itsInput.pre(), flags )) );
  else if( lookup->second.first )
  {
    lookup->second.second = filter_input( itsInput.pre(), flags );
    lookup->second.first = false;
  }

  return lookup->second.second;
}

// ============================================================================
// ============================================================================
// ============================================================================

double bd::sigmoid( double value )
{
  return 1.0 / ( 1.0 + std::exp( -value ) );
}

double bd::gaussian(double x, double mu, double sigma) {
  const double coeff = 1.0 / (sigma * std::sqrt(2 * M_PI));
  const double val = std::exp(-(x - mu) * (x - mu) / (2 * sigma * sigma));
  return coeff * val;
}

double bd::kumaraswamy(double x, double a, double b )
{
  return 1.0 - std::pow( 1.0 - std::pow( x, a ), b );
}

std::vector<size_t> bd::filter_input( std::vector<bd::Input> & input, int flags )
{
  std::vector<size_t> result;

  const auto sz = input.size();
  for( size_t i = 0; i < sz; ++i )
  {
    if( input[i].hasFlag( flags ) )
      result.push_back( i );
  }

  return result;
}

std::vector<std::reference_wrapper<const bd::Input>> bd::observer( std::vector<bd::Input> const & input, int flags )
{
  std::vector<std::reference_wrapper<const bd::Input>> result;

  const auto sz = input.size();
  for( size_t i = 0; i < sz; ++i )
  {
    if( input[i].hasFlag( flags ) )
      result.emplace_back( std::cref( input[i] ) );
  }

  return result;
}
