#ifndef NEURON_H_
#define NEURON_H_

#include <vector>
#include <functional>

#include <cereal/types/utility.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/archives/binary.hpp>
#include <cereal/archives/json.hpp>

#include "State.H"

namespace bd
{
  // ==========================================================================
  // ==========================================================================
  // ==========================================================================
  //! Returns sign of input (1 for positive or zero, -1 for negative)
  template <typename T> inline constexpr int sgn( T val )
  { return ((T(0) < val) - (val < T(0))) ? (T(0) < val) - (val < T(0)) : 1; }

  //! Clamps value between min and max
  inline constexpr double clamp( const double value, const double min = 0.0, const double max = 1.0 )
  {
    return value < min ? min : value > max ? max : value;
  }

  //! Performs single exponential averaging
  /*! @param newValue The new value to move towards
      @param oldValue The old value
      @param upSmoothing The smoothing factor when increasing
      @param downSmoothing The smoothing factor while decreasing */
  inline constexpr double exponentialAvg( double newValue, double oldValue, double upSmoothing, double downSmoothing )
  {
    return newValue * (newValue > oldValue ? upSmoothing : downSmoothing) + (1.0-(newValue > oldValue ? upSmoothing : downSmoothing)) * oldValue;
  }

  //! Performs single exponential averaging
  /*! Uses the same smoothing factor regardless of direction */
  inline constexpr double exponentialAvg( double newValue, double oldValue, double smoothing )
  {
    return exponentialAvg( newValue, oldValue, smoothing, smoothing );
  }

  inline constexpr double normalizeActivation( double activation, double /*maxActivation*/ )
  {
    return clamp( activation );
    //return activation / maxActivation;
  }

  // ==========================================================================
  // ==========================================================================
  // ==========================================================================
  class Neuron;

  // ==========================================================================
  // ==========================================================================
  // ==========================================================================
  #define ENUM_FLAG(x) (0x1 << x)

  //! Flags that describe a connection between two neurons
  enum ConnectionFlags
  {
    Dynamic = ENUM_FLAG(1),
    Static = ENUM_FLAG(2),
    Feedforward = ENUM_FLAG(3),
    Feedback = ENUM_FLAG(4),
    Lateral = ENUM_FLAG(5),
    Column = ENUM_FLAG(6),      // Used for within column competition
    Inhibitory = ENUM_FLAG(7),  // Inhibition from any other source
    Excitatory = ENUM_FLAG(8),  // Only necessary to distinguish from inhibitory (e.g. for lateral), otherwise assumed
    DisplayAsInhibitory = ENUM_FLAG(9),
    Direct = ENUM_FLAG(10),     // Connection comes directly form adjacent/same layer
    Skip = ENUM_FLAG(11),       // Connection skips at least one layer
    Silent = ENUM_FLAG(12)      // learns but not actively used
  };

  //! Convenience function to combine multiple flags
  /*! Ensures logical OR is correctly used instead of boolean OR */
  inline int make_connection_flag( ConnectionFlags h ){ return h; }
  template <class ... Args> inline int make_connection_flag( ConnectionFlags h, Args ... t )
  {
    return h | make_connection_flag( t... );
  }

  //! Removes a specific flag from a set of flags, if it exists
  inline int strip_connection_flag( const int flags, const ConnectionFlags f )
  {
    if( (flags & f) == f )
      return flags ^ f;
    else
      return flags;
  }

  #undef ENUM_FLAG

  // ==========================================================================
  // ==========================================================================
  // ==========================================================================
  //! Represents an incoming connection from another neuron
  struct Input
  {
    //! Initial weight used for accumulation, larger values smooth initial transitions
    static constexpr double INITIAL_ACC_WEIGHT = 5.0;
    static constexpr double INITIAL_LATERAL_ACC_WEIGHT = 5.0; // DEBUG: change this back to 5.0 for v1

    Input() = default;

    //! Create an input tied to a specific neuron
    /*! The accumulator value will be populated automatically (legacy functionality)
        @param n The neuron this input comes from
        @param w The initial weight for the input
        @param cf The connection flags describing the input */
    Input( std::shared_ptr<Neuron> const & n, double w, int cf ) :
      Input( n, w, (cf & ConnectionFlags::Lateral) == ConnectionFlags::Lateral ? INITIAL_LATERAL_ACC_WEIGHT : INITIAL_ACC_WEIGHT , cf )
    {
      // If this neuron is inhibitory and we have negative weights, make the initial accumulator negative as well
      if( w < 0.0 && hasFlag( ConnectionFlags::Feedback ) )
        accumulatedWeight *= -1.0;
    }

    //! Create an input tied to a specific neuron
    /*! @param n The neuron this input comes from
        @param w The initial weight for the input
        @param acc The initial accumulator value for the input
        @param cf The connection flags describing the input */
    Input( std::shared_ptr<Neuron> const & n, double w, double acc, int cf ) :
      neuron( n ), weight( w ), ltWeight( w ), initialWeight( w ),
      accumulatedWeight( acc ),
      flags( cf ) {}

    std::shared_ptr<Neuron> neuron; //!< The neuron this input comes from
    double weight; //!< Weight of this connection (volatile)
    double ltWeight; //!< Long term, stable weight
    double initialWeight; //!< Initial weight of this connection
    double accumulatedWeight; //!< Activation and weight over time
    int flags; //!< Connection flags

    //! Check for a given connection flag or set of flags (all must be present)
    inline bool hasFlag(const int flag) const
    {
      return (flags & flag) == flag;
    }

    // When serialized, we need to run an external step to
    // properly re-build this afterwards
    template <class Archive>
    void serialize( Archive & ar )
    {
      //ar( CEREAL_NVP(neuron),
      ar( cereal::defer(neuron), // to prevent deep recursion, defer connections
          CEREAL_NVP(weight),
          CEREAL_NVP(ltWeight),
          CEREAL_NVP(initialWeight),
          CEREAL_NVP(accumulatedWeight),
          CEREAL_NVP(flags) );
    }
  };

  extern int preIndex;  //!< Marks current pre index for all NeuronVariables
  extern int postIndex; //!< Marks current post index for all NeuronVariables

  //!< Finalize results for all NeuronVariables (swap pre and post index)
  inline static void finalizeNeuronVariables() { std::swap( preIndex, postIndex ); }

  // ==========================================================================
  // ==========================================================================
  // ==========================================================================
  //! A NeuronVariable has a pre and post state which is updated each iteration
  /*! @tparam T The type to wrap */
  template <class T>
  struct NeuronVariable
  {
    NeuronVariable() = default;
    NeuronVariable( T const & val )
    {
      initialize( val );
    }

    inline void initialize( T const & value ){ pre() = post() = value; };
    inline void update( T const & value ){ post() = value; };
    inline void updateSmooth( T const & value, double const smoothing ){ post() = exponentialAvg( value, pre(), smoothing ); }

    //! Retrieve the pre-value
    inline T & pre(){ return values[preIndex]; }
    //! Retrieve the pre-value
    inline T const & pre() const { return values[preIndex]; }

    //! Retrieve the post-value
    inline T & post(){ return values[postIndex]; }
    //! Retrieve the post-value
    inline T const & post() const { return values[postIndex]; }

    T values[2]; //!< Holds value before finalize is called (pre, used for computation), as well
                 //!< as after finalized (updated values go here, post)

    template <class Archive>
    void serialize( Archive & ar )
    {
      ar( CEREAL_NVP(values) );
    }
  };

  // ==========================================================================
  // ==========================================================================
  // ==========================================================================
  //! Abstract neuron class
  class Neuron
  {
    public:
      Neuron();

      //! Create at some position
      Neuron( std::pair<double, double> pos );

      //! Copy from some other neuron, placing at some position
      Neuron( Neuron const & other, std::pair<double, double> pos );

      //! Equal if memory address is the same
      bool operator==( Neuron const & other ) const;
      bool operator!=( Neuron const & other ) const;

      //! Propogate inputs to activation
      void update( SystemState const & state );

      //! Adjust incoming weights
      void learn( SystemState const & state );

      //! Get the activation of this neuron
      virtual double activation( bool useSigmoid = false ) const;

      //! Get the activation for use in Sanger's rule
      /*! Sanger's rule must be applied externally so this gives
          access to itsActivation.post (the in process value);
          activation() normally returns the pre (finalized) activation */
      double activationSanger() const;

      //! Get the modulatory input of this neuron
      double modulatoryInput() const;

      //! Get x,y location of this neuron in its sheet
      std::pair<double, double> position() const;

      //! Get the x,y location of the centroid of learned connections
      /*! Instead of returning the location of the neuron, this instead computes
          a weighted centroid of the learned connection type.

          @return The weighted centroid. If no connections of the requested type exist,
                  the centroid will be negative. */
      std::pair<double, double> centroid( const int flags ) const;

      //! Checks if this neuron is part of a competitive column
      bool hasCompetitors();

      //! Add a connection from another neuron to this one
      void addInput( Input in );

      //! The incoming weights
      NeuronVariable<std::vector<Input>> itsInput;

      //! Dump weights to file
      void dumpWeights( std::string filename ) const;

      template <class Archive>
      void serialize( Archive & ar )
      {
        ar( CEREAL_NVP(itsInput),
            CEREAL_NVP(itsBCMThreshold),
            CEREAL_NVP(itsMaxThreshold),
            CEREAL_NVP(itsFastThreshold),
            CEREAL_NVP(itsActivationThreshold),
            CEREAL_NVP(itsDecayThreshold),
            CEREAL_NVP(itsGainMultiplier),
            CEREAL_NVP(itsModulatoryGainMultiplier),
            CEREAL_NVP(itsNormalizationCoefficients),
            CEREAL_NVP(itsUninhibitedActivation),
            CEREAL_NVP(itsInhibitorySum),
            CEREAL_NVP(itsLateralInhibitorySum),
            CEREAL_NVP(itsModulatoryInput),
            CEREAL_NVP(itsActivation),
            CEREAL_NVP(itsPosition) );

        if( Archive::is_loading::value )
          rebuildCache();
      }

      //! Prints key metadata for the neuron such as its position, activation, and thresholds
      void printNeuronInfo() const;

      //! Print all weights satisfying check to the given stream
      void printWeights( std::ostream & os, std::function<bool(Input const &)> && check ) const;

      //! Print connections satisfying the funtion check, which is called on every input neuron
      /*! Values printed will be those that gave rise to the current activation of the neuron,
          NOT its current inputs.
          @param heading Heading for this output
          @param check Boolean condition each neuron must satisfy to be printed
          @param verbose Whether to print every connection weight or just metadata */
      virtual void printConnections( std::string const & heading, std::function<bool(Input const &)> && check, bool verbose = true ) const;

      //! Does long term potentiation rebalancing based upon values in flagged connections
      /*! Slightly modified variant of hebbianNormalization for column connections
          @param normalizationCoefficient The total weight to distribute to all weights
          @param currentActivation The current activation of the neuron
          @param state The system state
          @param flags The flags to filter inputs */
      void columnNormalization( double normalizationCoefficient, double currentActivation, SystemState const & state, int flags );

      //! Does long term potentiation rebalancing based upon values in flagged connections
      /*! @param normalizationCoefficient The total weight to distribute to all weights
          @param currentActivation The current activation of the neuron
          @param state The system state
          @param flags The flags to filter inputs */
      void hebbianNormalization( double normalizationCoefficient, double currentActivation, SystemState const & state, int flags );

      //! Does short term weight rebalancing based upon values in flagged connections
      /*! @param normalizationCoefficient The total weight to distribute to all weights
          @param currentActivation The current activation of the neuron
          @param state The system state
          @param flags The flags to filter inputs */
      void shortTermHebbianNormalization( double normalizationCoefficient, double currentActivation, SystemState const & state, int flags );

      //! Updates lateral connections, excluding within-column weights
      /*! @param threshold The value to use for the hybrid threshold
          @param normalizationCoefficient The total weight to distribute to all weights
          @param currentActivation The current activation of the neuron
          @param state The system state
          @param flags The flags to filter inputs */
      void updateLateralWeights( SystemState const & state, const int flags,
                                 const double threshold, const double normalizationCoefficient, const double currentActivation,
                                 const double inhibitorySum, const double alpha, const double beta );

      //! Performs updates to short term weights (conflict learning)
      /*! Short term weight updates interpolate between an excitatory and inhibitory update based on the
          amount of incoming inhibition.

          The short term update is essentially: Inhib * beta * target + (1-Inhib) * alpha * target

          The inflection point (amount of inhibition necessary for a zero net update) is controlled by
          the alpha and beta parameters:
            Inhib = alpha / (alpha + beta)
                  = 1 / (1 + beta / alpha) [alternate form]
              or in terms of their ratio:
            beta / alpha = 1 / Inhib - 1
                         = (1 - Inhib) / Inhib [alternate form]

          The function calculateShortTermAlpha can automatically pick alpha such that full activation from the most learned
          inhibitory input in a column results in an inflection point.

          Idea is for around 50% of the max inhib you can receive to be the inflection point.

          @relates calculateShortTermAlpha

          @param state The system state
          @param flags The flags to filter inputs to perform learning over
          @param maxLearnedFlags The flags to filter which neurons are considered for "strongly learned" (kappa)
          @param currentActivation The current activation of the neuron
          @param inhibitorySum The inhibitory sum to use
          @param localLearningRate A modifier on the global learning rate to apply locally (as a product)
          @param ltmRate A modifier on the LTM accumulation rate to apply locally (as a product on weight update)
          @param alpha Coefficient on positive learning
          @param beta Coefficient on negative learning */
      void shortTermUpdate( SystemState const & state,
                            const int flags,
                            const int maxLearnedFlags,
                            const double currentActivation, const double inhibitorySum,
                            const double localLearningRate, const double ltmRate,
                            const double alpha, const double beta );

      //! Propogates changes made to accumulated weights to short and long term weights
      /*! This performs the short and long term updates for conflict learning (equations 11 and 12 from Neural Networks
          paper)

          @param state The system state
          @param flags The flags to filter inputs
          @param maxWeight The maximum weight to distribute amongst the filtered inputs (should include appropriate fb/ff/etc normalization term)
          @param stChaseFactor The short term smoothing factor, [0 1]. Lower values prefer short-term.
          @param ltChaseFactor The base long term smoothing factor, [0 1]. Lower values prefer long-term.
          @param ltSlodown An extra slowdown applied to LT weights moving away from distribution */
      void propogateAccumulatedWeight( SystemState const & state, const int flags,
          const double maxWeight,
          const double stChaseFactor = 0.05, const double ltChaseFactor = 0.05, const double ltSlowdown = 1.0 );

      //! Updates the activation thresholds for this neuron
      /*! Three thresholds control the activation of a neuron:
            1. The activation threshold (itsActivationThreshold) determines the amount of driving input required to initially turn on.
            2. The fast threshold (itsFastThreshold) determines the amount of total input to stay on, once the activation threshold is passed.
            3. The decay threshold (itsDecayThreshold) determines when to lower the activation threshold.

          Neurons do not increase thresholds when under the effects of inhibition.

          @param state The system state
          @param ltFeedforward The incoming feedforward (driving) input using long term weights
          @param feedforward The incoming feedforward (driving) input using short term weights
          @param inhibition The controlling inhibitory signal
          @param inhibitionThreshold The threshold for inhibition to pass
          @param fastSmoothing Fast threshold adjustment rate
          @param slowSmoothing Slow threshold adjustment rate
          @param decayRate Decay rate for decay threshold for inactive neurons */
      void updateThresholds( SystemState const & state,
          const double ltFeedforward, const double feedforward,
          const double inhibition, const double inhibitionThreshold,
          const double fastSmoothing = 0.1, const double slowSmoothing = 0.01, const double decayRate = 0.0001 );

      //! Updates the variables for gain control
      /*! Gain control strives to keep the neuron at a maximum activation of 1.0 even if its input exceeds this amount.

          Neurons are normalized in a local neighborhood (controlled by flags) based roughly on the maximum value seen. */
      void updateGainControls( const int flags );

      //! Does a weight update using BCM
      /*! This uses itsThreshold to store the BCM expected activation threshold */
      void bcmUpdate( double currentActivation, SystemState const & state, int flags );

    //protected:
      //! BCM activation threshold
      NeuronVariable<double> itsBCMThreshold;

      //! Variables controlling activation threshold
      double itsMaxThreshold, itsFastThreshold, itsActivationThreshold, itsDecayThreshold;

      //! Variables for gain control
      double itsGainMultiplier;
      NeuronVariable<double> itsModulatoryGainMultiplier;

      //! A struct holding coefficients for normalization due to disparities in number of connections
      struct NormalizationCoefficients
      {
        double column      = 1.0;
        double lateral     = 1.0;
        double feedback    = 1.0;
        double feedbackSkip = 1.0;
        double feedforward  = 1.0;
        double feedforwardSkip = 1.0;

        template <class Archive>
        void serialize( Archive & ar )
        {
          ar( column, lateral, feedback, feedbackSkip, feedforward, feedforwardSkip );
        }
      };

      //! The normalization coefficients for this neuron (not all may be used)
      NormalizationCoefficients itsNormalizationCoefficients;

      //! Holds the uninhibited activation (always pre-activation)
      /*! Used in the calculation of the inhibitory sum */
      NeuronVariable<double> itsUninhibitedActivation;

      //! Total sum of columnar inhibition
      double itsInhibitorySum;

      //! Total sum of lateral inhibition
      double itsLateralInhibitorySum;

      //! Marks all input caches as dirty
      void rebuildCache();

    protected:
      //! Provides the actual update implementation
      /*! @return The new activation (pre and post inhibition) after update is complete */
      virtual double updateImpl( SystemState const & state );

      //! Adjust incoming weights
      /*! @param currentActivation The current activation of this neuron */
      virtual void learnImpl( SystemState const & state, double currentActivation );

      //! Force activation to some value
      void setActivation( double activation, double noiseScale = 0.0 );

      //! Calculates the inhibitory sum for the specified inputs
      /*! Inhibition is summed over the specified inputs assuming that their
          activation exceeds that of this neuron.

          @param state The system state
          @param prevInhibition The previously calculated inhibitory sum
          @param flags The input type to compute over
          @return The inhibitory sum */
      double calculateInhibitorySum( SystemState const & state, const double prevInhibition, const int flags );

      // TODO: back to const and make cache volatile, make all cache access const

      //! Calculates the alpha parameter for short-term learning
      /*! The alpha parameter can be automatically set such that the inflection point
          for learning vs unlearning is set to the strongest possible source of inhibition
          for a neuron in a competitive column. At early stages of learning this is a low value
          which grows as column weights are learned and fine tuned.

          @param inhibProportion The proportion of the maximum inhibitory input that must be present
                                 to trigger the inflection point. Must be in the range (0, 1].
          @param maxInflection The maximum acceptable inflection point. Must be in the range (0, 1).
          @return The alpha necessary to achieve the required inflection point with beta set to 1.0 */
      double calculateShortTermAlpha( const double inhibProportion = 1.0, const double maxInflection = 0.5 );

      //! Computes the maximum learned activation for the specified input type
      /*! This implements the "spreading" component of conflict learning.

          Value returned is in the range [0 1] even if activation value exceeds 1.0.

          @param type The input type to search over
          @param learnThreshold The ratio of the maximum learned weight to be considered strongly learned
          @return The coefficient (kappa in paper) to be applied to learning */
      double computeMaxLearnedActivation( const int type, const double learnThreshold = 0.5 );

      //! Computes so-called "hybrid threshold", which is the activation value of the neuron
      //! above a given percentile of activation amongst the filtered neurons
      /*! See Multimap Formation in Visual Cortex, Jain et al. 2015
          @param percentile Percentile activation to serve as a threshold, e.g.: .5 finds the median
          @param The input types to compute over
          @return The activation of the first neuron that is a member of the percentile */
      double computeHybridThreshold( double percentile, const int flags ) const;

      //! Checks if a neuron passes the hybrid threshold
      /*! This is a convenience function that computes the hybrid threshold and then checks if
          the passed activation is sufficient to place the neuron in the same percentile of activation */
      bool passesHybridThreshold( double currentActivation, double percentile, const int flags ) const;

      //! Looks up the indices for the inputs corresponding to the flags
      /*! This lookup will be cached.
          @param flags The input type to compute over.
          @param f Whether this should be the pre or post cache */
      std::vector<size_t> const & getInputIndices( int flags );

      //! The input the neuron receives via modulation
      NeuronVariable<double> itsModulatoryInput;

    private:
      //! The activation
      NeuronVariable<double> itsActivation;

      //! The location of this neuron in its sheet (x,y)
      std::pair<double, double> itsPosition;

      //! Caches sorted subsets of the input
      std::unordered_map<int, std::pair<bool, std::vector<size_t>>> itsCache;
  };

  //! Sigmoid function
  double sigmoid( double value );

  //! Gaussian function
  double gaussian(double x, double mu, double sigma);

  //! Kumaraswamy function
  /*! @input x bounded [0 1]
      @output bounded [0 1] */
  double kumaraswamy(double x, double a = 2.5, double b = 3.0);

  //! Return a filtered vector of indices to the input type
  /*! This function will filter the provided input such that only those inputs satisfying the flags are passed */
  std::vector<size_t> filter_input( std::vector<bd::Input> & input, int flags );

  //! Return an observation vector (constant) of filtered references to the input type
  /*! This function will filter the provided input such that only those inputs satisfying the flags are passed */
  std::vector<std::reference_wrapper<const bd::Input>> observer( std::vector<bd::Input> const & input, int flags );
}

#endif // NEURON_H_
