#ifndef NEURON_ALLOCATOR_H_
#define NEURON_ALLOCATOR_H_

#include <memory>
#include <mutex>
#include <cstdint>
#include <vector>

//! Holds all of the shared state for all NeuronAllocators
struct NeuronAllocatorMemory
{
  static constexpr size_t itsMaxSize = 1024 * 1024 * 100; // 100MB
  static std::mutex itsMutex;
  static std::vector<std::uint8_t> itsMemory;
  static size_t itsLastIndex;
};

//! An allocator for neurons
/*! This allocator is designed to be used for shared_ptr to
    various neuron types.  It allocates a large block of memory in which
    to contiguously assign memory for the shared_ptrs that each neuron
    uses.

    Since we don't deallocate neurons until the program exit, this allocator
    does not actually free up memory.  It uses a static amount of memory */
template <class T>
class NeuronAllocator : public std::allocator<T>, protected NeuronAllocatorMemory
{
  public:
    typedef typename std::allocator<T>::size_type       size_type;
    typedef typename std::allocator<T>::difference_type difference_type;
    typedef typename std::allocator<T>::pointer         pointer;
    typedef typename std::allocator<T>::const_pointer   const_pointer;
    typedef typename std::allocator<T>::reference       reference;
    typedef typename std::allocator<T>::const_reference const_reference;
    typedef typename std::allocator<T>::value_type      value_type;

    template<class U>
      struct rebind{ typedef NeuronAllocator<U> other; };

    NeuronAllocator() noexcept;

    NeuronAllocator( NeuronAllocator const & other ) noexcept;

    template <class U>
      NeuronAllocator( NeuronAllocator<U> const & other ) noexcept;

    ~NeuronAllocator() noexcept;

    pointer allocate( size_type n, const void * = 0 );

    void deallocate( pointer p, size_type );
}; // class NeuronAllocator

#include "NeuronAllocatorImpl.H"
#endif // NEURON_ALLOCATOR_H_
