#include "ProtoObjectNeuron.H"
#include "Random.H"

#include <iostream>
#include <cassert>
#include <cereal/types/polymorphic.hpp>
#include <mutex>

CEREAL_REGISTER_TYPE(bd::ProtoObjectNeuron)

double bd::ProtoObjectNeuron::updateImpl( SystemState const & state )
{
  const double noise = normalReal()() * state.noiseScale;

  double feedforward = 0;
  double ltFeedforward = 0;
  double feedforwardSkip = 0;
  double ltFeedforwardSkip = 0;
  double feedback_e = 0, feedback_i = 0;
  double lateralExcitation = 0;

  auto const & preInput = itsInput.pre();

  // Excitation
  for( auto const & preEdge : preInput )
  {
    const double incomingActivation = preEdge.neuron->activation();
    double currentInput = preEdge.weight * incomingActivation;
    const double ltCurrentInput = preEdge.ltWeight * incomingActivation;

    if( preEdge.hasFlag( ConnectionFlags::Silent ) )
      continue;

    if( preEdge.hasFlag( ConnectionFlags::Feedforward ) )
    {
      if( preEdge.hasFlag( ConnectionFlags::Skip ) && !state.dy_skip_feedforward )
        continue;
      if( preEdge.hasFlag( ConnectionFlags::Direct ) && !state.dy_grouping_to_proto_feedforward )
        continue;
      //if( preEdge.hasFlag( ConnectionFlags::Skip ) )
      //if( preEdge.hasFlag( ConnectionFlags::Direct ) )
      {
        feedforward += currentInput;
        ltFeedforward += ltCurrentInput;
      }
      //else
      //{
      //  feedforwardSkip += currentInput;
      //  ltFeedforwardSkip += ltCurrentInput;
      //}
    }
    else if( preEdge.hasFlag( ConnectionFlags::Lateral | ConnectionFlags::Excitatory ) )
      lateralExcitation += currentInput;
  }

  // Ambiguity and feedback
  const double feedback = std::max( feedback_e + feedback_i, 0.0 );
  const double ambiguity = state.cl_ambiguity ? std::min( feedback_e, -feedback_i ) : 0.0;

  // Calculates LT feedforward, properly gating the skip connection
  auto calculateLTFeedforward = []( double ff, double skipFF )
  {
    const double clampedFF = clamp(ff);
    const double ffGating = clampedFF * clampedFF;

    return ff + ffGating * skipFF;
  };

  // Calculates activation value before threshold and min/max clamping
  auto calculatePreClamp = [=]( double ff, double skipFF, double lateral, double fb, double inhib, double noise )
  {
    const double clampedFF = clamp(ff);
    const double ffGating = clampedFF * clampedFF;

    const double input = ff + (skipFF * ffGating) + (lateral * ffGating) + (fb * ffGating) + noise;

    double result = input / ( 1.0 + inhib );

    if( result <= 0 )
      return 0.0;

    result = std::min( result, 1.0 ) + std::max( 0.0, std::log10( result ) );
    return result / itsGainMultiplier;
  };

  // Inhibition
  itsInhibitorySum = calculateInhibitorySum( state, itsInhibitorySum, ConnectionFlags::Column );
  itsLateralInhibitorySum = calculateInhibitorySum( state, itsLateralInhibitorySum, ConnectionFlags::Lateral );

  updateGainControls( ConnectionFlags::Column );

  // Activation
  itsUninhibitedActivation.update( calculatePreClamp( feedforward, feedforwardSkip, lateralExcitation, feedback, ambiguity, noise ) );
  double preClamp = calculatePreClamp( feedforward, feedforwardSkip, lateralExcitation, feedback, itsInhibitorySum + itsLateralInhibitorySum + ambiguity, noise );

  updateThresholds( state, calculateLTFeedforward( ltFeedforward, ltFeedforwardSkip ), preClamp,
                    itsInhibitorySum,
                    0.4 * itsNormalizationCoefficients.column, // inhibitory threshold
                    0.01,       // fast
                    0.0025,     // slow
                    0.0001 );   // decay

  const double inverse = 1.0 / itsNormalizationCoefficients.feedforward;
  if( feedforward < itsFastThreshold * inverse )
  {
    preClamp = 0.0;
    itsUninhibitedActivation.update( 0.0 );
  }

  // Compute final output
  const double value = clamp( preClamp, 0.0, state.maxActivation );

  return value;
}

void bd::ProtoObjectNeuron::learnImpl( SystemState const & state, double currentActivation )
{
  if( state.learningRule == LearningRule::Hebbian )
  {
    shortTermHebbianNormalization( state.poMaxWeights.feedforward * itsNormalizationCoefficients.feedforward,
                                   currentActivation, state,
                                   ConnectionFlags::Dynamic | ConnectionFlags::Feedforward );
    return;
  }
  else if( state.learningRule == LearningRule::Conflict )
  {
    const double alpha = calculateShortTermAlpha(); // excitation
    const double beta = 1.0; // inhibition

    const int kappaFlagsFeedforward = ConnectionFlags::Feedforward | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

    // Feedforward Update from Grouping Neurons (Direct)
    {
      const int flags = ConnectionFlags::Direct | ConnectionFlags::Feedforward | ConnectionFlags::Dynamic;

      // Short term (skip if not on)
      if( currentActivation != 0.0 )
      {
        constexpr double localLearningRate = 0.01; // feedforward needs to be slower moving
        constexpr double ltmRate = 1.0; // how much faster LTM should adjust

        shortTermUpdate( state, flags, kappaFlagsFeedforward,
                         currentActivation, itsLateralInhibitorySum,
                         localLearningRate, ltmRate,
                         alpha, beta );
      }

      // Long term
      propogateAccumulatedWeight( state, flags,
          state.poMaxWeights.feedforward * itsNormalizationCoefficients.feedforward,
          0.5, // st
          0.05, // lt
          100.0 ); // slowdown factor for incorrect growth
    }

    // Feedforward Update from BO neurons (Skip)
    {
      const int flags = ConnectionFlags::Skip | ConnectionFlags::Feedforward | ConnectionFlags::Dynamic;

      // Short term (skip if not on)
      if( currentActivation != 0.0 )
      {
        constexpr double localLearningRate = 0.01; // feedforward needs to be slower moving
        constexpr double ltmRate = 1.0; // how much faster LTM should adjust

        shortTermUpdate( state, flags, kappaFlagsFeedforward,
                         currentActivation, itsLateralInhibitorySum,
                         localLearningRate, ltmRate,
                         alpha, beta );
      }

      // Long term
      propogateAccumulatedWeight( state, flags,
          state.poMaxWeights.feedforwardSkip * itsNormalizationCoefficients.feedforwardSkip,
          0.5,
          0.05,
          100.0 );
    }

    // Lateral Update
    //{
    //  const int flags = ConnectionFlags::Lateral | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic;

    //  // Short term (skip if not on)
    //  if( currentActivation != 0.0 )
    //  {
    //    constexpr double localLearningRate = 1.0;
    //    constexpr double ltmRate = 100.0; // how much faster LTM should adjust
    //    constexpr double alpha = 0.2;
    //    constexpr double beta  = 1.0;

    //    shortTermUpdate( state, flags, flags,
    //                     currentActivation, itsLateralInhibitorySum,
    //                     localLearningRate, ltmRate,
    //                     alpha, beta );
    //  }

    //  // Long term
    //  propogateAccumulatedWeight( state, ConnectionFlags::Lateral | ConnectionFlags::Excitatory | ConnectionFlags::Dynamic,
    //      state.poMaxWeights.lateral * itsNormalizationCoefficients.lateral,
    //      0.05,
    //      0.05,
    //      10.0 );
    //}
  } // end Conflict learning

  // Column connections learned using a modified LTM hebbian
  columnNormalization( state.poMaxWeights.column,
                       currentActivation,
                       state,
                       ConnectionFlags::Dynamic | ConnectionFlags::Column | ConnectionFlags::Inhibitory );

  //updateLateralWeights( 0.8, // hybrid threshold
  //                      state.gnLateralTotalWeight, currentActivation, state,
  //                      ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );
}
