#ifndef SHEET_H_
#define SHEET_H_

#include <vector>
#include <cstddef>
#include <opencv2/core/core.hpp>
#include <cereal/types/vector.hpp>
#include <cereal/types/polymorphic.hpp>

#include "NeuronAllocator.H"

namespace bd
{
  //! Controls how sheets are populated with neurons
  enum class PopulationStrategy
  {
    Grid,        //!< Uniform grid (non-random)
    Uniform,     //!< Uniform random
    PoissonDisk, //!< PoissonDisk distrubtion (maintain a minimum spacing)
    Empty        //!< No neurons
  };

  //! Controls how columns are constructed
  enum class CompetitionStrategy
  {
    Uniform,   //!< All columns have the same number of neurons
    Population //!< Columns have an indeterminate number of neurons controlled by
               //!<   where they end up due to the population strategy
  };

  //! Represents a "sheet" of a neuron type
  /*! A sheet of neurons is a discretized grid where each entry in the grid
      can contain any number of neurons.  Neurons hold onto a continuous position
      but are binned into the grid for fast retrieval. */
  template <class NeuronT>
  class Sheet
  {
    private:
      using NeuronGrid = std::vector<std::vector<std::shared_ptr<NeuronT>>>;
      using SheetIter = typename NeuronGrid::iterator;
      using SheetConstIter = typename NeuronGrid::const_iterator;

    public:
      Sheet() = default;
      //! Initialize with a uniform grid
      Sheet( size_t width, size_t height );
      //! Initialize with some given strategy
      Sheet( size_t width, size_t height, size_t count, PopulationStrategy ps );
      //! Initialize by copying some other sheet
      /*! Copies all neurons in the other sheet - should ideally be a sheet with no
          connections */
      Sheet( Sheet const & other );

      //! Trims away all neurons within a given distance of the boundaries
      /*! Useful for removing neurons in higher level sheets that may otherwise
          suffer from adverse effects from the experimental setup (boundaries) */
      void trimBoundary( size_t radius );

      size_t width() const;
      size_t height() const;
      size_t size() const;  //!< Total size of grid
      size_t count() const; //!< Number of neurons

      std::vector<NeuronT> & operator()(int x, int y);
      std::vector<NeuronT> & operator()(int x, int y) const;

      std::vector<std::shared_ptr<NeuronT>> & operator[](size_t index);
      std::vector<std::shared_ptr<NeuronT>> const & operator[](size_t index) const;

      //! Return an iterator over grid locations
      SheetIter begin();
      //! Return an iterator over grid locations
      SheetConstIter begin() const;

      //! Return an end iterator over grid locations
      SheetIter end();
      //! Return an end iterator over grid locations
      SheetConstIter end() const;

      //! Apply the given function to every neuron in the sheet
      template <class F> inline
      void for_each_neuron( F && func );

      void addNeuron( NeuronT const & neuron );

      template <class U>
      friend void windowCallback( int event, int mX, int mY, int flags, void* userdata );

      template <class Archive>
      void serialize( Archive & ar )
      {
        ar( CEREAL_NVP(itsNeurons),
            CEREAL_NVP(itsWidth),
            CEREAL_NVP(itsHeight),
            CEREAL_NVP(itsCount) );
      }

    protected:
      void addNeuron( size_t index, std::pair<double, double> pos );
      void addNeuronImpl( size_t index, std::pair<double, double> pos, std::true_type );
      void addNeuronImpl( size_t index, std::pair<double, double> pos, std::false_type );

      //! Removes all neurons at the specified grid index
      void removeNeuronsAtIndex( size_t index );

      static NeuronAllocator<NeuronT> itsAllocator;

      NeuronGrid itsNeurons;
      size_t itsWidth;
      size_t itsHeight;
      size_t itsCount;
  };

  template <class NeuronT>
  void windowCallback( int event, int mX, int mY, int flags, void* userdata );

  //! Combine two sheets together
  /*! This will merge two sheets together into a new sheet with the size of the largest
      of the two provided sheets.  One sheet must be strictly larger or equal
      to the other */
  template <class NeuronT>
  Sheet<NeuronT> mergeSheets( Sheet<NeuronT> const & s1, Sheet<NeuronT> const & s2 );
}

#include "SheetImpl.H"

#endif // SHEET_H_
