#include <tuple>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include "BorderOwnershipNeuron.H"
#include "Display.H"
#include "Random.H"

#ifndef USE_CUSTOM_ALLOCATOR
#define USE_CUSTOM_ALLOCATOR 1
#endif // USE_CUSTOM_ALLOCATOR

template <class NeuronT>
NeuronAllocator<NeuronT> bd::Sheet<NeuronT>::itsAllocator;

template <class NeuronT>
bd::Sheet<NeuronT>::Sheet( size_t width, size_t height ) : Sheet( width, height, width * height, PopulationStrategy::Grid )
{ }

template <class NeuronT>
bd::Sheet<NeuronT>::Sheet( size_t width, size_t height, size_t count, PopulationStrategy ps ) :
  itsNeurons( width * height ),
  itsWidth( width ),
  itsHeight( height ),
  itsCount( 0 )
{
  auto indToXY = [width](size_t index)
  { return std::make_pair( index % width, index / width); };

  auto XYtoInd = [width](size_t x, size_t y)
  { return y * width + x; };

  auto rng = uniformInt( 0lu, itsNeurons.size() - 1 );

  switch( ps )
  {
    case PopulationStrategy::Grid:
    {
      for( size_t i = 0; i < width * height; ++i )
        addNeuron( i, indToXY( i ) );
      break;
    }
    case PopulationStrategy::Uniform:
    {
      for( size_t i = 0; i < count; ++i )
      {
        const auto index = rng();
        addNeuron( index, indToXY( index ) );
      }
      break;
    }
    case PopulationStrategy::PoissonDisk:
    {
      const size_t k = 30; // maximum number of samples before rejection
      const double magic_number = 0.67; // used to solve the relationship
                                        // between radius, area, and count:
                                        // count = magic_number * area / r^2
                                        // determined via experimentation
                                        // NOTE: relationship turns out to not
                                        // actually be quite linear, this number
                                        // is not perfect
      const double radius = std::sqrt( magic_number * width * height / count );
      const double radius2 = radius * radius;

      // Initial location is random
      double x, y;
      std::tie( x, y ) = indToXY( rng() );

      std::vector<std::pair<double, double>> activeList = {{x, y}};
      addNeuron( XYtoInd( x, y ), {x, y} );

      // return a random index from the active list
      auto getRandomActive = [&activeList]()
      { return uniformInt( 0lu, activeList.size() - 1 )(); };

      auto rng2PI = uniformReal( 0.0, 2.0 * M_PI );
      auto rngRadius = uniformReal( 0.0, radius );

      // return a random x and y offset to use
      auto getRandomOffsets = [&]()
      {
        const double a = rng2PI();
        const double r = std::sqrt( rngRadius() + radius2 );

        return std::make_pair( r * std::cos( a ),
                               r * std::sin( a ) );
      };

      // Check if a given x,y is far enough away from other neurons
      // in the sheet
      auto far = [=]( double x, double y )
      {
        double i = x;
        double j = y;
        // create bounding box around current location
        const size_t j0 = std::max( i - radius, 0.0 );
        const size_t i0 = std::max( j - radius, 0.0 );
        const size_t j1 = std::min( i + radius, (double)width - 1 );
        const size_t i1 = std::min( j + radius, (double)height - 1 );

        // Loop over box around current location
        for( j = j0; j <= j1; ++j )
        {
          for( i = i0; i <= i1; ++i )
          {
            // Check if current cell has any neurons in it
            const auto & cell = itsNeurons[XYtoInd( j, i )];
            if( !cell.empty() )
            {
              // Check against all neurons in current cell
              for( const auto & neuron : cell )
              {
                // position of already inserted neuron
                const auto pos = neuron->position();

                const double dx = pos.first  - x;
                const double dy = pos.second - y;
                const double dist2 = dx * dx + dy * dy;

                if( dist2 < radius2 )
                  return false;
              }
            }
          }
        }

        return true;
      };

      // While activeList is not empty, search around it up to K
      // times randomly trying new neurons
      while( !activeList.empty() )
      {
        size_t current = getRandomActive();
        bool found = false;

        // Make a new candidate between [radius, 2*radius] from existing sample
        for( size_t i = 0; i < k; ++i )
        {
          const auto offsets = getRandomOffsets();
          x = activeList[current].first + offsets.first;
          y = activeList[current].second + offsets.second;

          // Reject candidates that are outside the allowed extent,
          // or closer than 2 * radius to any existing sample
          if( 0 <= x && x < width && 0 <= y && y < height && far( x, y ) )
          {
            // we found something!
            found = true;
            addNeuron( XYtoInd(x, y), {x, y} );
            activeList.emplace_back( x, y );

            break;
          }
        }

        if( !found )
          activeList.erase( activeList.begin() + current );
      }

      std::cout << "Poisson initialized with " << itsCount << " neurons (target was " << count << ")" << std::endl;

      break;
    } // end poisson disk
    case PopulationStrategy::Empty:
    {
      break;
    } // end empty
  } // end switch
}

template <class NeuronT> inline
bd::Sheet<NeuronT>::Sheet( Sheet const & other ) :
  itsNeurons( other.itsNeurons.size() ),
  itsWidth( other.itsWidth ),
  itsHeight( other.itsHeight ),
  itsCount( 0 )
{
  for( const auto & cell : other )
    for( auto const & neuron : cell )
      addNeuron( *neuron );
}

template <class NeuronT> inline
void bd::Sheet<NeuronT>::trimBoundary( size_t radius )
{
  auto indToXY = [this](size_t index)
  { return std::make_pair( index % itsWidth, index / itsWidth); };

  for( size_t i = 0; i < itsWidth * itsHeight; ++i )
  {
    double x, y;
    std::tie( x, y ) = indToXY( i );
    if( x < radius ||
        x >= itsWidth - radius ||
        y < radius ||
        y >= itsHeight - radius )
      removeNeuronsAtIndex( i );
  }
}

template <class NeuronT> inline
size_t bd::Sheet<NeuronT>::width() const
{
  return itsWidth;
}

template <class NeuronT> inline
size_t bd::Sheet<NeuronT>::height() const
{
  return itsHeight;
}

template <class NeuronT> inline
size_t bd::Sheet<NeuronT>::size() const
{
  return itsNeurons.size();
}

template <class NeuronT> inline
std::vector<NeuronT> & bd::Sheet<NeuronT>::operator()(int x, int y)
{
  return itsNeurons[y * itsWidth + x];
}

template <class NeuronT> inline
std::vector<NeuronT> & bd::Sheet<NeuronT>::operator()(int x, int y) const
{
  return itsNeurons[y * itsWidth + x];
}

template <class NeuronT> inline
std::vector<std::shared_ptr<NeuronT>> & bd::Sheet<NeuronT>::operator[](size_t index)
{
  return itsNeurons[index];
}

template <class NeuronT> inline
std::vector<std::shared_ptr<NeuronT>> const & bd::Sheet<NeuronT>::operator[](size_t index) const
{
  return itsNeurons[index];
}

template <class NeuronT> inline
auto bd::Sheet<NeuronT>::begin() -> SheetIter
{
  return itsNeurons.begin();
}

template <class NeuronT> inline
auto bd::Sheet<NeuronT>::begin() const -> SheetConstIter
{
  return itsNeurons.begin();
}

template <class NeuronT> inline
auto bd::Sheet<NeuronT>::end() -> SheetIter
{
  return itsNeurons.end();
}

template <class NeuronT> inline
auto bd::Sheet<NeuronT>::end() const -> SheetConstIter
{
  return itsNeurons.end();
}

template <class NeuronT>
template <class Func> inline
void bd::Sheet<NeuronT>::for_each_neuron( Func && f )
{
  for( const auto & cell : *this )
    for( const auto & neuron : cell )
      f( neuron );
}

template <class NeuronT> inline
void bd::Sheet<NeuronT>::addNeuron( NeuronT const & neuron )
{
  int x, y;
  std::tie(x, y) = neuron.position();
#ifdef USE_CUSTOM_ALLOCATOR
  itsNeurons[y * itsWidth + x].push_back( std::allocate_shared<NeuronT>( itsAllocator, neuron ) );
#else // (not) USE_CUSTOM_ALLOCATOR
  itsNeurons[y * itsWidth + x].push_back( std::make_shared<NeuronT>( neuron ) );
#endif // USE_CUSTOM_ALLOCATOR

  ++itsCount;
}

template <class NeuronT> inline
void bd::Sheet<NeuronT>::addNeuron( size_t index, std::pair<double, double> pos )
{
  addNeuronImpl( index, pos, std::is_same<NeuronT, BorderOwnershipNeuron>() );
  ++itsCount;
}

// special case for border ownership neuron
template<class NeuronT> inline
void bd::Sheet<NeuronT>::addNeuronImpl( size_t index, std::pair<double, double> pos, std::true_type )
{
  // create neurons
#ifdef USE_CUSTOM_ALLOCATOR
  itsNeurons[index].push_back( std::allocate_shared<NeuronT>( itsAllocator, pos ) );
  itsNeurons[index].push_back( std::allocate_shared<NeuronT>( itsAllocator, pos ) );
#else // (not) USE_CUSTOM_ALLOCATOR
  itsNeurons[index].push_back( std::make_shared<NeuronT>( pos ) );
  itsNeurons[index].push_back( std::make_shared<NeuronT>( pos ) );
#endif // USE_CUSTOM_ALLOCATOR

  //wireEnemies( *(itsNeurons[index].rbegin()), *(itsNeurons[index].rbegin() + 1) );
}

template<class NeuronT> inline
void bd::Sheet<NeuronT>::addNeuronImpl( size_t index, std::pair<double, double> pos, std::false_type )
{
  //itsNeurons[index].push_back( std::make_shared<NeuronT>( pos ) );
  itsNeurons[index].push_back( std::allocate_shared<NeuronT>( itsAllocator, pos ) );
}

template<class NeuronT> inline
void bd::Sheet<NeuronT>::removeNeuronsAtIndex( size_t index )
{
  itsNeurons[index].clear();
}

template <class NeuronT>
bd::Sheet<NeuronT> bd::mergeSheets( Sheet<NeuronT> const & s1, Sheet<NeuronT> const & s2 )
{
  Sheet<NeuronT> sheet = s1.size() > s2.size() ? s1 : s2;
  Sheet<NeuronT> const & other = s1.size() <= s2.size() ? s1 : s2;

  double widthScaling = sheet.width() / other.width();
  double heightScaling = sheet.height() / other.height();

  for( const auto & cell : other )
    for( auto const & neuron : cell )
    {
      double x, y;
      std::tie(x, y) = neuron->position();
      NeuronT scaledNeuron( *neuron, { x * widthScaling, y * heightScaling } );
      sheet.addNeuron( scaledNeuron );
    }

  return sheet;
}
