#ifndef STATE_H_
#define STATE_H_

#include <cereal/cereal.hpp>
#include <cstddef>

namespace bd
{
  //! The learning rule to use for plasticity
  enum class LearningRule
  {
    Conflict,
    Hebbian,
    BCM,
    Sanger
  };

  //! Holds the maximum weight for different connection types for a neuron layer
  struct MaxWeights
  {
    double column = 1.0;
    double lateral = 1.0, lateralInhibitory = 1.0;
    double feedback = 1.0, feedbackInhibitory = 1.0;
    double feedbackSkip = 1.0, feedbackSkipInhibitory = 1.0;
    double feedforward = 1.0;
    double feedforwardSkip = 1.0;

    template <class Archive>
    void serialize( Archive & ar )
    {
      ar( column, lateral, lateralInhibitory,
          feedback, feedbackInhibitory,
          feedbackSkip, feedbackSkipInhibitory,
          feedforward, feedforwardSkip );
    }
  };

  //! The state of the system (overall, persistent)
  /*! This also includes some parameters that can't easily be
      passed to neurons */
  class SystemState
  {
    public:
      SystemState() :
        learningRate( 0.1 ),
        //groupingTotalWeight( groupingMaxWeight ),
        learningRule( LearningRule::Conflict ),
        useLateralConnections( true ),
        // V1 parameters
        v1MaxFFWeight( 4.0 ),
        v1lateralTotalWeight( 2.0 ),
        noiseScale( 0.01 ),
        staticNoiseScale( 0.0 ),
        // Global parameters
        maxActivation( 4.0 ),
        maxWeight( 1.0 ),
        // CL learning rules
        cl_unlearning( true ),
        cl_spreading( true ),
        cl_slt( true ),
        cl_ambiguity( true ),
        // Feedback/feedforward control
        dy_skip_feedback( true ),
        dy_skip_feedforward( true ),
        dy_grouping_to_proto_feedforward( true ),
        dy_proto_to_grouping_feedback( true ),
        dy_bo_to_grouping_feedforward( true ),
        dy_grouping_to_bo_feedback( true )
      { }

    public:
      //! Learning rate modifier for all neurons
      double learningRate;

      //! The total weight normalized across a single
      //! grouping neuron
      //double groupingTotalWeight;

      bool debugPrint;

      //! The learning rule to apply
      LearningRule learningRule;

      //! Whether lateral connections should be used
      /*! Turning off lateral connections is mainly used for visualization in V1 */
      bool useLateralConnections;

      MaxWeights boMaxWeights; //!< Highest weight distributed amongst border ownership neurons
      MaxWeights gnMaxWeights; //!< Highest weight distributed amongst grouping neurons
      MaxWeights poMaxWeights; //!< Highest weight distributed amongst proto object neurons

      //static double const groupingMaxWeight;   //!< Highest weight distributed amongst FF connections (per Grouping neuron)
      //static double const groupingMinWeight;   //!< Lowest weight distributed amongst FF connections (per Grouping neuron)

      //static double feedbackTotalWeight; //!< Total weight distributed amongst FB connections (per BO neuron)
      //static double feedbackInhibitoryTotalWeight; //!< Total weight distributed amongst FB inhibitory connections (per BO neuron)

      //// Border ownership lateral parameters
      //static double const boColumnTotalWeight; //!< Total weight distributed amongst lateral inhibitory connections (per BO column)
      //static double const boLateralTotalWeight; //!< Total weight distributed amongst lateral connections (excludes column)

      //static double const gnColumnTotalWeight; //!< Total weight distributed amongst lateral inhibitory connections (per Grouping column)
      //static double const gnLateralTotalWeight; //!< Total weight distributed amongst lateral connections (excludes column)
      //static double const gnSkipTotalWeight; //!< Total weight distributed amongst feedforward (skip) connections

      double v1MaxFFWeight;
      double v1lateralTotalWeight;

      double noiseScale = 0.01;
      double staticNoiseScale = 0.0;

      double maxActivation; //!< Max any neuron can activate
      double maxWeight; //!< Max any neuron connection can weigh

      //! Used to turn off specific portions of the conflict learning for testing
      bool cl_unlearning;
      bool cl_spreading;
      bool cl_slt;
      bool cl_ambiguity;

      //! Used to turn off specific portions of neuron dynamics for testing
      bool dy_skip_feedback;
      bool dy_skip_feedforward;
      bool dy_grouping_to_proto_feedforward;
      bool dy_proto_to_grouping_feedback;
      bool dy_bo_to_grouping_feedforward;
      bool dy_grouping_to_bo_feedback;

      template <class Archive>
      void serialize( Archive & ar )
      {
        ar( CEREAL_NVP(learningRate),
            //CEREAL_NVP(groupingTotalWeight),
            CEREAL_NVP(learningRule),
            CEREAL_NVP(useLateralConnections),
            CEREAL_NVP(boMaxWeights),
            CEREAL_NVP(gnMaxWeights),
            CEREAL_NVP(poMaxWeights),
            CEREAL_NVP(noiseScale),
            CEREAL_NVP(staticNoiseScale),
            CEREAL_NVP(maxActivation),
            CEREAL_NVP(maxWeight) );
      }
  };
}
#endif // STATE_H_
