#include "V1.H"

#include "Wiring.H"
#include "Util.H"

#include <fstream>
#include <cereal/archives/binary.hpp>
#include <cereal/types/vector.hpp>

#include <opencv2/imgproc/imgproc.hpp>

#include "ThreadingUtil.H"

//#define V1_SINGLE_NEURON

bd::V1::V1( size_t gridWidth,
            size_t gridHeight,
            SystemState const & state,
            Options const & Options ) :
  itsState( state ),
  itsGridWidth( gridWidth ),
  itsGridHeight( gridHeight ),
  itsOnOffResponses( gridWidth, gridHeight, gridHeight * gridWidth * 2.0, PopulationStrategy::PoissonDisk ),
  itsOffOnResponses( itsOnOffResponses ),
#ifdef V1_SINGLE_NEURON
  itsEdgeResponses( gridWidth, gridHeight, 0, PopulationStrategy::Empty ),
#else
  itsEdgeResponses( gridWidth, gridHeight, gridHeight * gridWidth * 8.0, PopulationStrategy::PoissonDisk ),
  //itsEdgeResponses( gridWidth, gridHeight )
#endif
  itsOptions( Options )
{
  const double radius = 0;
  const double spread = 2.5;
  const double lateralRadiusCoefficient = 1.0;
  const double probability = 1.0;

  #ifdef V1_SINGLE_NEURON
  // Single neuron version
  //  Create 4 edge neurons at the same location with the same receptive field + inhibition
  itsEdgeResponses.addNeuron( {{gridWidth/2.0 + .007, gridHeight/2.0}} );
  itsEdgeResponses.addNeuron( {{gridWidth/2.0 - .007, gridHeight/2.0}} );
  itsEdgeResponses.addNeuron( {{gridWidth/2.0, gridHeight/2.0 + .007}} );
  itsEdgeResponses.addNeuron( {{gridWidth/2.0, gridHeight/2.0 - .007}} );

  itsEdgeResponses.addNeuron( {{gridWidth/2.0 + .005, gridHeight/2.0 + .005}} );
  itsEdgeResponses.addNeuron( {{gridWidth/2.0 - .005, gridHeight/2.0 - .005}} );
  itsEdgeResponses.addNeuron( {{gridWidth/2.0 - .005, gridHeight/2.0 + .005}} );
  itsEdgeResponses.addNeuron( {{gridWidth/2.0 + .005, gridHeight/2.0 - .005}} );

  wireSheetRingUniform( itsEdgeResponses, itsOnOffResponses,
      radius, spread, probability, 0.5, true,
      ConnectionFlags::Dynamic | ConnectionFlags::Feedforward );

  wireSheetRingUniform( itsEdgeResponses, itsOffOnResponses,
      radius, spread, probability, 0.5, true,
      ConnectionFlags::Dynamic | ConnectionFlags::Feedforward | ConnectionFlags::DisplayAsInhibitory );

  wireSheetRingUniform( itsEdgeResponses, itsEdgeResponses, 0, spread * 0.85, probability, -itsState.v1lateralTotalWeight, true,
                        ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );

  //wireSheetRingUniform( itsEdgeResponses, itsEdgeResponses, 0, spread * 0.37, probability, 1.2 * itsState.v1lateralTotalWeight, true, ConnectionFlags::Static | ConnectionFlags::Lateral );
  #else // not V1_SINGLE_NEURON
  // Normal version

  // Feedforward connections
  wireSheetRingUniform( itsEdgeResponses, itsOnOffResponses,
      radius, spread, probability, 1.0 / 2.0, true,
      ConnectionFlags::Dynamic | ConnectionFlags::Feedforward );

  wireSheetRingUniform( itsEdgeResponses, itsOffOnResponses,
      radius, spread, probability, 1.0 / 2.0, true,
      ConnectionFlags::Dynamic | ConnectionFlags::Feedforward | ConnectionFlags::DisplayAsInhibitory );

  const double excitationRadius = itsOptions.excitationRadius * spread * lateralRadiusCoefficient;
  const double inhibitionRadius = itsOptions.inhibitionRadius * spread * lateralRadiusCoefficient;
  const double excitationSpread = itsOptions.excitationSpread * spread * lateralRadiusCoefficient;
  const double inhibitionSpread = itsOptions.inhibitionSpread * spread * lateralRadiusCoefficient;

  // Inhibitory lateral connections, initially equal weight
  if( state.learningRule == LearningRule::Conflict )
  {
    if( itsOptions.useGaussianRing )
      wireSheetRing( itsEdgeResponses, itsEdgeResponses,
                     inhibitionRadius, inhibitionSpread, probability, -itsOptions.inhibitionStrength,
                     ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );
    else
      wireSheetRingUniform( itsEdgeResponses, itsEdgeResponses,
                            inhibitionRadius, inhibitionSpread, probability, -itsOptions.inhibitionStrength, true,
                            ConnectionFlags::Dynamic | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );
  }
  else if( state.learningRule == LearningRule::Hebbian )
  {
    if( itsOptions.useGaussianRing )
      wireSheetRing( itsEdgeResponses, itsEdgeResponses,
                     inhibitionRadius, inhibitionSpread, probability, -itsOptions.inhibitionStrength,
                     ConnectionFlags::Static | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );
    else
      wireSheetRingUniform( itsEdgeResponses, itsEdgeResponses,
                            inhibitionRadius, inhibitionSpread, probability, -itsOptions.inhibitionStrength, false,
                            ConnectionFlags::Static | ConnectionFlags::Lateral | ConnectionFlags::Inhibitory );
  }

  // Normalize connections
  calculateV1Coefficients( itsEdgeResponses, itsState );

  // Excitatory lateral connections
  if( itsOptions.useGaussianRing )
    wireSheetRing( itsEdgeResponses, itsEdgeResponses,
                   excitationRadius, excitationSpread, probability, itsOptions.excitationStrength,
                   ConnectionFlags::Static | ConnectionFlags::Lateral );
  else
    //                            to             from,   radius, spread,        probability, scaling, flags
    wireSheetRingUniform( itsEdgeResponses, itsEdgeResponses,
                          excitationRadius, excitationSpread, probability, itsOptions.excitationStrength, true,
                          ConnectionFlags::Static | ConnectionFlags::Lateral );

  // Normalize lateral excitatory connections
  normalizeV1LateralExcitatoryConnections( itsEdgeResponses, itsState );
  #endif // not V1_SINGLE_NEURON
}

void bd::V1::onset( cv::Mat const & image )
{
  const int imgW = image.cols;
  const int imgH = image.rows;

  assert( imgW >= (int)itsGridWidth );
  assert( imgH >= (int)itsGridHeight );

  const double xScaleFactor = imgW / itsGridWidth;
  const double yScaleFactor = imgH / itsGridHeight;
  const double alpha = std::max( xScaleFactor, yScaleFactor );

  // alpha is proportional to scale factor
  // kernels have a 1.2:1 std ratio
  // outer:
  //  alpha * 3, s=.3333*alpha
  // inner:
  //  alpha * 3, s=0.4*alpha (negative)

  const int kernelSzTemp = static_cast<int>( alpha * 3.0 );
  const int kernelSz = (kernelSzTemp % 2 == 0) ? kernelSzTemp : kernelSzTemp + 1;

  // create DoG
  cv::Mat outer = cv::getGaussianKernel( kernelSz, 1.0 / 3.0 * alpha, CV_32F );
  cv::Mat inner = cv::getGaussianKernel( kernelSz, 0.4 * alpha, CV_32F );

  cv::Mat outerResult, innerResult;
  cv::sepFilter2D( image, outerResult, -1, outer, outer );
  cv::sepFilter2D( image, innerResult, -1, inner, inner );

  const float kernelBoost = 20.0; // boosts result so that we can actually see something
  cv::Mat convResult = (outerResult - innerResult) * kernelBoost;

  for( auto & vec : itsOnOffResponses )
    for( auto & neuron : vec )
    {
      double x, y;
      std::tie(x, y) = neuron->position();
      x *= xScaleFactor; y *= yScaleFactor;

      neuron->setActivation( clamp( convResult.at<float>( y, x ) ), itsState.staticNoiseScale );
    }

  for( auto & vec : itsOffOnResponses )
    for( auto & neuron : vec )
    {
      double x, y;
      std::tie(x, y) = neuron->position();
      x *= xScaleFactor; y *= yScaleFactor;

      neuron->setActivation( clamp( -convResult.at<float>( y, x ) ), itsState.staticNoiseScale );
    }
}

void bd::V1::iterate( int numIters )
{
  auto iterateImpl = [&]()
  {
    threading::parallel_for( 0lu, itsEdgeResponses.size(), [&](size_t i)
    {
      auto & vec = itsEdgeResponses[i];
      for( auto & neuron : vec )
        neuron->update( itsState );
    } );
    threading::parallel_for( 0lu, itsEdgeResponses.size(), [&](size_t i)
    {
      auto & vec = itsEdgeResponses[i];
      for( auto & neuron : vec )
        neuron->learn( itsState );
    } );

    finalizeNeuronVariables();
//    threading::parallel_for( 0lu, itsEdgeResponses.size(), [&](size_t i)
//    {
//      auto & vec = itsEdgeResponses[i];
//      for( auto & neuron : vec )
//        neuron->finalize();
//    } );
  };

  for( int i = 0; i < numIters; ++i )
    iterateImpl();
}
