#ifndef V1_H_
#define V1_H_

#include <opencv2/core/core.hpp>

#include "Gabor.H"
#include "Sheet.H"
#include "BOSheet.H"
#include "State.H"

#include "StaticNeuron.H"
#include "EdgeNeuron.H"

namespace bd
{
  //! Overall border ownership class
  class V1
  {
    public:
      //! Options for V1 specific settings
      struct Options
      {
        Options() :
          useGaussianRing( false ),
          excitationRadius( 0.0 ),
          inhibitionRadius( 1.0 ),
          excitationSpread( 0.27 ),
          inhibitionSpread( 1.0 ),
          excitationStrength( 1.0 ),
          inhibitionStrength( 0.5 )
        {}

        //! Whether to use Gaussian or uniform center-surround weights
        bool useGaussianRing = false;

        //! Radius for uniform, mean for Gaussian
        double excitationRadius = 0.27;
        double inhibitionRadius = 1.0;

        //! Spread for uniform, std for Gaussian
        double excitationSpread = 0.27;
        double inhibitionSpread = 1.0;

        //! Amount of weight distributed amongst connection type
        double excitationStrength = 1.0;
        double inhibitionStrength = 0.5;

        template <class Archive>
        void serialize( Archive & ar )
        {
          ar( CEREAL_NVP(useGaussianRing),
              CEREAL_NVP(excitationRadius),
              CEREAL_NVP(inhibitionRadius),
              CEREAL_NVP(excitationSpread),
              CEREAL_NVP(inhibitionSpread),
              CEREAL_NVP(excitationStrength),
              CEREAL_NVP(inhibitionStrength) );
        }
      };

      /*! @param gridWidth Width of simulation (think input pixels)
          @param gridHeight Height of simulation */
      V1( size_t gridWidth  = 40,
          size_t gridHeight = 40,
          SystemState const & state = SystemState(),
          Options const & options = Options() );

      //! Present a new image for the first time (recomputes v1)
      void onset( cv::Mat const & image );

      //! Advances the network
      void iterate( int numIters = 1 );

      template <class Archive>
      void serialize( Archive & ar )
      {
        ar( CEREAL_NVP(itsState),
            CEREAL_NVP(itsOnOffResponses),
            CEREAL_NVP(itsOffOnResponses),
            CEREAL_NVP(itsEdgeResponses),
            CEREAL_NVP(itsOptions) );
      }

    public:
      SystemState itsState;

      size_t itsGridWidth;
      size_t itsGridHeight;

      //! Input layer
      Sheet<StaticNeuron> itsOnOffResponses;
      Sheet<StaticNeuron> itsOffOnResponses;

      //! V1 layer
      Sheet<EdgeNeuron>  itsEdgeResponses;

      //! General options
      Options itsOptions;
  };
}

#endif // V1_H_
