#include "Wiring.H"

void bd::wireToV1Dense( Sheet<BorderOwnershipNeuron> & sheet, Sheet<StaticNeuron> const & v1, double weight )
{
  auto sheetIter = sheet.begin();
  auto v1Iter = v1.begin();

  for( ; sheetIter != sheet.end(); ++sheetIter, ++v1Iter )
  {
    // grab border ownership neurons ptrs
    auto left =  (*sheetIter)[0];
    auto right = (*sheetIter)[1];

    // get V1 neuron
    const auto n = (*v1Iter)[0];

    left->addInput(  {n, weight, ConnectionFlags::Static | ConnectionFlags::Feedforward} );
    right->addInput( {n, weight, ConnectionFlags::Static | ConnectionFlags::Feedforward} );
  }
}

void bd::calculateBorderOwnershipFBCoefficients( std::vector<Sheet<BorderOwnershipNeuron>> & boSheets, SystemState const & state )
{
  // Calculate target number of connections (max, avg, whatever)
  size_t maxConnections = 0;
  for( auto const & oSheet : boSheets ) // loop orientations
  {
    for( auto const & vec : oSheet )
      for( auto const & n : vec )
      {
        size_t fbCount = 0;
        // get number of FB connections
        for( const auto & input : n->itsInput.pre() )
          if( input.hasFlag( ConnectionFlags::Feedback ) )
            ++fbCount;

        maxConnections = std::max( maxConnections, fbCount );
      }
  }

  // Set normalization coefficient to numFBWeights / maxConnections
  for( auto const & oSheet : boSheets ) // loop orientations
  {
    for( auto & vec : oSheet )
      for( auto & n : vec )
      {
        size_t fbCount = 0;
        // get number of FB connections
        for( const auto & input : n->itsInput.pre() )
          if( input.hasFlag( ConnectionFlags::Feedback ) )
            ++fbCount;

        const double coeff = (double)fbCount / (double)maxConnections;

        // normalize initial FF connections
        for( auto & input : n->itsInput.pre() )
          if( input.hasFlag( ConnectionFlags::Feedback ) )
            input.weight = input.ltWeight = input.initialWeight = input.weight * coeff;

        // This is hacky but normally learn would update post before calling
        // propogateAccumulatedWeight, so we are calling this
        // ahead of learn running for the first time
        n->itsInput.post() = n->itsInput.pre();
        n->itsNormalizationCoefficients.feedback = coeff;
        // call prop to re-normalize using coefficient
        n->propogateAccumulatedWeight( state,
            ConnectionFlags::Feedback | ConnectionFlags::Dynamic,
            0.0, 0.0, 1.0 ); // expects to read from post values
      }
  }
}

void bd::calculateV1Coefficients( Sheet<EdgeNeuron> & sheet, SystemState const & state )
{
  // Calculate target number of connections (max, avg, whatever)
  auto maxConnections = [&sheet]( int flags )
  {
    size_t maxConnections = 0;
    for( auto const & vec : sheet )
      for( auto const & n : vec )
      {
        size_t count = 0;
        for( const auto & input : n->itsInput.pre() )
          if( input.hasFlag( flags ) )
            ++count;

        maxConnections = std::max( maxConnections, count );
      }

    return maxConnections;
  };

  size_t maxFFConnections  = maxConnections( ConnectionFlags::Feedforward );
  size_t maxLatConnections = maxConnections( ConnectionFlags::Lateral | ConnectionFlags::Dynamic );

  // Set normalization coefficient to numFFWeights / maxConnections
  for( auto & vec : sheet )
    for( auto & n : vec )
    {
      size_t ffCount  = 0;
      size_t latCount = 0;
      // get number of FF and lat connections
      for( const auto & input : n->itsInput.pre() )
        if( input.hasFlag( ConnectionFlags::Feedforward ) )
          ++ffCount;
        else if( input.hasFlag( ConnectionFlags::Lateral | ConnectionFlags::Dynamic ) )
          ++latCount;

      const double coeffFF  = (double)ffCount  / (double)maxFFConnections;
      const double coeffLat = (double)latCount / (double)maxLatConnections;

      // This is hacky but normally learn would do this, we are calling this
      // ahead of learn running for the first time
      n->itsInput.post() = n->itsInput.pre(); // TODO
      n->itsNormalizationCoefficients.feedforward = coeffFF;
      n->itsLateralNormalizationCoefficient = coeffLat;
      n->hebbianNormalization( state.v1lateralTotalWeight * coeffLat,
                               0.0,
                               state,
                               ConnectionFlags::Dynamic | ConnectionFlags::Lateral );
    }
}

void bd::normalizeV1LateralExcitatoryConnections( Sheet<EdgeNeuron> & sheet, SystemState const & /*state*/ )
{
  // Calculate target number of connections (max, avg, whatever)
  size_t maxExcitConnections = 0;
  for( auto const & vec : sheet )
    for( auto const & n : vec )
    {
      size_t posCount = 0;
      for( const auto & input : n->itsInput.pre() )
        if( input.hasFlag( ConnectionFlags::Lateral ) )
        {
          if( input.weight >= 0 )
            ++posCount;
        }

      maxExcitConnections = std::max( maxExcitConnections, posCount );
    }

  // Normalize
  for( auto & vec : sheet )
    for( auto & n : vec )
    {
      size_t posCount = 0;
      for( const auto & input : n->itsInput.pre() )
        if( input.hasFlag( ConnectionFlags::Lateral ) )
        {
          if( input.weight >= 0 )
            ++posCount;
        }

      const double excitCoeff = (double)posCount / (double)maxExcitConnections;

      for( auto & input : n->itsInput.pre() )
        if( input.hasFlag( ConnectionFlags::Lateral ) )
        {
          if( input.weight >= 0 )
          {
            input.weight *= excitCoeff;
          }
        }
    }
}

void bd::wireSheetRandom( Sheet<BorderOwnershipNeuron> & s1, Sheet<EdgeNeuron> const & s2,
                          double radius,
                          int numConnections,
                          double weightScaling,
                          int connectionFlags )
{
  auto rng = uniformReal( 0.0, 1.0 );

  const double widthScaling = s1.width() / s2.width();
  const double heightScaling = s1.height() / s2.height();

  // Loop over s1 neurons and find all s2 neurons to connect from
  // Loop neurons in s2
  for( auto & vec : s1 )
    for( auto & n : vec )
    {
      std::vector<std::shared_ptr<EdgeNeuron>> incoming;

      // Step 1: find all incoming connections
      for_each_neuron( n, s2, [&]( std::shared_ptr<BorderOwnershipNeuron> const & n1, std::shared_ptr<EdgeNeuron> const & n2)
      {
        const double n1X = n1->position().first;
        const double n1Y = n1->position().second;

        const double n2X = n2->position().first * widthScaling;
        const double n2Y = n2->position().second * heightScaling;

        const double dist2 = std::pow(n1X - n2X, 2) + std::pow(n1Y - n2Y, 2);

        const bool radiusCheck = dist2 <= radius;

        if( radiusCheck &&
            *n1 != *n2 )
          incoming.push_back( n2 );
      } );

      // Step 2: wire up connections
      for( int i = 0; i < numConnections; ++i )
      {
        int r = uniformInt( 0, (int) incoming.size() - 1 )();
        n->addInput( {incoming[r], weightScaling, connectionFlags} );
        incoming.erase( incoming.begin() + r );
      }
    }
}
