#ifndef WIRING_H_
#define WIRING_H_

#include "State.H"
#include "Sheet.H"
#include "BorderOwnershipNeuron.H"
#include "EdgeNeuron.H"
#include "StaticNeuron.H"

namespace bd
{
  //! Adds incoming connections from V1 to Border Ownership neurons
  /*! This creates a 1 to 1 connection */
  void wireToV1Dense( Sheet<BorderOwnershipNeuron> & sheet, Sheet<StaticNeuron> const & v1, double weight );

  //! Add incoming connections from s2 to s1 in a ring pattern
  /*! Creates connections in thick disk with uniform random
      probability.  All connections are weighted equally.

      @param s1 Create connections TO this sheet
      @param s2 Create connections FROM this sheet
      @param radius The radius of the ring
      @param spread The delta to add to each side of the radius. Cuts off at zero.
      @param probability The probability of creating a connection
      @param weightScaling If normalization is turned on, this is the total strength
                           divided amongst all incoming connections to a neuron. Otherwise,
                           this is the flat weight given to each incoming connection.
      @param normalize Whether to normalize the weight scaling
      @param initialAccumulator The value to initialize each incoming connection's accumulator with.
      @param connectionFlags The flags to apply to all created connections */
  template <class T1, class T2>
  void wireSheetRingUniform( Sheet<T1> & s1, Sheet<T2> const & s2,
                             double radius,
                             double spread,
                             double probability,
                             double weightScaling = 1.0,
                             bool normalize = false,
                             const int connectionFlags = ConnectionFlags::Static,
                             const int excludedFlags = 0 );

  //! Add incoming connections from s2 to s1 in a ring pattern with Gaussian weighting
  /*! Creates connections in thick disk with uniform random
      probability.  Connections are weighted via gaussian distribution.

      @param s1 Create connections TO this sheet
      @param s2 Create connections FROM this sheet
      @param mean The middle radius (mean) of the Gaussian.
      @param stdDev The standard deviation of the Gaussian. Spread will be 3*stdDev.
      @param spread The delta to add to each side of the radius
      @param probability The probability of creating a connection
      @param weightScaling The proportion of total strength divided up amongst
                           all incoming connections to a neuron
      @param connectionFlags The flags to apply to all created connections */
  template <class T1, class T2>
  void wireSheetRing( Sheet<T1> & s1, Sheet<T2> const & s2,
                      double mean,
                      double stdDev,
                      double probability,
                      double weightScaling = 1.0,
                      int connectionFlags = ConnectionFlags::Static );

  //! Add incoming connections from s2 to s1 in a random pattern
  /*! Creates a specified number of random connections in disk
      pattern.  Connections are weighted via flat weight.

      @param radius The radius of the potential connection
      @param numConnections The number of random connections to create
      @param weightScaling The proportion of total strength divided up amongst
                           all incoming connections to a neuron */
  void wireSheetRandom( Sheet<BorderOwnershipNeuron> & s1, Sheet<EdgeNeuron> const & s2,
                        double radius,
                        int numConnections,
                        double weightScaling = 1.0,
                        int connectionFlags = ConnectionFlags::Static );

  //! This calculates all normalization coefficients for the specified sheet of neurons
  template <class NeuronT>
  void calculateCoefficients( Sheet<NeuronT> & sheet );

  //! Calculate the individual coefficients for all border ownership cells
  /*! This coefficient is used to scale the normalized feedback input
      to ensure that border ownership cells near the edge of the simulation
      receive the same amount of activation as one in the center */
  void calculateBorderOwnershipFBCoefficients( std::vector<Sheet<BorderOwnershipNeuron>> & boSheets, SystemState const & state );

  //! Calculate the individual feedforward and dynamic lateral coefficients for all v1 edge neuron cells */
  void calculateV1Coefficients( Sheet<EdgeNeuron> & sheet, SystemState const & state );

  //! Normalizes only the unlearned excitatory V1 connections
  void normalizeV1LateralExcitatoryConnections( Sheet<EdgeNeuron> & sheet, SystemState const & state );
}

#include "WiringImpl.H"

#endif // WIRING_H_
