#include "Random.H"

namespace
{
  //! Applies a function for every neuron in the sheet depending on the passed in neuron
  /*! @param neuron One of the neurons the binary function considers
      @param sheet  A sheet from which every neuron will be sampled
      @param f      A Binary function that will be passed two neurons */
  template <class T1, class T2, class BinaryFunction> inline
  void for_each_neuron( std::shared_ptr<T1> const & neuron, bd::Sheet<T2> const & sheet, BinaryFunction f )
  {
    for( auto const & vec : sheet )
      for( auto const & n : vec )
        f( neuron, n );
  }

  //! Checks if the neuron n2 is already connected to n1
  //! with an exclusive connection type
  template <class T1, class T2> inline
  bool excluded( std::shared_ptr<T1> const & n1, std::shared_ptr<T2> const & n2, const int excludedFlags )
  {
    if( excludedFlags == 0 )
      return false;

    for( auto const & input : n1->itsInput.pre() )
      if( *input.neuron == *n2 && input.hasFlag( excludedFlags ) )
        return true;

    return false;
  }
}

template <class T1, class T2> inline
void bd::wireSheetRingUniform( Sheet<T1> & s1, Sheet<T2> const & s2,
                               double radius,
                               double spread,
                               double probability,
                               double weightScaling,
                               bool normalize,
                               const int connectionFlags,
                               const int excludedFlags )
{
  auto rng = uniformReal( 0.0, 1.0 );

  const double radiusInner2 = (radius - spread) > 0 ? (radius - spread) * (radius - spread) : 0.0;
  const double radiusOuter2 = (radius + spread) * (radius + spread);

  const double widthScaling = s1.width() / s2.width();
  const double heightScaling = s1.height() / s2.height();

  // Loop over s1 neurons and find all s2 neurons to connect from
  for( auto & vec : s1 )
    for( auto & n : vec )
    {
      std::vector<std::shared_ptr<T2>> incoming;

      // Step 1: find all incoming connections
      for_each_neuron( n, s2, [&]( std::shared_ptr<T1> const & n1, std::shared_ptr<T2> const & n2)
      {
        const double n1X = n1->position().first;
        const double n1Y = n1->position().second;

        const double n2X = n2->position().first * widthScaling;
        const double n2Y = n2->position().second * heightScaling;

        const double dist2 = std::pow(n1X - n2X, 2) + std::pow(n1Y - n2Y, 2);

        const bool radiusCheck = dist2 >= radiusInner2 && dist2 <= radiusOuter2;

        if( radiusCheck &&
            (rng() < probability) &&
            (*n1 != *n2) &&
            !excluded( n1, n2, excludedFlags ) )
          incoming.push_back( n2 );
      } );

      // Step 2: wire up all connections
      for( const auto & in : incoming )
      {
        if ( normalize )
          n->addInput( {in, weightScaling / incoming.size(), connectionFlags} );
        else
          n->addInput( {in, weightScaling, connectionFlags} );
      }
    }
}

template <class T1, class T2> inline
void bd::wireSheetRing( Sheet<T1> & s1, Sheet<T2> const & s2,
                        double mean,
                        double stdDev,
                        double probability,
                        double weightScaling,
                        int connectionFlags )
{
  auto rng = uniformReal( 0.0, 1.0 );

  const double radius = mean;
  const double spread = 3 * stdDev;
  const double radiusInner2 = (radius - spread) > 0 ? (radius - spread) * (radius - spread) : 0.0;
  const double radiusOuter2 = (radius + spread) * (radius + spread);

  const double widthScaling = s1.width() / s2.width();
  const double heightScaling = s1.height() / s2.height();

  // Loop over s1 neurons and find all s2 neurons to connect from
  // Loop neurons in s2
  for( auto & vec : s1 )
    for( auto & n : vec )
    {
      double weightSum = 0.0;
      std::vector<std::pair<std::shared_ptr<T2>, double>> incoming;

      // Step 1: find all incoming connections
      for_each_neuron( n, s2, [&]( std::shared_ptr<T1> const & n1, std::shared_ptr<T2> const & n2)
      {
        const double n1X = n1->position().first;
        const double n1Y = n1->position().second;

        const double n2X = n2->position().first * widthScaling;
        const double n2Y = n2->position().second * heightScaling;

        const double dist2 = std::pow(n1X - n2X, 2) + std::pow(n1Y - n2Y, 2);

        const bool radiusCheck = dist2 >= radiusInner2 && dist2 <= radiusOuter2;

        if( radiusCheck &&
            (rng() < probability) &&
            n1 != n2 )
        {
          const double weight = gaussian(std::sqrt(dist2), mean, stdDev);
          incoming.emplace_back( n2, weight );
          weightSum += weight;
        }
      } );

      // Step 2: wire up all connections
      for( const auto in : incoming )
        n->addInput( {in.first, in.second * weightScaling / weightSum, connectionFlags} );
    }
}

template <class NeuronT> inline
void bd::calculateCoefficients( Sheet<NeuronT> & sheet )
{
  // Computes count of given type of connection
  auto getCount = [](std::shared_ptr<Neuron> const & n, const int flags)
  {
    size_t count = 0;
    for( const auto & input : n->itsInput.pre() )
      if( input.hasFlag( flags ) )
        ++count;

    return count;
  };

  // Get max counts
  size_t maxColumnConn = 1,
         maxLateralConn = 1,
         maxFeedbackConn = 1,
         maxFeedbackSkipConn = 1,
         maxFeedforwardConn = 1,
         maxFeedforwardSkipConn = 1;

  for( auto const & vec : sheet )
    for( auto const & n : vec )
    {
      maxColumnConn = std::max( maxColumnConn, getCount( n, ConnectionFlags::Column ) );
      maxLateralConn = std::max( maxLateralConn, getCount( n, ConnectionFlags::Lateral ) );
      maxFeedbackConn = std::max( maxFeedbackConn, getCount( n, ConnectionFlags::Direct | ConnectionFlags::Feedback ) );
      maxFeedbackSkipConn = std::max( maxFeedbackConn, getCount( n, ConnectionFlags::Skip | ConnectionFlags::Feedback ) );
      maxFeedforwardConn = std::max( maxFeedforwardConn, getCount( n, ConnectionFlags::Direct | ConnectionFlags::Feedforward ) );
      maxFeedforwardSkipConn = std::max( maxFeedforwardSkipConn, getCount( n, ConnectionFlags::Skip | ConnectionFlags::Feedforward ) );
    }

  // Normalizes matching connection using coefficient
  auto normalizeConn = [](Input & in, const double coeff, const int flags)
  {
    if( in.hasFlag( flags ) )
      in.weight = in.ltWeight = in.initialWeight = in.weight * coeff;
  };

  // Set normalization coefficient based on count relative to max
  for( auto & vec : sheet )
    for( auto & n : vec )
    {
      const double coeffColumn = (double)getCount( n, ConnectionFlags::Column ) / (double)maxColumnConn;
      const double coeffLateral = (double)getCount( n, ConnectionFlags::Lateral ) / (double)maxLateralConn;
      const double coeffFeedback = (double)getCount( n, ConnectionFlags::Direct | ConnectionFlags::Feedback ) / (double)maxFeedbackConn;
      const double coeffFeedbackSkip = (double)getCount( n, ConnectionFlags::Skip | ConnectionFlags::Feedback ) / (double)maxFeedbackSkipConn;
      const double coeffFeedforward = (double)getCount( n, ConnectionFlags::Direct | ConnectionFlags::Feedforward ) / (double)maxFeedforwardConn;
      const double coeffFeedforwardSkip = (double)getCount( n, ConnectionFlags::Skip | ConnectionFlags::Feedforward ) / (double)maxFeedforwardSkipConn;

      for( auto & input : n->itsInput.pre() )
      {
        normalizeConn( input, coeffColumn, ConnectionFlags::Column );
        normalizeConn( input, coeffLateral, ConnectionFlags::Lateral );
        normalizeConn( input, coeffFeedback, ConnectionFlags::Direct | ConnectionFlags::Feedback );
        normalizeConn( input, coeffFeedbackSkip, ConnectionFlags::Skip | ConnectionFlags::Feedback );
        normalizeConn( input, coeffFeedforward, ConnectionFlags::Direct | ConnectionFlags::Feedforward );
        normalizeConn( input, coeffFeedforwardSkip, ConnectionFlags::Skip | ConnectionFlags::Feedforward );
      }

      // This is hacky but normally learn would update post before calling
      // propogateAccumulatedWeight, so we are calling this
      // ahead of learn running for the first time
      n->itsInput.post() = n->itsInput.pre();

      // Set actual coefficients
      n->itsNormalizationCoefficients.column = coeffColumn;
      n->itsNormalizationCoefficients.lateral = coeffLateral;
      n->itsNormalizationCoefficients.feedback = coeffFeedback;
      n->itsNormalizationCoefficients.feedbackSkip = coeffFeedbackSkip;
      n->itsNormalizationCoefficients.feedforward = coeffFeedforward;
      n->itsNormalizationCoefficients.feedforwardSkip = coeffFeedforwardSkip;
    }
}

