#include "CombinedTests.H"
#include "../src/Generate.H"
#include "../src/Random.H"
#include "Util.H"
#include "../config/Options.H"

#include <fstream>
#include <cereal/archives/binary.hpp>

#include <boost/filesystem.hpp>

extern OptionsStruct Options;

void testFullCombinedConvergence( bd::Combined & network, cv::Mat const & mask )
{
  const int height = (int)network.itsGridHeight;
  const int width = (int)network.itsGridWidth;

  auto shapeRNG = bd::uniformInt( 0, 1 );

  auto getRandomShape = [&]()
  {
    bd::Shape s;
    do
    {
      const double theta = bd::uniformReal( 0.0, 2 * M_PI )();
      const double scale = bd::uniformReal( 1.0, 1.6 )();

      switch( shapeRNG() )
      {
        case 0: // rectangle
          s = bd::generateRectangle( bd::uniformInt( -13, width - 1 )(),
                                     bd::uniformInt( -13, height - 1)(),
                                     10, 10,
                                     scale );
          break;
        case 1: // line
          s = bd::generateLine( bd::uniformInt( 1 - width, width - 1 )(),
                                bd::uniformInt( 1 - height, height - 1)(),
                                std::max(height, width) );
          break;
        case 2: // triangle
          s = bd::generateEquilateralTriangle( bd::uniformInt( 0, width - 1 )(),
                                               bd::uniformInt( 0, height - 1)(),
                                               4 );
          break;
        default:
          throw "Not implemented";
          break;
      }

      s = bd::rotateShape( s, theta );
    }
    while( !bd::isValid( network, s ) );
    return s;
  };

  // choose a random direction in radians, 0 to 2pi
  auto randomDxDy = []()
  {
    const double theta = bd::uniformReal( 0.0, 2 * M_PI )();
    return cv::Point2d( std::cos(theta), -std::sin(theta) );
  };

  // main loop
  const int numPresentIters = 10; // how long to show each square at each location
  const int numOverallIters = 100000;

  //bd::showImage( bd::sheetNeuronImage( network.itsGNResponses[0], 2.0 ), false );

  for( int i = 0; i < numOverallIters; ++i )
  {
    auto shape = getRandomShape();
    auto dxdy = randomDxDy();

    // present square and move it steps in some direction
    for( int steps = 0; steps < 20; ++steps )
    {
      cv::Mat stimulus = cv::Mat::zeros( height, width, CV_32FC1 );

      cv::Mat stimulus8U = cv::Mat::zeros( height, width, CV_8UC1 );
      bd::drawShape( stimulus8U, shape, 255.0, false, 1, CV_AA );

      stimulus8U.convertTo( stimulus, CV_32FC1 );
      stimulus /= 255.0;

      bd::test::presentStimulus( network, stimulus, numPresentIters );

      // translate along random direction
      shape = bd::translateShape( shape, dxdy );

      if( !bd::isValid( network, shape ) )
        break;
    }

    bd::test::doMask( network, mask );

    const int numInspectIters = Options.iters;
    if( i % numInspectIters == 0 && i != 0 )
    {
      bd::test::saveBD( network, Options.output + "_" + std::to_string(i) + "_combined_full_converge.cereal" );

      std::cout << "=================================" << std::endl;
      std::cout << "=================================" << std::endl;
      std::cout << "Done with " << i << " iterations " << std::endl;
      std::cout << "=================================" << std::endl;
      std::cout << "=================================" << std::endl;
    }
    else
      std::cout << "Counter: " << (i % numInspectIters) << std::endl;
  }
}
