#include "ConvergenceTests.H"
#include "../src/Generate.H"
#include "../src/Random.H"
#include "Util.H"
#include "../config/Options.H"

#include <fstream>
#include <cereal/archives/binary.hpp>

#include <boost/filesystem.hpp>

#include <chrono>

extern OptionsStruct Options;

void testSlideConvergence( bd::Border & border, cv::Mat const & mask, bool save )
{
  const int height = border.itsGridHeight;
  const int width = border.itsGridWidth;

  auto image = bd::generateRectangle( height, width, 10, 10, 10, 10 );

  auto doMask = [&]()
  {
    border.onset( mask );
    border.iterate( 20 );
  };

  auto presentSquare = [&]( int x, int y, int numIters, std::function<void()> postFunc )
  {
    auto localimage = bd::generateRectangle( height, width, x, y, 10, 10 );

    border.onset( localimage );
    border.iterate(numIters);

    postFunc();
  };

  if( save )
  {
    for( int y = height - 16; y >= 10; --y )
      presentSquare( 10, y, 20, [&]() { } );

    bd::test::saveBD( border );
  }
  else
    bd::test::loadBD( border );

  doMask();
  displayPolarity(border.itsBOResponses[0], border.itsGNResponses, false) ;
  bd::showImage( bd::sheetNeuronImage( border.itsGNResponses, 2.0 ), false );

  for( int y = 10; y < height - 15; ++y )
  {
    presentSquare( 10, y, 20, [&]()
    {
      showConnections( border.itsGNResponses, false );
      showConnections( border.itsBOResponses[0], border.itsGNResponses, false );
      bd::showImage( visualizeInhibitoryPairs( border.itsBOResponses[0] ), false );
      displayPolarity(border.itsBOResponses[0], border.itsGNResponses, false) ;
    } );
  }

  doMask();

  for( int y = 10; y < height - 15; ++y )
    presentSquare( 10, y, 20, [&]()
    {
      showConnections( border.itsGNResponses, false );
      showConnections( border.itsBOResponses[0], border.itsGNResponses, false );
      bd::showImage( visualizeInhibitoryPairs( border.itsBOResponses[0] ), false );
      displayPolarity(border.itsBOResponses[0], border.itsGNResponses, false) ;
      doMask();
    } );
}

void testSimpleConvergence( bd::Border & border )
{
  const int numiters = 10;
  const float resolutionScaling = border.itsResolutionScaling;
  const int height = (int)border.itsGridHeight * resolutionScaling;
  const int width = (int)border.itsGridWidth * resolutionScaling;
  const int yloc = height / 2 - 5 * resolutionScaling;

  cv::Mat mask = cv::Mat::zeros( height, width, CV_32FC1 );

  auto presentSquare = [&]( int x, int y, int numIters, std::function<void()> postFunc )
  {
    const int squareSize = 10 * resolutionScaling;
    auto rectangle = bd::generateRectangle( x, y, squareSize, squareSize );

    cv::Mat stimulus = cv::Mat::zeros( height, width, CV_32FC1 );
    bd::drawShapesInterpolated( stimulus, {{rectangle, 255.0, false, (int)resolutionScaling}} );

    bd::test::presentStimulus( border, stimulus, numIters );

    if( Options.display )
      bd::showImage(stimulus, false, true, "snarf");

    postFunc();
  };

  auto doMask = [&]()
  {
    border.onset( mask );
    border.iterate( 20 );
  };

  auto goLeft = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int x = width - 11 * resolutionScaling; x >= 0; x -= step )
      presentSquare( x, yloc, numiters, postFunc );
  };

  //auto goLeftLoc = [&](int xx, std::function<void()> postFunc)
  //{
  //  const int step = resolutionScaling;
  //  for( int x = width - 11; x >= 0; x -= step )
  //  {
  //    if( (x == xx) || (x + 9 == xx) )
  //    {
  //      presentSquare( x, yloc, 5, postFunc );
  //      presentSquare( x, yloc, numiters - 5, postFunc );
  //    }
  //    else
  //      presentSquare( x, yloc, numiters, [](){});
  //  }
  //};

  auto goRight = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int x = 0; x < width - 10 * resolutionScaling; x += step )
      presentSquare( x, yloc, numiters, postFunc );
  };

  //auto goRightLoc = [&](int xx, std::function<void()> postFunc)
  //{
  //  const int step = resolutionScaling;
  //  for( int x = 0; x < width - 10; x += step )
  //  {
  //    if( (x == xx) || (x + 9 == xx) )
  //    {
  //      presentSquare( x, yloc, 5, postFunc );
  //      presentSquare( x, yloc, numiters - 5, postFunc );
  //    }
  //    else
  //      presentSquare( x, yloc, numiters, [](){} );
  //  }
  //};

  auto display = [&](int x, int y, bool clear = true)
  {
    if( clear )
    {
      bd::test::doMask( border, mask );
      presentSquare( x, y, yloc, [](){} );
    }

    showConnections( border.itsGNResponses, false, false, "grouping1" );
    {
      int maxPerGrid = 0;
      for( auto const & cell : border.itsGNResponses )
        maxPerGrid = std::max( maxPerGrid, (int)cell.size() );
      const double scaling = maxPerGrid;

      bd::showImage( sheetNeuronValueImage( border.itsGNResponses, [](std::shared_ptr<bd::GroupingNeuron> const & val){ return val->itsActivationThreshold; }, scaling ),
          false, true, "grouping1_activationThresh");
    }
    //showConnections( border.itsGNResponses[1], false, false, "grouping2" );

    //showConnections( border.itsBOResponses[0], border.itsGNResponses, false, "bo_0_0" );
    //showConnections( border.itsBOResponses[1], border.itsGNResponses, false, "bo_45_0" );
    //showConnections( border.itsBOResponses[2], border.itsGNResponses, false, "bo_90_0" );
    //showConnections( border.itsBOResponses[3], border.itsGNResponses, false, "bo_135_0" );

    displayPolarity( border.itsBOResponses[0], border.itsGNResponses, true, "polarity_0_0" );
    displayPolarity( border.itsBOResponses[1], border.itsGNResponses, true, "polarity_45_0" );
    displayPolarity( border.itsBOResponses[2], border.itsGNResponses, true, "polarity_90_0" );
    displayPolarity( border.itsBOResponses[3], border.itsGNResponses, true, "polarity_135_0" );

    //showConnections( border.itsBOResponses[0], border.itsGNResponses[1], false, "bo_0_1" );
    //showConnections( border.itsBOResponses[1], border.itsGNResponses[1], false, "bo_45_1" );
    //showConnections( border.itsBOResponses[2], border.itsGNResponses[1], false, "bo_90_1" );
    //showConnections( border.itsBOResponses[3], border.itsGNResponses[1], false, "bo_135_1" );
  };

  auto displayIf = [&](int i, int mod)
  {
    if( i % mod == 0 && i != 0 )
      display( 0, 0, false );
  };

  for( int i = 0; i < 99999; ++i )
  {
    //goLeft([](){});
    goLeft([&](){ display(0,0,false); });
    doMask();

    //goRight([](){});
    goRight([&](){ display(0,0,false); });
    doMask();

    // testing reset
    //border.onset( mask );
    //border.iterate( 200 );

    const int numSaveIters = 2;
    if( i % numSaveIters == 0 && i != 0 )
    {
      bd::test::saveBD( border, Options.output + "_" + std::to_string(i) + "_simple_converge.cereal" );
      std::cout << "Iteration " << i << std::endl;
    }

    if( Options.display )
      displayIf( i, 1 );

    //std::cout << "Which x position?" << std::endl;
    //int xx;
    //std::cin >> xx;
    //std::cout << "received " << xx << std::endl;

    //goLeftLoc( xx, [&](){display(0, 0, false);} );
    //doMask();
    //display( 0,0,false );

    //std::cout << "Which x position?" << std::endl;
    //std::cin >> xx;
    //std::cout << "received " << xx << std::endl;

    //goRightLoc( xx, [&](){display(0, 0, false);} );
    //doMask();
    //display( 0,0,false );



    //goRight([](){});
    //doMask();
    //display( 0,0,false );

    //goRight([&](){ display(0,0,false); });
    //doMask();

    //goLeft([&](){doMask();});
    //goRight([&](){doMask();});

    //display( width / 2, 10 );
    //display( 0,0,false );
  }
}

void testCrossConvergence( bd::Border & border )
{
  const int numiters = 10;
  const float resolutionScaling = border.itsResolutionScaling;
  const int height = (int)border.itsGridHeight * resolutionScaling;
  const int width = (int)border.itsGridWidth * resolutionScaling;
  const int yloc = height / 2 - 5 * resolutionScaling;
  const int xloc = width / 2 - 5 * resolutionScaling;

  assert(height == width); // only works with square input for now

  cv::Mat mask = cv::Mat::zeros( height, width, CV_32FC1 );

  auto presentSquare = [&]( int x, int y, int numIters, std::function<void()> postFunc )
  {
    const int squareSize = 10 * resolutionScaling;
    auto rectangle = bd::generateRectangle( x, y, squareSize, squareSize );

    cv::Mat stimulus = cv::Mat::zeros( height, width, CV_32FC1 );
    bd::drawShapesInterpolated( stimulus, {{rectangle, 255.0, false, (int)resolutionScaling}} );

    bd::test::presentStimulus( border, stimulus, numIters );

    if( Options.display )
      bd::showImage(stimulus, false, true, "snarf");

    postFunc();
  };

  auto doMask = [&]()
  {
    border.onset( mask );
    border.iterate( 20 );
  };

  auto goLeft = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int x = width - 11 * resolutionScaling; x >= 0; x -= step )
      presentSquare( x, yloc, numiters, postFunc );
  };

  auto goRight = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int x = 0; x < width - 10 * resolutionScaling; x += step )
      presentSquare( x, yloc, numiters, postFunc );
  };

  auto goUp = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int y = height - 11 * resolutionScaling; y >= 0; y -= step )
      presentSquare( xloc, y, numiters, postFunc );
  };

  auto goDown = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int y = 0; y < height - 10 * resolutionScaling; y += step )
      presentSquare( xloc, y, numiters, postFunc );
  };

  auto goDownRight = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int y = 0, x = 0; y < height - 10 * resolutionScaling &&
                           x < width  - 10 * resolutionScaling;
         y += step, x += step )
      presentSquare( x, y, numiters, postFunc );
  };

  auto goDownLeft = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int y = 0, x = width - 11 * resolutionScaling;
         y < height - 10 * resolutionScaling &&
         x >= 0;
         y += step, x -= step )
      presentSquare( x, y, numiters, postFunc );
  };

  auto goUpRight = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int y = height - 11 * resolutionScaling, x = 0;
         y >= 0 &&
         x < width  - 10 * resolutionScaling;
         y -= step, x += step )
      presentSquare( x, y, numiters, postFunc );
  };

  auto goUpLeft = [&](std::function<void()> postFunc)
  {
    const int step = resolutionScaling;
    for( int y = height - 11 * resolutionScaling, x = width - 11 * resolutionScaling;
         y >= 0 &&
         x >= 0;
         y += step, x -= step )
      presentSquare( x, y, numiters, postFunc );
  };

  auto display = [&](int x, int y, bool clear = true)
  {
    if( clear )
    {
      bd::test::doMask( border, mask );
      presentSquare( x, y, numiters, [](){} );
    }

    showConnections( border.itsGNResponses, false, false, "grouping1" );
    {
      int maxPerGrid = 0;
      for( auto const & cell : border.itsGNResponses )
        maxPerGrid = std::max( maxPerGrid, (int)cell.size() );
      const double scaling = maxPerGrid;

      bd::showImage( sheetNeuronValueImage( border.itsGNResponses, [](std::shared_ptr<bd::GroupingNeuron> const & val){ return val->itsActivationThreshold; }, scaling ),
          false, true, "grouping1_activationThresh");
    }
    //showConnections( border.itsGNResponses[1], false, false, "grouping2" );

    //showConnections( border.itsBOResponses[0], border.itsGNResponses, false, "bo_0_0" );
    //showConnections( border.itsBOResponses[1], border.itsGNResponses, false, "bo_45_0" );
    //showConnections( border.itsBOResponses[2], border.itsGNResponses, false, "bo_90_0" );
    //showConnections( border.itsBOResponses[3], border.itsGNResponses, false, "bo_135_0" );

    displayPolarity( border.itsBOResponses[0], border.itsGNResponses, true, "polarity_0_0" );
    displayPolarity( border.itsBOResponses[1], border.itsGNResponses, true, "polarity_45_0" );
    displayPolarity( border.itsBOResponses[2], border.itsGNResponses, true, "polarity_90_0" );
    displayPolarity( border.itsBOResponses[3], border.itsGNResponses, true, "polarity_135_0" );

    //showConnections( border.itsBOResponses[0], border.itsGNResponses[1], false, "bo_0_1" );
    //showConnections( border.itsBOResponses[1], border.itsGNResponses[1], false, "bo_45_1" );
    //showConnections( border.itsBOResponses[2], border.itsGNResponses[1], false, "bo_90_1" );
    //showConnections( border.itsBOResponses[3], border.itsGNResponses[1], false, "bo_135_1" );
  };

  auto displayIf = [&](int i, int mod)
  {
    if( i % mod == 0 && i != 0 )
      display( 0, 0, false );
  };

  for( int i = 0; i < 99999; ++i )
  {
    goDownLeft([&](){ display(0,0,false); });
    doMask();

    goDownRight([&](){ display(0,0,false); });
    doMask();

    goLeft([&](){ display(0,0,false); });
    doMask();

    goRight([&](){ display(0,0,false); });
    doMask();

    goUp([&](){ display(0,0,false); });
    doMask();

    goDown([&](){ display(0,0,false); });
    doMask();

    goUpLeft([&](){ display(0,0,false); });
    doMask();

    goUpRight([&](){ display(0,0,false); });
    doMask();

    const int numSaveIters = 4;
    if( i % numSaveIters == 0 && i != 0 )
    {
      bd::test::saveBD( border, Options.output + "_" + std::to_string(i) + "_cross_converge.cereal" );
      std::cout << "Iteration " << i << std::endl;
    }

    //if( Options.display )
    //  displayIf( i, 1 );
  }
}

void testFullConvergence( bd::Border & border, double groupingScale )
{
  const float resolutionScaling = border.itsResolutionScaling;
  const int height = (int)border.itsGridHeight * resolutionScaling;
  const int width = (int)border.itsGridWidth * resolutionScaling;

  const int maxSteps = (int)std::max( border.itsGridHeight, border.itsGridWidth );

  const double minScale = 0.7; // .9
  const double maxScale = 2.3; // 1.1

  cv::Mat mask = cv::Mat::zeros( height, width, CV_32FC1 );

  auto shapeRNG = bd::uniformInt( 0, 0 );

  auto getRandomShape = [&]()
  {
    bd::Shape s;
    do
    {
      const double theta = bd::uniformReal( 0.0, 2 * M_PI )();
      const double scale = bd::uniformReal( minScale, maxScale )();
      const int squareSize = 10 * groupingScale * resolutionScaling;

      switch( shapeRNG() )
      {
        case 0:
          // square
          s = bd::generateRectangle( bd::uniformInt( -squareSize - 1, width - 1 )(),
                                     bd::uniformInt( -squareSize - 1, height - 1)(),
                                     squareSize, squareSize,
                                     scale );
          break;
        case 1:
          {
            // rectangle with random aspect ratio
            const double aspect = bd::uniformReal( 1.0, 2.0 )();
            const int sWidth = squareSize * aspect;
            const int sHeight = squareSize;
            s = bd::generateRectangle( bd::uniformInt( -sWidth - 1, width - 1 )(),
                                       bd::uniformInt( -sHeight - 1, height - 1)(),
                                       sWidth, sHeight,
                                       scale );
            break;
          }
        case 2:
          // triangle
          s = bd::generateEquilateralTriangle( bd::uniformInt( -squareSize/2 - 1, width - 1 )(),
                                               bd::uniformInt( -squareSize/2 - 1, height - 1)(),
                                               squareSize / 2.0 );
          break;
        case 3:
          // line
          s = bd::generateLine( bd::uniformInt( 1 - width, width - 1 )(),
                                bd::uniformInt( 1 - height, height - 1)(),
                                std::max(height, width) );
          break;
        default:
          throw "Not implemented";
          break;
      }

      s = bd::rotateShape( s, theta );
    }
    while( !bd::isValid( border, s, resolutionScaling ) );
    return s;
  };

  // choose a random direction in radians, 0 to 2pi
  auto randomDxDy = []()
  {
    const double theta = bd::uniformReal( 0.0, 2 * M_PI )();
    return cv::Point2d( std::cos(theta), -std::sin(theta) );
  };

  // choose a random direction towards any other quadrant
  auto otherQuadrantDxDy = [](double width, double height, bd::Shape const & shape)
  {
    const double x = shape.centroid.x;
    const double y = shape.centroid.y;

    double minRNG = 0.0;
    double maxRNG = 2 * M_PI;

    if( x < width / 2 )
    {
      if( y < height / 2 ) // upper left
      {
        minRNG = 3.0 / 2.0 * M_PI;
        maxRNG = 2.0 * M_PI;
      }
      else // bottom left
      {
        minRNG = 0.0;
        maxRNG = M_PI / 2.0;
      }
    }
    else
    {
      if( y < height / 2 ) // upper right
      {
        minRNG = M_PI;
        maxRNG = 3.0 / 2.0 * M_PI;
      }
      else // bottom right
      {
        minRNG = M_PI / 2.0;
        maxRNG = M_PI;
      }
    }

    const double theta = bd::uniformReal( minRNG, maxRNG )();
    return cv::Point2d( std::cos(theta), -std::sin(theta) );
  };

  // main loop
  const int numPresentIters = 13; // how long to show each square at each location
  const int numOverallIters = 80001;
  //const int numOverallIters = 10000;

  //bd::showImage( bd::sheetNeuronImage( border.itsGNResponses, 2.0 ), false );

  for( int i = 0; i < numOverallIters; ++i )
  {
    auto shape = getRandomShape();
    const double scaleStep = std::exp( std::log( bd::uniformReal( minScale, maxScale )() ) / maxSteps );
    //auto dxdy = randomDxDy() * resolutionScaling;
    auto dxdy = otherQuadrantDxDy(width, height, shape) * resolutionScaling;

    // present square and move it steps in some direction
    //for( int steps = 0; steps < maxSteps; ++steps )
    bool displayShape = true;
    while( bd::isValid( border, shape, resolutionScaling ) )
    {
      cv::Mat stimulus = cv::Mat::zeros( height, width, CV_32FC1 );
      bd::drawShapesInterpolated( stimulus, {{shape, 255.0, false, (int)resolutionScaling}} );

      //bd::test::presentStimulus( border, stimulus, numPresentIters );

      bool finishDisplay = true;
      for( int z = 0; z < numPresentIters; ++z )
      {
        bd::test::presentStimulus( border, stimulus, 1 );

        if( Options.display && finishDisplay && displayShape )
        { // DEBUG
          bd::showImage(stimulus, false, false, "input");

          showConnections( border.itsGNResponses, false, false, "grouping" );
          {
            int maxPerGrid = 0;
            for( auto const & cell : border.itsGNResponses )
              maxPerGrid = std::max( maxPerGrid, (int)cell.size() );
            const double scaling = maxPerGrid;

            bd::showImage( sheetNeuronValueImage( border.itsGNResponses, [](std::shared_ptr<bd::GroupingNeuron> const & val){ return val->itsActivationThreshold; }, scaling ),
                           false, false, "grouping1_activationThresh");
          }

          showConnections( border.itsProtoResponses, false, false, "proto" );

          //showConnections( border.itsBOResponses[0], border.itsGNResponses, false, "bo_0_0" );
          //showConnections( border.itsBOResponses[1], border.itsGNResponses, false, "bo_45_0" );
          //showConnections( border.itsBOResponses[2], border.itsGNResponses, false, "bo_90_0" );
          //showConnections( border.itsBOResponses[3], border.itsGNResponses, false, "bo_135_0" );

          displayPolarity( border.itsBOResponses[0], border.itsGNResponses, false, "polarity_0_0" );
          displayPolarity( border.itsBOResponses[1], border.itsGNResponses, false, "polarity_45_0" );
          displayPolarity( border.itsBOResponses[2], border.itsGNResponses, false, "polarity_90_0" );
          displayPolarity( border.itsBOResponses[3], border.itsGNResponses, false, "polarity_135_0" );

          bd::showImage( bd::visualizeAllDirections( border.itsBOResponses, resolutionScaling ), false, true, "bo_dirs" );

          char key = cv::waitKey(0);
          if( key == 'n' )
            finishDisplay = false;
          else if (key == 'k' )
            displayShape = false;
        } // end DEBUG
      } // end iterated display version

      // translate along random direction
      shape = bd::translateShape( shape, dxdy );
      // scale adjustment
      shape = bd::scaleShape( shape, scaleStep );

      if( !bd::isValid( border, shape, resolutionScaling ) )
        break;
    }

    bd::test::doMask( border, mask );

    const int numInspectIters = Options.iters;
    if( i % numInspectIters == 0 && i != 0 )
    {
      bd::test::saveBD( border, Options.output + "_" + std::to_string(i) + "_full_converge.cereal" );
      if( Options.display )
      {
        showConnections( border.itsGNResponses, false );
        displayPolarity(border.itsBOResponses[0], border.itsGNResponses, false) ;
        showConnections( border.itsBOResponses[0], border.itsGNResponses, false );
      }

      std::cout << "=================================" << std::endl;
      std::cout << "=================================" << std::endl;
      std::cout << "Done with " << i << " iterations " << std::endl;
      std::cout << "=================================" << std::endl;
      std::cout << "=================================" << std::endl;
    }
    else
      std::cout << "Counter: " << (i % numInspectIters) << std::endl;
  }
}

void testFullConvergenceSpiral( bd::Border & border, cv::Mat const & mask )
{
  const int height = (int)border.itsGridHeight;
  const int width = (int)border.itsGridWidth;

  const int xMid = width / 2 - 5;
  const int yMid = height / 2 - 5;

  auto doMask = [&]()
  {
    border.onset( mask );
    border.iterate( 20 );
  };

  // presents a square at some location
  auto presentSquare = [&]( int x, int y, int numIters, std::function<void()> postFunc )
  {
    auto localimage = bd::generateRectangle( height, width, x, y, 10, 10 );

    border.onset( localimage );
    border.iterate(numIters);

    postFunc();
  };

  auto isLocValid = [&]( const int x, const int y )
  {
    if( x < 10 || x > width - 21 )
      return false;

    if( y < 10 || y > height - 21 )
      return false;

    return true;
  };

  // main loop
  const int numPresentIters = 9; // how long to show each square at each location
  const int numOverallIters = 4;

  //bd::showImage( bd::sheetNeuronImage( border.itsGNResponses, 2.0 ), false );

  std::array<int, 4> dx = {{1, 0, -1, 0}};
  std::array<int, 4> dy = {{0, -1, 0, 1}};

  // One iteration is an entirely presentation of the spiral
  for( int i = 0; i < numOverallIters; ++i )
  {
    int curX = xMid;
    int curY = yMid;
    int length = 1;
    int count = 0;
    size_t direction = 0;

    while( isLocValid( curX, curY ) )
    {
      // Do one side length of presentation
      for( int j = 0; j < length; ++j )
      {
        presentSquare( curX, curY, numPresentIters, [](){} );
        curX += dx[direction];
        curY += dy[direction];
      }

      direction = (direction + 1) % 4;

      if( ++count == 2 )
      {
        count = 0;
        ++length;
      }
    }

    doMask();

    const int numInspectIters = 1;
    if( i % numInspectIters == 0 && i != 0 )
    {
      bd::test::saveBD( border );
      displayPolarity(border.itsBOResponses[0], border.itsGNResponses, false) ;
      showConnections( border.itsBOResponses[0], border.itsGNResponses, false );
    }
    else
      std::cout << "Counter: " << (i % numInspectIters) << std::endl;
  }
}

void testConvergenceSpecificRow( bd::Border & border, cv::Mat const & mask )
{
  const int height = (int)border.itsGridHeight;
  const int width = (int)border.itsGridWidth;

  const int numPresentIters = 9; // how long to show each square at each location

  auto doMask = [&]()
  {
    border.onset( mask );
    border.iterate( 20 );
  };

  // presents a square at some location
  auto presentSquare = [&]( int x, int y, int numIters, std::function<void()> postFunc )
  {
    auto localimage = bd::generateRectangle( height, width, x, y, 10, 10 );

    border.onset( localimage );
    border.iterate(numIters);

    postFunc();
  };

  auto goRightRow = [&](int yy, std::function<void()> postFunc)
  {
    for( int x = 10; x < width - 20; ++x )
    {
      std::cerr << "Square upper left corner is at " << x << ", " << yy << std::endl;
      std::cerr << "=============================================" << std::endl;
      presentSquare( x, yy, numPresentIters, postFunc );
    }
  };

  auto goLeftRow = [&](int yy, std::function<void()> postFunc)
  {
    for( int x = width - 21; x >= 10; --x )
    {
      std::cerr << "Square upper left corner is at " << x << ", " << yy << std::endl;
      std::cerr << "=============================================" << std::endl;
      presentSquare( x, yy, numPresentIters, postFunc );
    }
  };

  auto displayIf = [&](int i, int mod)
  {
    if( i % mod == 0 )
      displayPolarity(border.itsBOResponses[0], border.itsGNResponses, false) ;
  };

  bd::test::loadBD( border );
  displayPolarity(border.itsBOResponses[0], border.itsGNResponses, false);

  std::cout << "Which row?" << std::endl;
  int row;
  std::cin >> row;
  std::cout << "Doing row " << row << std::endl;

  for( int i = 0; i < 999999; ++i )
  {
    goRightRow( row, [&]() { } );
    doMask();

    goLeftRow( row, [&]() { } );
    doMask();
    displayIf( i, 20 );
  }
}

namespace
{
  // Returns a list of all files in a directory
  std::vector<boost::filesystem::path> dir(const boost::filesystem::path& root, const std::string& ext)
  {
    namespace fs = boost::filesystem;

    if(!fs::exists(root) || !fs::is_directory(root))
      return {};

    fs::recursive_directory_iterator it(root);
    fs::recursive_directory_iterator endit;

    std::vector<boost::filesystem::path> ret;

    while(it != endit)
    {
      if(fs::is_regular_file(*it) && it->path().extension() == ext)
        ret.push_back(it->path().string());
      ++it;
    }

    return ret;
  }

  // Loads an image and returns a color grayscale pair
  std::pair<cv::Mat, cv::Mat> loadImage( std::string const & file )
  {
    auto image = cv::imread( file );
    // convert to float, leave in 0 to 255 range
    if( image.type() != CV_32FC3 )
      image.convertTo( image, CV_32F );

    cv::Mat grayscale;

    cv::cvtColor( image, grayscale, CV_BGR2GRAY );
    cv::normalize( grayscale, grayscale, 0, 1, cv::NORM_MINMAX );

    return {image, grayscale};
  }

  // Rotates a mat by some number of degrees, potentially making the image larger
  void rotate2D(const cv::Mat & src, cv::Mat & dst, const double degrees)
  {
    cv::Mat frame, frameRotated;

    int diagonal = (int)sqrt(src.cols * src.cols + src.rows * src.rows);
    int newWidth = diagonal;
    int newHeight = diagonal;

    int offsetX = (newWidth - src.cols) / 2;
    int offsetY = (newHeight - src.rows) / 2;
    cv::Mat targetMat(newWidth, newHeight, src.type(), cv::Scalar(0));
    cv::Point2f src_center(targetMat.cols / 2.0f, targetMat.rows / 2.0f);

    src.copyTo(frame);

    frame.copyTo(targetMat.rowRange(offsetY, offsetY +
          frame.rows).colRange(offsetX, offsetX + frame.cols));
    cv::Mat rot_mat = cv::getRotationMatrix2D(src_center, degrees, 1.0);
    cv::warpAffine(targetMat, frameRotated, rot_mat, targetMat.size(), cv::INTER_NEAREST);

    cv::Rect bound_Rect(frame.cols, frame.rows, 0, 0);
    int x1 = offsetX;
    int x2 = offsetX + frame.cols;
    int x3 = offsetX;
    int x4 = offsetX + frame.cols;
    int y1 = offsetY;
    int y2 = offsetY;
    int y3 = offsetY + frame.rows;
    int y4 = offsetY + frame.rows;
    cv::Mat co_Ordinate = (cv::Mat_<double>(3, 4) << x1, x2, x3, x4,
        y1, y2, y3, y4,
        1,  1,  1,  1 );

    cv::Mat RotCo_Ordinate = rot_mat * co_Ordinate;

    for (int i = 0; i < 4; ++i) {
      if (RotCo_Ordinate.at<double>(0, i) < bound_Rect.x)
        bound_Rect.x = (int)RotCo_Ordinate.at<double>(0, i);
      if (RotCo_Ordinate.at<double>(1, i) < bound_Rect.y)
        bound_Rect.y = RotCo_Ordinate.at<double>(1, i);
    }

    for (int i = 0; i < 4; ++i) {
      if (RotCo_Ordinate.at<double>(0, i) > bound_Rect.width)
        bound_Rect.width = (int)RotCo_Ordinate.at<double>(0, i);
      if (RotCo_Ordinate.at<double>(1, i) > bound_Rect.height)
        bound_Rect.height = RotCo_Ordinate.at<double>(1, i);
    }

    bound_Rect.width = bound_Rect.width - bound_Rect.x;
    bound_Rect.height = bound_Rect.height - bound_Rect.y;

    if (bound_Rect.x < 0)
      bound_Rect.x = 0;
    if (bound_Rect.y < 0)
      bound_Rect.y = 0;
    if (bound_Rect.width > frameRotated.cols)
      bound_Rect.width = frameRotated.cols;
    if (bound_Rect.height > frameRotated.rows)
      bound_Rect.height = frameRotated.rows;

    cv::Mat ROI = frameRotated(bound_Rect);
    ROI.copyTo(dst);
  }

  // Performs a tight crop of an image
  cv::Mat autoCrop( cv::Mat const & image )
  {
    int xMin = image.cols;
    int xMax = 0;
    int yMin = image.rows;
    int yMax = 0;

    for( int y = 0; y < image.rows; ++y )
      for( int x = 0; x < image.cols; ++x )
      {
        if( image.at<float>( y, x ) > 0.0f )
        {
          xMin = std::min( xMin, x );
          xMax = std::max( xMax, x );
          yMin = std::min( yMin, y );
          yMax = std::max( yMax, y );
        }
      }

    const int w = xMax - xMin + 1;
    const int h = yMax - yMin + 1;

    cv::Mat result = cv::Mat::zeros( w, h, CV_32FC1 );
    image(cv::Rect(xMin, yMin, w, h)).copyTo( result );

    return result;
  }
}

void testFullConvergenceImages( bd::Border & border, std::string const & directory )
{
  const int height = (int)border.itsGridHeight * border.itsResolutionScaling;
  const int width = (int)border.itsGridWidth * border.itsResolutionScaling;

  cv::Mat mask = cv::Mat::zeros( height, width, CV_32FC1 );

  // Get images
  auto imagePaths = dir(directory, ".png");

  struct Image
  {
    cv::Mat image; //!< The image
    double x, y;   //!< top left XY coordinate of subwindow
  };

  //! Get a random top left xy location to paste the image
  auto randomXY = [&]( cv::Mat const & image )
  {
    return std::make_pair( bd::uniformInt( -image.cols + 1, width - 1 )(),
                           bd::uniformInt( -image.rows + 1, height - 1 )() );

  };

  //! Valid if any overlap between presentation image and canvas
  auto isValid = [&]( cv::Mat const & image, int x, int y )
  {
    const bool xValid = (x < width)  && (x + image.cols > 0);
    const bool yValid = (y < height) && (y + image.rows > 0);
    return xValid && yValid;
  };

  //! image is "good" if enough pixels are on
  auto isGood = [&]( cv::Mat const & image )
  {
    const double threshold = 10;

    if( cv::sum( image )[0] >= threshold )
      return true;
    else
      return false;
  };

  //! Load a random image from the collection, rotate and crop it
  auto getRandomImage = [&]()
  {
    Image im;
    const std::string fileAsString = imagePaths[bd::uniformInt(0, (int)imagePaths.size() - 1)()].string();
    cv::Mat temp = loadImage( fileAsString ).second;

    do
    {
      const double deg = bd::uniformReal( 0.0, 360.0 )();
      rotate2D( temp, im.image, deg );
      im.image = autoCrop(im.image);
      std::tie( im.x, im.y ) = randomXY(im.image);
    }
    while( !isValid( im.image, im.x, im.y ) );
    return im;
  };

  // choose a random direction in radians, 0 to 2pi
  auto randomDxDy = []()
  {
    const double theta = bd::uniformReal( 0.0, 2 * M_PI )();
    return cv::Point2d( std::cos(theta), -std::sin(theta) );
  };

  //! draws the image to im.x, im.y in a blank environment
  auto renderImage = [&]( Image const & im )
  {
    cv::Mat render = cv::Mat::zeros( height, width, CV_32FC1 );

    // im.x/y are in canvas relative coordinates
    //  negative values mean we're moving the image to the left/up
    //  the largest positive value for these is the extremity of the canvas (simulation)
    // This function figures out where to grab from the image and where to paste in the canvas

    const int imX = im.x;
    const int imY = im.y;

    const int leftCrop = std::max(-imX, 0);
    const int upCrop   = std::max(-imY, 0);
    const int numCols = std::min(im.image.cols, width - imX) - leftCrop;
    const int numRows = std::min(im.image.rows, height - imY) - upCrop;

    const auto fromRect = cv::Rect( leftCrop, upCrop, numCols, numRows );
    const auto destRect = cv::Rect( imX + leftCrop, imY + upCrop, numCols, numRows );

    im.image( fromRect ).copyTo( render( destRect ) );

    return render;
  };

  // main loop
  const int numPresentIters = 9; // how long to show each square at each location
  const int numOverallIters = 100000;

  for( int i = 0; i < numOverallIters; ++i )
  {
    auto image = getRandomImage();
    auto dxdy = randomDxDy() * border.itsResolutionScaling;
    bool prevGood = false;

    // present square and move it steps in some direction
    while( true )
    {
      cv::Mat stimulus = renderImage( image );

      if( isGood( stimulus ) )
      {
        bd::test::presentStimulus( border, stimulus, numPresentIters );
        prevGood = true;

        //{ // DEBUG
        //  bd::showImage(stimulus, false, true, "snarf");
        //  showConnections( border.itsGNResponses, false, false, "grouping" );
        //  showConnections( border.itsBOResponses[0], border.itsGNResponses, false, "bo_0_" );
        //  showConnections( border.itsBOResponses[1], border.itsGNResponses, false, "bo_45_" );
        //  showConnections( border.itsBOResponses[2], border.itsGNResponses, false, "bo_90_" );
        //  showConnections( border.itsBOResponses[3], border.itsGNResponses, false, "bo_135_" );
        //} // end DEBUG
      }
      else if( prevGood )
      {
        bd::test::doMask( border, mask );
        prevGood = false;
      }

      // translate along random direction
      image.x += dxdy.x;
      image.y += dxdy.y;

      if( !isValid( image.image, image.x, image.y ) )
        break;
    }

    bd::test::doMask( border, mask );

    const int numInspectIters = Options.iters;
    if( i % numInspectIters == 0 && i != 0 )
    {
      bd::test::saveBD( border, Options.output + "_" + std::to_string(i) + "_full_converge_images.cereal" );
      if( Options.display )
      {
        showConnections( border.itsGNResponses, false );
        displayPolarity(border.itsBOResponses[0], border.itsGNResponses, false) ;
        showConnections( border.itsBOResponses[0], border.itsGNResponses, false );
      }

      std::cout << "=================================" << std::endl;
      std::cout << "=================================" << std::endl;
      std::cout << "Done with " << i << " iterations " << std::endl;
      std::cout << "=================================" << std::endl;
      std::cout << "=================================" << std::endl;
    }
    else
      std::cout << "Counter: " << (i % numInspectIters) << std::endl;
  }
}
