#include "GaborTests.H"
#include "../src/Generate.H"
#include "../src/Util.H"

void gaborOrientationTest( bd::Gabor & gabor )
{
  auto makeStimulus = []( double radians )
  {
    cv::Mat stimulus = cv::Mat::zeros( 40, 40, CV_32FC1 );

    auto shape = bd::generateLine( 10, 20, 21 );
    //auto shape = bd::generateRectangle( 10, 10, 10, 10 );
    shape = bd::rotateShape( shape, radians );

    bd::drawShape( stimulus, shape, 1.0, false );

    return stimulus;
  };

  // 360 x orientation x scale
  std::vector<std::vector<std::vector<double>>> sums( 180 );
  for( auto & v : sums )
  {
    v.resize( gabor.itsNumOrientations );
    for( auto & vv : v )
      vv.resize( gabor.itsNumScales );
  }

  for( size_t i = 0; i < 180; ++i )
  {
    const double rad = i * (2 * M_PI) / 360.0;
    auto stimulus = makeStimulus( rad );
    auto results = gabor.getFilterResponses( stimulus );
    bd::showImage( stimulus, false, true, "stimulus" );

    for( size_t o = 0; o < results.size(); ++o )
      for( size_t s = 0; s < results[o].size(); ++s )
        sums[i][o][s] = cv::sum( results[o][s] )[0];
  }

  auto doSum = [&](size_t o, size_t s)
  {
    double sum = 0.0;
    for( const auto & orientation : sums )
      sum += orientation[o][s];

    return sum;
  };

  for( size_t o = 0; o < gabor.itsNumOrientations; ++o )
  {
    const float theta = o * ( M_PI / static_cast<float>( gabor.itsNumOrientations ) );
    std::cout << "Orientation: " << rad2deg( theta ) << std::endl;

    for( size_t s = 0; s < gabor.itsNumScales; ++s )
      std::cout << "   Scale[" << s << "]: " << doSum(o, s) << std::endl;
  }

}

void fftTests( cv::Mat const & image, bd::Gabor & gabor )
{
  // Test forward FFT followed by inverse FFT results in the same image
  int sz = bd::uniformInt( 1, 99 )(); if( sz % 2 == 0 ) ++sz;
  cv::Size filterSize( sz, sz );

  // Forward dft
  auto paddedSize = bd::getPaddedSize( image.size(), filterSize );
  cv::Mat padded  = bd::padImageFFT( image, filterSize, paddedSize );
  cv::dft( padded, padded );

  // inverse DFT
  cv::Mat result;
  cv::idft( padded, result );

  std::vector<cv::Mat> realImag(2);
  cv::split( result, realImag );

  auto real = bd::uncropFFT( realImag[0], image.size() );
  auto imag = bd::uncropFFT( realImag[1], image.size() );

  //bd::showImage( image, false, true, "before" );
  //bd::showImage( real, false, true, "real" );
  //bd::showImage( imag, false, true, "imag" );

  auto results = gabor.getFilterResponses( image, bd::Gabor::ResponseType::real );
  bd::showImage( image, false, true, "input" );

  for( auto const & vec : results )
    for( const auto & im : vec )
      bd::showImage( im );
}

void centerSurroundTest()
{
  auto makeStimulus = []( double radians )
  {
    cv::Mat stimulus = cv::Mat::zeros( 40, 40, CV_32FC1 );

    auto shape = bd::generateLine( 10, 20, 21 );
    //auto shape = bd::generateRectangle( 10, 10, 10, 10 );
    shape = bd::rotateShape( shape, radians );

    bd::drawShape( stimulus, shape, 1.0, false );

    return stimulus;
  };

  // 360 x orientation x scale
  std::vector<std::vector<double>> sums( 180 );
  for( auto & v : sums )
    v.resize( 2 );

  for( size_t i = 0; i < 180; ++i )
  {
    const double rad = i * (2 * M_PI) / 360.0;
    auto stimulus = makeStimulus( rad );

    // Convolution for onOff/offOn
    cv::Mat onOffKernel = (cv::Mat_<float>( 3, 3 ) <<
        -.125, -.25, -.125,
        -.25, 1.5, -.25,
        -.125, -.25, -.125);
    cv::Mat offOnKernel = -onOffKernel;

    cv::Mat onOffResult, offOnResult;

    cv::filter2D( stimulus, onOffResult, -1, onOffKernel );
    cv::filter2D( stimulus, offOnResult, -1, offOnKernel );

    for( auto iter = onOffResult.begin<float>(), end = onOffResult.end<float>(); iter != end; ++iter )
      *iter = bd::clamp(*iter);

    for( auto iter = offOnResult.begin<float>(), end = offOnResult.end<float>(); iter != end; ++iter )
      *iter = bd::clamp(*iter);

    bd::showImage( stimulus, false, true, "stimulus" );

    sums[i][0] = cv::sum( onOffResult )[0];
    sums[i][1] = cv::sum( offOnResult )[0];
  }

  auto doSum = [&](size_t i)
  {
    double sum = 0.0;
    for( const auto & v : sums )
      sum += v[i];

    return sum;
  };

  std::cout << "OnOff: " << doSum(0) << std::endl;
  std::cout << "OffOn: " << doSum(1) << std::endl;
}
