#include "Util.H"
#include "../src/V1.H"
#include "../src/Combined.H"
#include "../src/Generate.H"

#include <fstream>
#include <iostream>
#include <cereal/archives/binary.hpp>

template <class Network>
void bd::test::doMask( Network & network, cv::Mat const & mask, int numIters )
{
  network.onset( mask );
  network.iterate( numIters );
}

template void bd::test::doMask<bd::Border>( bd::Border &, cv::Mat const &, int );
template void bd::test::doMask<bd::V1>( bd::V1 &, cv::Mat const &, int );
template void bd::test::doMask<bd::Combined>( bd::Combined &, cv::Mat const &, int );

template <class Network>
void bd::test::presentStimulus( Network & network, cv::Mat const & stimulus, int numIters,
                                std::function<void()> postFunc )
{
  network.onset( stimulus );
  network.iterate( numIters );
  postFunc();
}

template void bd::test::presentStimulus<bd::Border>( bd::Border &, cv::Mat const &, int, std::function<void()> );
template void bd::test::presentStimulus<bd::V1>( bd::V1 &, cv::Mat const &, int, std::function<void()> );
template void bd::test::presentStimulus<bd::Combined>( bd::Combined &, cv::Mat const &, int, std::function<void()> );

void bd::test::presentSquare( bd::Border & border, int numIters, int x, int y, int sideLength, std::function<void()> postFunc )
{
  presentRectangle( border, x, y, sideLength, sideLength, numIters, postFunc );
};

void bd::test::presentRectangle( bd::Border & border, int numIters, int x, int y, int lengthX, int lengthY, std::function<void()> postFunc )
{
  const auto height = border.itsGridHeight;
  const auto width = border.itsGridWidth;

  auto localimage = bd::generateRectangle( height, width, x, y, lengthX, lengthY );

  border.onset( localimage );
  border.iterate(numIters);

  postFunc();
};

std::tuple<double, double, double, double> bd::test::learnedPolarityScore( Sheet<BorderOwnershipNeuron> const & sheet,
    std::string const & prefix )
{
  double totalScore = 0, average = 0, stddev = 0;
  int count = 0;

  const auto height = sheet.height();
  const auto width = sheet.width();
  const int cropX = width  * 0.34;
  const int cropY = height * 0.34;

  std::vector<double> scores;

  // Compute average
  for( const auto & cell : sheet )
  {
    const auto & left = cell[0];
    const auto & right = cell[1];

    double x, y;
    std::tie(x, y) = left->position();

    if( (x >= cropX ) &&
        (y >= cropY ) &&
        (x < (width - cropX)) &&
        (y < (height - cropY)) )
    {
      auto leftPair = display::getDirectionValues( left );
      auto rightPair = display::getDirectionValues( right );

      const auto score = std::abs((leftPair.first - leftPair.second) + (rightPair.second - rightPair.first));

      scores.push_back( score );

      totalScore += score;
      ++count;
    }
  }

  average = totalScore / static_cast<double>( count );

  // Compute Standard Deviation
  for( const auto & cell : sheet )
  {
    const auto & left = cell[0];
    const auto & right = cell[1];

    double x, y;
    std::tie(x, y) = left->position();

    if( (x >= cropX ) &&
        (y >= cropY ) &&
        (x < (width - cropX)) &&
        (y < (height - cropY)) )
    {
      auto leftPair = display::getDirectionValues( left );
      auto rightPair = display::getDirectionValues( right );

      const auto score = std::abs((leftPair.first - leftPair.second) + (rightPair.second - rightPair.first));

      stddev += std::pow( score - average, 2.0 );
    }
  }

  stddev = std::sqrt( stddev / static_cast<double>(count) );

  // Compute median
  std::nth_element( scores.begin(), scores.begin() + scores.size()/2, scores.end() );
  const double median = scores[scores.size()/2];

  // Dump scores to text file
  {
    std::ofstream out( prefix + "_polarity_scores.txt" );
    for( auto s : scores )
      out << std::to_string(s) << std::endl;
  }

  return std::make_tuple(totalScore, average, stddev, median);
}

std::vector<double> bd::test::getLearnedPolarityScores( const Border & border )
{
  std::vector<std::vector<double>> scores;
  for( const auto & orientation : border.itsBOResponses )
  {
    scores.emplace_back();
    auto &ref = scores.back();
    ref.push_back( std::get<0>( learnedPolarityScore( orientation ) ) );
  }

  return {};
}

std::string debugGetImageType2(int number)
{
    // find type
    int imgTypeInt = number%8;
    std::string imgTypeString;

    switch (imgTypeInt)
    {
        case 0:
            imgTypeString = "8U";
            break;
        case 1:
            imgTypeString = "8S";
            break;
        case 2:
            imgTypeString = "16U";
            break;
        case 3:
            imgTypeString = "16S";
            break;
        case 4:
            imgTypeString = "32S";
            break;
        case 5:
            imgTypeString = "32F";
            break;
        case 6:
            imgTypeString = "64F";
            break;
        default:
            break;
    }

    // find channel
    int channel = (number/8) + 1;

    std::stringstream type;
    type<<"CV_"<<imgTypeString<<"C"<<channel;

    return type.str();
}

//! Checks if the angle is within +/- delta of the target
bool angleInBounds( double angle, double target, double delta )
{
  auto normalize = [](double x)
  {
    x = std::fmod(x, 360.0);
    return x >= 180.0 ? x - 360.0 : x < -180.0 ? x + 360.0 : x;
  };

  const double diff = normalize(angle - target);
  return diff < delta && diff > -delta;
}

std::tuple<double, double, double, double, double, double>
bd::test::activePolarityScore( cv::Mat const & groundTruth, cv::Mat const & allDirections, const int resolutionScaling )
{
  cv::Mat gt_HSV = groundTruth;
  cv::Mat ad_HSV = allDirections;

  cv::cvtColor( gt_HSV, gt_HSV, CV_BGR2HSV );
  cv::cvtColor( ad_HSV, ad_HSV, CV_BGR2HSV );

  std::vector<double> scores;
  double minScore = 0.0, maxScore = 0.0;
  double totalScore = 0.0;
  int count = 0;
  int countWithin180 = 0; // used to calculate number of answers that are facing the same direction
                          // (plus or minus 90 degrees) of the ground truth

  for( int row = 0; row < allDirections.rows; ++row )
  {
    const int y = row * resolutionScaling;

    for( int col = 0; col < allDirections.cols; ++col )
    {
      const int x = col * resolutionScaling;

      // Ignore values where there is no ground truth available
      auto gtPixel = gt_HSV.at<cv::Vec3f>( y, x );
      if( gtPixel[2] == 0 )
        continue;

      const double gtDegrees = gtPixel[0];
      const double adDegrees = ad_HSV.at<cv::Vec3f>( row, col )[0];
      const double score = std::abs( gtDegrees - adDegrees );

      if( angleInBounds( adDegrees, gtDegrees, 90.0 ) )
        ++countWithin180;

      scores.push_back(score);
      minScore = std::min( minScore, score );
      maxScore = std::max( maxScore, score );
      totalScore += score;
      ++count;
    }
  }

  const double averageScore = totalScore / count;
  double stddev = 0.0;

  // compute standard deviation
  for( int row = 0; row < allDirections.rows; ++row )
  {
    const int y = row * resolutionScaling;

    for( int col = 0; col < allDirections.cols; ++col )
    {
      const int x = col * resolutionScaling;

      // Ignore values where there is no ground truth available
      auto gtPixel = gt_HSV.at<cv::Vec3f>( y, x );
      if( gtPixel[2] == 0 )
        continue;

      const double score = std::abs( gtPixel[0] - ad_HSV.at<cv::Vec3f>( row, col )[0] );

      stddev += std::pow( score - averageScore, 2.0 );
    }
  }

  stddev = std::sqrt( stddev / static_cast<double>(count) );

  // Compute median
  std::nth_element( scores.begin(), scores.begin() + scores.size()/2, scores.end() );
  const double median = scores[scores.size()/2];

  const double score180 = countWithin180 / static_cast<double>(count);

  std::cout << "180Score " << score180 << ", AvgScore " << averageScore << ", min/max: " << minScore << "/" << maxScore << ", std: " << stddev << ", med: " << median << std::endl;

  return std::make_tuple(score180, averageScore, minScore, maxScore, stddev, median);
}

template <class Network>
void bd::test::loadBD( Network & network, std::string const & name )
{
  std::cout << "Loading " << name << std::endl;
  std::ifstream file(name, std::ios::binary );
  cereal::BinaryInputArchive ar( file );
  ar( network );
}

template void bd::test::loadBD<bd::Border>( bd::Border &, std::string const & );
template void bd::test::loadBD<bd::V1>( bd::V1 &, std::string const & );
template void bd::test::loadBD<bd::Combined>( bd::Combined &, std::string const & );

template <class Network>
void bd::test::saveBD( Network & network, std::string const & name )
{
  std::cout << "Saving " << name << std::endl;
  std::ofstream file(name, std::ios::binary );
  cereal::BinaryOutputArchive ar( file );
  ar( network );
}

template void bd::test::saveBD<bd::Border>( bd::Border &, std::string const & );
template void bd::test::saveBD<bd::V1>( bd::V1 &, std::string const & );
template void bd::test::saveBD<bd::Combined>( bd::Combined &, std::string const & );

void bd::test::dumpBO( bd::Border & border, std::string const & prefix )
{
  const size_t numFiles = border.itsV1NumOrientations * 2;
  const auto & sheets = border.itsBOResponses;
  std::vector<std::stringstream> streams( numFiles );

  // Goes from orientation, offset (0/1 for left/right) to index in
  // vector of streams
  auto index = [&](int o, int offset)
  { return (o * 2) + offset; };

  // Fill streams with space delimited values
  for( size_t o = 0; o < sheets.size(); ++o )
  {
    size_t counter = 0;
    for( const auto & cell : sheets[o] )
    {
      // For each sheet, we'll create two outputs.
      //  One will correspond to the activation of all neurons that are 'green',
      //  the other for all neurons that are 'red'
      //
      //  A neuron is assigned to the red or green side based on its maximum
      //  direction value being larger than the maximum of its pair.  The pair
      //  is then automatically assigned the opposite side.
      const auto & left = cell[0];
      const auto & right = cell[1];

      const auto valsLeft = display::getDirectionValues(left);
      const auto valsRight = display::getDirectionValues(right);

      bool maxIsLeft = false;
      auto maxVals = valsRight;

      if(std::max(valsLeft.first, valsLeft.second) >= std::max(valsRight.first, valsRight.second))
      {
        maxIsLeft = true;
        maxVals = valsLeft;
      }

      const int maxIndex = maxVals.first >= maxVals.second ? 0 : 1;
      const int leftIndex = maxIsLeft ? maxIndex : 1 - maxIndex;
      const int rightIndex = 1 - leftIndex;

      streams[index(o, leftIndex)] << left->activation() << " ";
      streams[index(o, rightIndex)] << right->activation() << " ";

      if(++counter == border.itsGridWidth)
      {
        streams[index(o, 0)] << std::endl;
        streams[index(o, 1)] << std::endl;
        counter = 0;
      }
    }
  }

  // Write out files
  for( size_t o = 0; o < sheets.size(); ++o )
  {
    std::ofstream out1( prefix + "_bo_" + std::to_string(o) + "_1.txt" );
    std::ofstream out2( prefix + "_bo_" + std::to_string(o) + "_2.txt" );

    out1 << streams[index(o, 0)].rdbuf();
    out2 << streams[index(o, 1)].rdbuf();
  }
}

template <class NeuronT>
void bd::test::dumpWeights( Sheet<NeuronT> & sheet, std::string const prefix )
{
  std::ofstream outFF( prefix + "_FF.txt" );
  std::ofstream outFB( prefix + "_FB.txt" );
  std::ofstream outLAT( prefix + "_LAT.txt" );

  for( const auto & cell : sheet )
  {
    for( const auto & neuron : cell )
    {
      neuron->printWeights( outFF, [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedforward ); } );
      neuron->printWeights( outFB, [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Feedback ); } );
      neuron->printWeights( outLAT, [](bd::Input const & in){ return in.hasFlag( bd::ConnectionFlags::Lateral ); } );
    }
  }
}

template void bd::test::dumpWeights( Sheet<EdgeNeuron> &, std::string const );
