#ifndef TEST_UTIL_H_
#define TEST_UTIL_H_

#include "../src/Border.H"
#include <tuple>

namespace bd
{
  namespace test
  {
    //! Apply a mask to the network
    template <class Network>
    void doMask( Network & network, cv::Mat const & mask, int numIters = 30 );

    //! Present some stimulus and optionally call something afterwards
    template <class Network>
    void presentStimulus( Network & network, cv::Mat const & stimulus, int numIters,
                          std::function<void()> postFunc = [](){} );

    //! Present a square for some number of iterations
    /*! Presents a square with its upper-left corner exactly on the coordinates provided and optionally
        executes a function after completion */
    void presentSquare( bd::Border & border, int numIters,
                        int x, int y, int sideLength,
                        std::function<void()> postFunc = [](){} );

    //! Present a rectangle for some number of iterations
    /*! Presents a rectangle with its upper-left corner exactly on the coordinates provided and optionally
        executes a function after completion */
    void presentRectangle( bd::Border & border, int numIters,
                           int x, int y, int lengthX, int lengthY,
                           std::function<void()> postFunc = [](){} );

    //! Computes a score indicating how well the BO sheet has learned its connections
    /*! @return The total score, the average, the standard deviation, and the median */
    std::tuple<double, double, double, double> learnedPolarityScore( Sheet<BorderOwnershipNeuron> const & sheet, std::string const & prefix = "" );

    //! Computes scores for every orientation
    std::vector<double> getLearnedPolarityScores( const Border & border );

    //! Computes a score indicating how well the current BO polarity assignments match the ground truth
    /*! @param groundTruth The ground truth image
        @param allDirections The visualization of all polarity directions
        @return A tuple of {score180, averageScore, minScore, maxScore, stddev, median} */
    std::tuple<double, double, double, double, double, double>
    activePolarityScore( cv::Mat const & groundTruth, cv::Mat const & allDirections, const int rotationScaling );

    //! Load a border ownership network
    template <class Network>
    void loadBD( Network & network, std::string const & name = "data.cereal" );

    //! Load a border ownership network
    template <class Network>
    void saveBD( Network & network, std::string const & name = "data.cereal" );

    //! Dumps all BO activations to disk in plain text files
    /*! Two files per orientation of BO neurons will be created, with values corresponding
        to their activation, signed depending on which direction the neuron prefers.
        Positive is used for green, which rotates counter-clockwise by orientation increments */
    void dumpBO( bd::Border & border, std::string const & prefix );

    //! Dump all weights to disk in plain text files
    /*! One file will be created per type of connection, which will contain
        all weights coming into one neuron per line */
    template <class NeuronT>
    void dumpWeights( Sheet<NeuronT> & sheet, std::string const prefix );
  } // namespace test
} // namespace bd

//! C++11 version of sprintf, dumps directly into a string
template< typename... Args > inline
std::string string_sprintf( const char* format, Args... args )
{
  int length = std::snprintf( nullptr, 0, format, args... );
  assert( length >= 0 );

  char* buf = new char[length + 1];
  std::snprintf( buf, length + 1, format, args... );

  std::string str( buf );
  delete[] buf;
  return std::move(str);
}

#endif // TEST_UTIL_H_
